/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file dist_list.h
 * @brief The header for parallel skeleton on distributed list.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 * @author Kiminori MATSUZAKI (kmatsu@ipl.t.u-tokyo.ac.jp)
 */

#ifndef __DIST_LIST_H__
#define __DIST_LIST_H__

#include <memory.h>
#include <mpi.h>
#include "../util/skeleton.h"
#include "../util/template_mpitype.h"

typedef enum {
  TAG_OSTREAM,
  TAG_DISTRIBUTE,
  TAG_GATHER, 
  TAG_REDUCE1, 
  TAG_REDUCE2, 
  TAG_PRINT1, 
  TAG_PRINT2, 
  TAG_SCAN1, 
  TAG_SCAN2,
  TAG_SCANR1,
  TAG_SCANR2,
  TAG_ACCUMULATE1_1,
  TAG_ACCUMULATE1_2,
  TAG_ACCUMULATE2_1,
  TAG_ACCUMULATE2_2,
  TAG_SHIFTR1,
  TAG_SHIFTR2,
  TAG_SHIFTL1,
  TAG_SHIFTL2,
  TAG_CATAJ1,
  TAG_CATAJ2,
} dist_list_tag_t;

template< typename A >
class template_mpitype;

class list_skeletons;

template<typename A>
class dist_list
{
  friend class list_skeletons;

private:
  /** @brief The total size of list. */
  int global_size;

  /** @brief The MPI type for the elements. */
  MPI_Datatype mpitype;

  /** @brief The really allocated data on the processes. */
  A *data;

  /* following three fields (@c begin, @c end, @c local_size) are computed */
  /** @brief The first index in the global list. */
  int begin;
  /** @brief The (last + 1) index in the global list. */
  int end;
  /** @brief The size of locally handled list. */
  int local_size;

  //------------------------------------------------------------------------
  // Access methods

  /** @brief Returns the total size of list */
public:
  int get_global_size( ) const;
  /**
   * @brief Local access to data
   * @param index : local index [ in ]
   * @return The reference to local data
   */
public:
  A& at( int index );
  /**
   * @brief Constant local access to data
   * @param index : local index [ in ]
   * @return The constant reference to local data
   */
public:
  const A& at( int index ) const;

  /**
   * @brief Computes the first index. 
   * @param rank : the rank to obtain the first index for.
   */
public:
  int get_begin_index( int rank ) const;

  /**
   * @brief Computes the (last + 1) index. 
   * @param rank : the rank to obtain the (last + 1) index for.
   */
public:
  int get_end_index( int rank ) const;

  /**
   * @brief Computes the local size.
   * @param rank : the rank to obtain the local size for.
   */
public:
  int get_local_size( int rank ) const;
  
  //------------------------------------------------------------------------
  // Constructors

  /**
   * @brief Makes an empty list.
   * @param list_size_ : The total size of list. [ in ]
   */
public:
  dist_list( int list_size_ );

  /**
   * @brief Makes an list by distributing the @c list on master process. 
   * @param array_size_ : The total size of array. [ in ]
   * @param array_ : The array to distribute. [ in, master ]
   */
public:
  dist_list( int array_size_,
	     const A *array_ );

  /**
   * @brief Makes an list by generating with @c func. 
   * @param f : The generating function. [ in ]
   * @param list_size_ : The total size of list. [ in ]
   */
public:
  template< typename F >
  dist_list( const F &f,
	     int list_size_ );

  /**
   * @brief Copy constructor.
   * @param as : The original list.
   */
public:
  dist_list( const dist_list< A > &as );

  /**
   * @brief The assignment operator.
   * @param as : The list to copy from.
   */
public:
  dist_list< A > &operator=( const dist_list< A > &as );

  /**
   * @brief Initializes the fields. 
   *
   * @attention This function should be called from the constructor only.
   * @attention Before calling this function, the field @c global_size must
   *            be correctly set. 
   */
private:
  void init( );

  //------------------------------------------------------------------------
  // Destructor
  
public:
  ~dist_list();
};

#include "dist_list.tpp"
#endif // __DIST_LIST_H__
