/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file matrix2.tpp
 * @brief The template implementation for 2x2 matrices.
 * @author Kiminori MATSUZAKI
 */

#ifdef _MATRIX2_H_
#ifndef _MATRIX2_TPP_
#define _MATRIX2_TPP_

#include <assert.h>

template< typename A >
Matrix2< A >::Matrix2( A a11, A a12, A a21, A a22 )
{
  at( 0, 0 ) = a11;
  at( 0, 1 ) = a12;
  at( 1, 0 ) = a21;
  at( 1, 1 ) = a22;
}

template< typename A >
Matrix2< A >::Matrix2( bool )
{
  at( 0, 0 ) = 0;
  at( 0, 1 ) = 0;
  at( 1, 0 ) = 0;
  at( 1, 1 ) = 0;
}
  
template< typename A >
inline A&
Matrix2< A >::at( int x, int y )
{
  assert( x >= 0 );
  assert( x < 2 );
  assert( y >= 0 );
  assert( y < 2 );

  return data[ x ][ y ];
}

template< typename A >
inline const A&
Matrix2< A >::at( int x, int y ) const
{
  assert( x >= 0 );
  assert( x < 2 );
  assert( y >= 0 );
  assert( y < 2 );

  return data[ x ][ y ];
}

template< typename A >
inline Matrix2< A >
Matrix2< A >::operator+( const Matrix2< A >& other ) const
{
  return Matrix2< A >( this->at( 0, 0 ) + other.at( 0, 0 ),
		       this->at( 0, 1 ) + other.at( 0, 1 ),
		       this->at( 1, 0 ) + other.at( 1, 0 ),
		       this->at( 1, 1 ) + other.at( 1, 1 ) );
}

template< typename A >
inline Matrix2< A >
Matrix2< A >::operator-( const Matrix2< A >& other ) const
{
  return Matrix2< A >( this->at( 0, 0 ) - other.at( 0, 0 ),
		       this->at( 0, 1 ) - other.at( 0, 1 ),
		       this->at( 1, 0 ) - other.at( 1, 0 ),
		       this->at( 1, 1 ) - other.at( 1, 1 ) );
}

template< typename A >
inline Matrix2< A >
Matrix2< A >::operator*( const Matrix2< A >& other ) const
{
  const A a11 = this->at(0,0)*other.at(0,0) + this->at(0,1)*other.at(1,0);
  const A a12 = this->at(0,0)*other.at(0,1) + this->at(0,1)*other.at(1,1);
  const A a21 = this->at(1,0)*other.at(0,0) + this->at(1,1)*other.at(1,0);
  const A a22 = this->at(1,0)*other.at(0,1) + this->at(1,1)*other.at(1,1);
  return Matrix2< A >( a11, a12, a21, a22 );
}

template< typename A >
inline Matrix2< A >
Matrix2< A >::operator*( A val ) const
{
  return Matrix2< A >( val * this->at( 0, 0 ),
		       val * this->at( 0, 1 ),
		       val * this->at( 1, 0 ),
		       val * this->at( 1, 1 ) );
}

template< typename A >
inline bool
Matrix2< A >::operator==( const Matrix2< A >& other ) const
{
  return ( this->at( 0, 0 ) == other.at( 0, 0 ) &&
	   this->at( 0, 1 ) == other.at( 0, 1 ) &&
	   this->at( 1, 0 ) == other.at( 1, 0 ) &&
	   this->at( 1, 1 ) == other.at( 1, 1 ) );
}

/* public, static */
template< typename A >
inline const MPI_Datatype
Matrix2< A >::get_mpi_type( ) {
  static bool is_already_called = false;
  static MPI_Datatype typeval;
  if ( !is_already_called ) {
    is_already_called = true;
    
    Matrix2< A > temp;
    int blength_list[ 2 ] = { 2, 2 };
    MPI_Datatype datatype_list[ 2 ] = { template_mpitype< A >::type( ),
					template_mpitype< A >::type( ) };
    MPI_Aint displacement_list[ 2 ];
    MPI_Aint start_address, address;
    MPI_Address( &temp, &start_address );
    
    MPI_Address( &temp.data[ 0 ][ 0 ], &address );
    displacement_list[ 0 ] = address - start_address;
    
    MPI_Address( &temp.data[ 1 ][ 0 ], &address );
    displacement_list[ 1 ] = address - start_address;
    
    MPI_Datatype newtype;
    MPI_Type_struct( 2, blength_list, displacement_list,
		     datatype_list, &newtype );
    MPI_Type_commit( &newtype );
    
    typeval = newtype;
  }
  return typeval;
}

template< typename A >
std::ostream&
operator<<( std::ostream& s, const Matrix2< A >& x )
{
  s << x.at( 0, 0 ) << " " << x.at( 0, 1 ) << " "
    << x.at( 1, 0 ) << " " << x.at( 1, 1 );
  return s;
}

template< typename A >
inline std::istream& operator>>( std::istream& s, Matrix2< A >& x )
{
  s >> x.at( 0, 0 ) >> x.at( 0, 1 )
    >> x.at( 1, 0 ) >> x.at( 1, 1 );
  return s;
}

#endif // _MATRIX2_TPP_
#endif // _MATRIX2_H_

