/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file Shadower.h
 * @brief Template utility for using shadow(i.e. an element without entity).
 * @author Kento EMOTO
 */
#ifndef _SHADOWER_H_
#define _SHADOWER_H_
#include <utility>

/**
 * @brief A default template class to make shadow.
 *
 * This is default implementation and invokes A::isShadowable() and A::shadow(A&);
 *
 * Instances for some primitive types are written in this file.
 * If you use other primitive types, you should make instances for them.
 * For classes used in skeletons, i.e. dist_pair and matrix, 
 * you don't have to are also written in this file.
 * If you use classes except those used in skeletons, i.e. dist_pair and matrix, , you should do either of the following:
 *  1. add members @c isShadowable() and @c shadow() which are used in default Shadower,
 *  2. write an instance of Shadower for the class.
 */
template <typename A>
class Shadower {
public:
	/**
	 * @brief Make sh shadow of org.
	 * 
	 * Default implementation invokes just A::shadow.
	 * @param sh is made to be shadow of @a org.
	 * @param org the object to be shadowed.
	 * @see matrix<A>::shadow()
	 */
	static void shadow(const A &org, A* sh){ org.shadow(sh); }
	/**
	 * @brief Returns whether class A is shadowable.
	 * @retval true class A is shadowable.
	 * @retval false class A is not shadowable. It should have entity.
	 * @see matrix<A>::isShadowable()
	 */
	static const bool isShadowable(){ return A::isShadowable(); }
};

/**
 * @brief Instance of Shadower for std::pair
 */
template <typename A,  typename B>
class Shadower<std::pair<A,B> > {
public:
	/**
	 * @brief Just copy.
	 */
	static void shadow(const std::pair<A,B> &org, std::pair<A,B>* sh){ *sh = org; }
	/**
	 * @brief Primitives are not shadowable.
	 */
	static const bool isShadowable(){ return false; }
};
/**
 * @brief Instance of Shadower for int
 */
template <>
class Shadower<int> {
public:
	/**
	 * @brief Just copy.
	 */
	static void shadow(const int &org, int* sh){ *sh = org; }
	/**
	 * @brief Primitives are not shadowable.
	 */
	static const bool isShadowable(){ return false; }
};
/**
 * @brief Instance of Shadower for unsigned int
 */
template <>
class Shadower<unsigned int> {
public:
	/**
	 * @brief Just copy.
	 */
	static void shadow(const unsigned int &org, unsigned int* sh){ *sh = org; }
	/**
	 * @brief Primitives are not shadowable.
	 */
	static const bool isShadowable(){ return false; }
};
/**
 * @brief Instance of Shadower for char
 */
template <>
class Shadower<char> {
public:
	/**
	 * @brief Just copy.
	 */
	static void shadow(const char &org, char *sh){ *sh = org; }
	/**
	 * @brief Primitives are not shadowable.
	 */
	static const bool isShadowable(){ return false; }
};
/**
 * @brief Instance of Shadower for unsigned char
 */
template <>
class Shadower<unsigned char> {
public:
	/**
	 * @brief Just copy.
	 */
	static void shadow(const unsigned char &org, unsigned char *sh){ *sh = org; }
	/**
	 * @brief Primitives are not shadowable.
	 */
	static const bool isShadowable(){ return false; }
};
/**
 * @brief Instance of Shadower for double
 */
template <>
class Shadower<double> {
public:
	/**
	 * @brief Just copy.
	 */
	static void shadow(const double &org, double *sh){ *sh = org; }
	/**
	 * @brief Primitives are not shadowable.
	 */
	static const bool isShadowable(){ return false; }
};
/**
 * @brief Instance of Shadower for float
 */
template <>
class Shadower<float> {
public:
	/**
	 * @brief Just copy.
	 */
	static void shadow(const float &org, float *sh){ *sh = org; }
	/**
	 * @brief Primitives are not shadowable.
	 */
	static const bool isShadowable(){ return false; }
};


#endif
