/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file dist_matrix.h
 * @brief The header of class @a dist_matrix, the data for distributed
 *        matrices and read/write functions. 
 * @author Kento EMOTO
 */

#ifndef __DIST_MATRIX_H__
#define __DIST_MATRIX_H__

#if 0   // for C++ bindings
#define SKEL_MPI_COMM_WORLD MPI::COMM_WORLD
#define SKEL_MPI_Request MPI::Request
#define SKEL_MPI_Wait(request) request.Wait()
#define SKEL_MPI_Send(buf, count, datatype, dest, tag, comm) comm.Send(buf, count, datatype, dest, tag)
#define SKEL_MPI_Recv(buf, count, datatype, source, tag, comm) comm.Recv(buf, count, datatype, source, tag)
#define SKEL_MPI_Isend( buf, count, datatype, dest, tag, comm, request) request = comm.Isend(buf, count, datatype, dest, tag)
#define SKEL_MPI_Irecv( buf, count, datatype, source, tag, comm, request) request = comm.Irecv(buf, count, datatype, source, tag)

#else           // for C 
#define SKEL_MPI_COMM_WORLD MPI_COMM_WORLD
#define SKEL_MPI_Request MPI_Request
#define SKEL_MPI_Wait(request) MPI_Wait (&request, MPI_STATUS_IGNORE)
#define SKEL_MPI_Send(buf, count, datatype, dest, tag, comm) MPI_Send(buf, count, datatype, dest, tag, comm)
#define SKEL_MPI_Recv(buf, count, datatype, source, tag, comm) MPI_Recv(buf, count, datatype, source, tag, comm, MPI_STATUS_IGNORE)
#define SKEL_MPI_Isend( buf, count, datatype, dest, tag, comm, request) MPI_Isend( (void*)(buf), count, datatype, dest, tag, comm, &request )
#define SKEL_MPI_Irecv( buf, count, datatype, source, tag, comm, request) MPI_Irecv( (void*)(buf), count, datatype, source, tag, comm, &request )
#endif


#include <mpi.h>
#include <iostream>
#include <iomanip>

#include "Serializer.h"
#include "matrix.h"
#include "template_mpitype.h"
#include "../util/skeleton.h"
/**
 * @brief Tags for communication in skeletons.
 */
typedef enum {
	TAG_NOTHING,             /**< nothing(not used).*/
	TAG_DISTRIBUTE,          /**< distribute(data). */
	TAG_DISTRIBUTE2,         /**< distribute(size). */
	TAG_GATHER1,             /**< gather (structure).*/
	TAG_GATHER2,             /**< gather (data).*/
	TAG_GATHER3,             /**< gather (size).*/
	TAG_PERMUTE1,            /**< permute/shift/rotate (structure). */
	TAG_PERMUTE2,            /**< permute/shift/rotate (data). */
	TAG_PERMUTE3,            /**< permute/shift/rotate (size). */
	TAG_SCAN12,              /**< scan (step 1.2). */
	TAG_SCAN12s,             /**< scan (step 1.2, size). */
	TAG_SCAN22,              /**< scan (step 2.2). */
	TAG_SCAN22s,             /**< scan (step 2.2, size). */
	TAG_REDUCE1,             /**< reduce (row). */
	TAG_REDUCE2,             /**< reduce (col). */
	TAG_REDUCE3,             /**< reduce (row, size). */
	TAG_REDUCE4,             /**< reduce (col, size). */
	TAG_ALLREDUCEROW1,          /**< allreduce_row. */
	TAG_ALLREDUCEROW2,          /**< allreduce_row (size). */
	TAG_ALLREDUCEROW3,          /**< allreduce_row. */
	TAG_ALLREDUCEROW4,          /**< allreduce_row (size). */
	TAG_ALLREDUCECOL1,          /**< allreduce_col. */
	TAG_ALLREDUCECOL2,          /**< allreduce_col (size). */
	TAG_ALLREDUCECOL3,          /**< allreduce_col. */
	TAG_ALLREDUCECOL4,          /**< allreduce_col (size). */
	TAG_GLOBAL_GET1,             /**< global_get (flag).*/
	TAG_GLOBAL_GET2,             /**< global_get (data).*/
	TAG_GLOBAL_GET22,             /**< global_get (size).*/
	TAG_GLOBAL_GET3,             /**< global_get (data).*/
	TAG_GLOBAL_GET32,             /**< global_get (size).*/
	TAG_GLOBAL_SET1,             /**< global_set (flag).*/
	TAG_GLOBAL_SET2,             /**< global_set (data).*/
	TAG_GLOBAL_SET22,             /**< global_set (size).*/
	TAG_GLOBAL_SET3,             /**< global_set (data).*/
	TAG_GLOBAL_SET32,             /**< global_set (size).*/
} dist_matrix_tag_t;

/**
 * @brief Datatype used to transmit serialized binary data.
 */
static const MPI_Datatype BYTE_TYPE = MPI_BYTE;


class matrix_skeletons;

/** @brief Structure of submatrix */
struct matrixStructure{
	int idxRow;          /**< Global row-index of the upper-left element */
	int idxCol;          /**< Global column-index of the upper-left element */
	int localRows;       /**< Size(#rows) of submatrix(actual) */
	int localCols;       /**< Size(#columns)  of submatrix(actual) */
};





/**
 * @brief A template class for handling matrix distributed on processors.
 *
 * The matrix is consists of three constructors 'abv'(above), 'bsd'(beside) and 'sing'(singleton).
 * So, in haskell notation, datatype matrix is as follows:
 * @code
 * data Matrix a = Sing a
 *         | Abv (Matrix a) (Matrix a)
 *         | Bsd (Matrix a) (Matrix a)
 * @endcode
 * A singleton matrix with an element E is 'Sing E' or '|E|' for brevity.
 * A matrix made by placing matrix B on the left of matrix C is 'Bsd B C' or 'B `Bsd` C'. The heights of B and C should be the same.
 * A matrix made by placing matrix B on the top of matrix C is 'Abv B C' or 'B `Abv` C'. The widths of B and C should be the same.
 * The constructors 'Bsd' and 'Abv' are associative and have abiding property:
 * '(B `Bsd` C) `Abv` (E `Bsd` F) = (B `Abv` E) `Bsd` (C `Abv` F)'.
 * Skeletons are defined as recursive functions on the structure.
 * Indices are starting from 0.
 *
 * @see matrix
 */
template< typename A >
class dist_matrix
{
  friend class matrix_skeletons; /**< The skeletons. */
private:
	////////////////////////////////////////////////////////////////////////
	// Member variables
private:
	/** @brief Rank of this process. */
	int myrank;	
	/** @brief Global size(#rows) of the matrix (@a n x @a m). */
	int n;				
	/** @brief Global size(#columns) of the matrix (@a n x @a m). */
	int m;
	/** @brief Size(#rows) of distributed submatrix (@a rows x @a cols). */
	int rows;		
	/** @brief Size(#columns) of distributed submatrix (@a rows x @a cols). */
	int cols;
	/** @brief Size(#rows) of distributed submatrix (actual) (@a localRows x @a localCols) for the case that n%rows!=0 or m%cols!=0. */
	int localRows;		
	/** @brief Size(#columns) of distributed submatrix (actual) (@a localRows x @a localCols) for the case that n%rows!=0 or m%cols!=0. */
	int localCols;
	/** @brief Total blocks in matrix. */
	int blocks;
	/** @brief Blocks in rows. */
	int bir;			
	/** @brief Blocks in cols. */
	int bic;
	/** @brief Local elements in 1D array. */
	A *ele;
	/** @brief Pointer to each row/column. */
	A **idxEle;
	/** @brief Global index of left-upper elelment of this submatrix (@a idxRow, @a idxCol).*/
	int idxRow;
	/** @brief Global index of left-upper elelment of this submatrix (@a idxRow, @a idxCol).*/
	int idxCol;
	/** @brief Position of this process (@a posX, @a posY). */
	int posX; 
	/** @brief Position of this process (@a posX, @a posY). */
	int posY;

	/** @brief The MPI type for the elements. */
	MPI_Datatype mpitype;


public:
	////////////////////////////////////////////////////////////////////////
	// constructor and destructor
public:
	/**
	 * @brief Default constructor.
	 *
	 * This constructor do nothing.
	 */
    dist_matrix();
	/**
	 * @brief constructor specifying its shape but elements.
	 *
	 * The number of processors to be used is @a r * @a c.
	 * If r or c is less than 0 (default), 
	 * then r and c are calcuated by distribution policy 
	 * which can be spedified by matrix_skeletons.
	 * 
	 * @a r * @a c should not exceed the total number of processors.
	 * @param nn the number of rows of global matrix.
	 * @param mm the number of columns of global matrix.
	 * @param r  the number of partitions in rows.
	 * @param c  the number of partitions in columns.
	 */
	dist_matrix(int nn, int mm, int r = -1, int c = -1);
	/**
	 * @brief Constructor with generator function.
	 *
	 * The number of processors to be used is @a r * @a c.
	 * If r or c is less than 0 (default), 
	 * then r and c are calcuated by distribution policy 
	 * which can be spedified by matrix_skeletons.
	 * @a r * @a c should not exceed the total number of processors.
	 * @param nn the number of rows of global matrix.
	 * @param mm the number of columns of global matrix.
	 * @param r  the number of partitions in rows.
	 * @param c  the number of partitions in columns.
	 * @param f  the generator function object. 
	 *     the value of element at (i, j) is initialized with @a f (i, j).
	 *     @a f should have a member void f(int i, int j, A *r) const. 
	 * A sample of function object is as follows: 
	 * (the value of element at (i, j) is i + j)
	 * @code
	 * class Gen {
	 * public:
	 *     void operator()(int i, int j, A *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 */
	template<typename F>
	dist_matrix(int nn, int mm, const F &f, int r = -1, int c = -1);
 	/**
	 * @brief Constructor with distributing root's local matrix.
	 *
	 * The number of processors to be used is @a r * @a c.
	 * If r or c is less than 0 (default), 
	 * then r and c are calcuated by distribution policy 
	 * which can be spedified by matrix_skeletons.
	 * @a r * @a c should not exceed the total number of processors.
	 * @param mat matrix on root process to be distributed.
	 * @param r  the number of partitions in rows.
	 * @param c  the number of partitions in columns.
	 */
	dist_matrix(const matrix<A> *mat, int r = -1, int c = -1);
 	/**
	 * @brief A destructor.
	 */
 public:
	~dist_matrix();


	////////////////////////////////////////////////////////////////////////
	// Private Utilities
private:
 	/**
	 * @brief Initialize some fields(process alignment, allocate memory for elements).
	 * @param r  the number of partitions in rows.
	 * @param c  the number of partitions in columns.
	 */
	void init(int r, int c);
 	/**
	 * @brief Tnitialize some fields(submatrix alignment, allocate memory for row/col index).
	 */
	void init2();
 	/**
	 * @brief Return matrixStructure of rank @a rnk.
	 * @param rnk  rank of process of which structure is returned.
	 * @return the structure of rank @a rnk.
	 */
	matrixStructure createStructure(int rnk) const;
	/**
	 * @brief Seding data with serialization.
	 *
	 * This function returns the sending buffer, so you should free it.
	 * @param target the target prosess id(rank).
	 * @param dat data to send.
	 * @param num the number of data.
	 * @param tag1 tag of sending buffer size.
	 * @param tag2 tag of sending serialized data.
	 * @param reqs an instance of SKEL_MPI_Request which Isend returns is stored.
	 * @return The sending buffer. User should free it after the transfer finishes.
	 */
	template <typename B>
	static unsigned char *sendSerializingData(int target, B *dat, int num, dist_matrix_tag_t tag1, dist_matrix_tag_t tag2, SKEL_MPI_Request &reqs);
	/**
	 * @brief Seding all elements of this matrix with serialization.
	 *
	 * This function returns the sending buffer, so you should free it.
	 * @param target the target prosess id(rank).
	 * @param tag1 tag of sending buffer size.
	 * @param tag2 tag of sending serialized data.
	 * @param reqs SKEL_MPI_Request Isend returns is stored.
	 * @return The sending buffer. User should free it after the transfer finishes.
	 */
	unsigned char *sendSerializingData(int target, dist_matrix_tag_t tag1, dist_matrix_tag_t tag2, SKEL_MPI_Request &reqs) const;
	/**
	 * @brief Receiving elements with serialization and storing them to an array.
	 *
	 * @param target the target prosess id(rank).
	 * @param dat an array to which the received data are stored.
	 * @param num the number of data.
	 * @param tag1 tag of receiving buffer size.
	 * @param tag2 tag of receiving serialized data.
	 */
	template <typename B>
	static void recvSerializingData(int target, B *dat, int num, dist_matrix_tag_t tag1, dist_matrix_tag_t tag2);
	/**
	 * @brief Receiving elements with serialization and they are stored in this matrix as its elements.
	 * @param target the target prosess id(rank).
	 * @param tag1 tag of receiving buffer size.
	 * @param tag2 tag of receiving serialized data.
	 */
	void recvSerializingData(int target, dist_matrix_tag_t tag1, dist_matrix_tag_t tag2);

	////////////////////////////////////////////////////////////////////////
	// Public Utilities
 public:
	/**
	 * @brief Generate elements by generator function.
	 *
	 * @param f  the generator function object. 
	 *     the value of element at (i, j) is initialized with @a f (i, j).
	 *     @a f should have a member void f(int i, int j, A *r) const. 
	 * A sample of function object is as follows: 
	 * (the value of element at (i, j) is i + j)
	 * @code
	 * class Gen {
	 * public:
	 *     void operator()(int i, int j, int *r) const {
	 *         *r = i + j;
	 *     }
	 * };
	 * @endcode
	 */
	template <typename F>
	void generate(const F &f);
 	/**
	 * @brief Distribute a matrix at root process.
	 *
	 * Arguments of other process are ignored.
	 * @param mat a matrix to be distributed.
	 */
	void distribute(const matrix<A> *mat);
	/**
	 * @brief Gather block-matrices distributed to processors into a matrix.
	 *
	 * The whole matrix is stored in the root process.
	 * The arguments of other process are unknown.
	 * @param mat a matrix to which the whole matrix is stored.
	 */
	void gather(matrix<A> *mat) const;
	/**
	 * @brief Gather with broadcast.
	 *
	 * This function gather the block-matrices distributed to processors into a matrix.
	 * The resulting matrix is broadcasted among all processors.
	 * @param mat a matrix to which the whole matrix is stored.
	 * @see gather(matrix<A> &)
	 */
	void gatherB(matrix<A> *mat) const;
	/**
	 * @brief Print whole matrix to output-stream in root process.
	 *
	 * This function gathers distributed matrix 
	 *  and outputs it to the output stream of root process.
	 * @param os output-stream to which the matrix is output.
	 */
	void print(std::ostream& os) const;
	/**
	 * @brief Shape copy.
	 *
	 * Shape (size of matrix and the distribution pattern) is copied from the argument matrix @a mat.
 	 * @param mat dist_matrix to be copied (shape only).
	 */
	template <typename B>
	void shapeCopy(const dist_matrix<B> *mat);



	////////////////////////////////////////////////////////////////////////
	// access methods
public:
	/**
	 * @brief The number of rows of total matrix.
	 */
	int getRows() const {return n;}
	/**
	 * @brief The number of columns of total matrix.
	 */
	int getCols() const {return m;}
	/**
	 * @brief The number of rows of local submatrix.
	 */
	int getLocalRows() const {return localRows;}
	/**
	 * @brief The number of columns of local submatrix.
	 */
	int getLocalCols() const {return localCols;}
	/**
	 * @brief The number of submatrices in rows.
	 */
	int getBlocksInRow() const {return bir;}
	/**
	 * @brief the number of submatrices in columns.
	 */
	int getBlocksInCol() const {return bic;}
	/**
	 * @brief Global row-index of upper-left element of this matrix.
	 */
	int getIdxRow() const {return idxRow;}
	/**
	 * @brief Global column-index of upper-left element of this matrix.
	 */
	int getIdxCol() const {return idxCol;}
	/**
	 * @brief Global x position of this submatrix.
	 */
	int getPosX() const {return posX;}
	/**
	 * @brief Global y position of this submatrix.
	 */
	int getPosY() const {return posY;}
	/**
	 * @brief Return reference of the element at (@a i, @a j) of local submatrix.
	 *
	 * @param i row-index of the element.
	 * @param j column-index of the element.
	 */
	A &get(int i, int j) const {return idxEle[i][j];}
	/**
	 * @brief Return reference of the element at (@a i, @a j) of local submatrix.
	 *
	 * @param i row-index of the element.
	 * @param j column-index of the element.
	 */
	A &at(int i, int j) {return idxEle[i][j];}
	/**
	 * @brief Return const reference of the element at (@a i, @a j) of local submatrix.
	 *
	 * @param i row-index of the element.
	 * @param j column-index of the element.
	 */
	const A &at(int i, int j) const {return idxEle[i][j];}
	/**
	 * @brief Put value @a v at (@a i, @a j) of local submatrix.
	 *
	 * @param i row-index of the element.
	 * @param j column-index of the element.
	 * @param v the value to be stored.
	 */
	void set(int i, int j, const A &v) {idxEle[i][j] = v;}
	/**
	 * @brief Rank of this process.
	 */
	int getRank() const { return myrank;}
	/**
	 * @brief Return elements of this dist_matrix as an object of matrix.
	 */
	matrix<A> *getMatrix() const { 	return new matrix<A>(n, m, ele, idxEle, false); }
	/**
	 * @brief MPI datatype of elements.
	 *
	 * Users usually should not use this function.
	 */
	const MPI_Datatype &getType() const { return mpitype; }

	/**
	 * @brief get an element at (@a i, @a j) of global index of the matrix.
	 *
	 * This function returns a value of the element at (@a i, @a j) wherever the element is stored.
	 * The result is broadcasted to all processes.
	 * User should free the resulting pointer.
	 * @param i row-index of the element
	 * @param j coljmn-index of the element
	 */
	A* globalGet(int i, int j) const;
	/**
	 * @brief set an element at (@a i, @a j) of global index of the matrix.
	 *
	 * This function puts a value in root process into an element at (@a i, @a j) wherever the element is stored.
	 * @param i row-index of the element
	 * @param j coljmn-index of the element
	 * @param val the value to be put (the value is ignored in all processes except the root)
	 */
	void globalSet(int i, int j, const A& val) const;
	/**
	 * @brief broadcast (from root to the other processes).
	 *
	 * This function usually used to broadcast results of recude skeletons.
	 * For example, to divide all elements by their summation, we use reduce-skeleton with addition-operators to get the summation 'S', broadcast it by this bcast, and apply map-skeleton with a function '( / S)' to the matrix.
	 * @param res the value to be broadcasted in root process, the received value in other processes.   
	 * @param mat a dist_matrix used to determine members of broadcasting.
	 */
	template<typename B>
	static void bcast(B *res, const dist_matrix<A> *mat);
	
};

template<typename A>
class template_mpitype< matrix < A > > {
	/** 
	 * @brief Returns MPI_Datatype of matrix<A>. 
	 * 
	 * This function returns just MPI_Datatype of A as a dummy type,
	 * because MPI_Datatype of matrix<A> is not used in skeletons.
	 * @return MPI_Datatype of A. 
	 */
 public:
    static const MPI_Datatype type() { return template_mpitype< A >::type( );}
};


#include "dist_matrix.tpp"

#endif // __DIST_MATRIX_H__
