/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file dist_matrix.tpp
 * @brief The implementation of class @a dist_matrix, the data for distributed
 *        matrices and read/write functions. 
 * @author Kento EMOTO
 */

#ifdef __DIST_MATRIX_H__
#ifndef __DIST_MATRIX_TPP__
#define __DIST_MATRIX_TPP__

////////////////////////////////////////////////////////////////////////
// constructor and destructor


template< typename A >
dist_matrix< A >::dist_matrix()
{
	mpitype = template_mpitype< A >::type( );
	myrank = skeleton::rank;
	n = 0;
	m = 0;
	ele = NULL;
	idxEle = NULL;
	bic = 1;
	bir = 1;
	idxRow = 0;
	idxCol = 0;
	localRows = 0;
	localCols = 0;	
}

template< typename A >
dist_matrix< A >::dist_matrix(int nn, int mm, int r, int c)
{
	mpitype = template_mpitype< A >::type( );
	myrank = skeleton::rank;
	n = nn;
	m = mm;
	if(r < 0 || c < 0) matrix_skeletons::calcDivision(r, c);
	init(r, c);
	init2();
}

	
template< typename A > template<typename F>
dist_matrix< A >::dist_matrix(int nn, int mm, const F &f, int r, int c)
{
	mpitype = template_mpitype< A >::type( );
	myrank = skeleton::rank;
	n = nn;
	m = mm;
	if(r < 0 || c < 0) matrix_skeletons::calcDivision(r, c);
	init(r, c);
	generate(f);
}

template< typename A >
dist_matrix< A >::dist_matrix(const matrix<A> *mat, int r, int c)
{
	mpitype = template_mpitype< A >::type( );
	myrank = skeleton::rank; 
	if(r < 0 || c < 0) matrix_skeletons::calcDivision(r, c);
	bir = r;
	bic = c;
	ele = NULL;
	distribute(mat);
}
template< typename A >
dist_matrix< A >::~dist_matrix()
{
	if(ele!=NULL){
		delete [] ele;
	}
	if(idxEle!=NULL) delete [] idxEle;
}


////////////////////////////////////////////////////////////////////////
// Private Utilities






template< typename A >
void
dist_matrix< A >::init(int r, int c)
{
	blocks = r * c;
	bir = r;
	bic = c;
	rows = (n + bir - 1) / bir;
	cols = (m + bic - 1) / bic;
	posY = myrank / bic;
	posX = myrank % bic;
	ele = new A[rows*cols];
	idxEle = NULL;
}

template< typename A >
void
dist_matrix< A >::init2()
{
	matrixStructure ms = createStructure(myrank);
	idxRow = ms.idxRow;
	idxCol = ms.idxCol;
	localRows = ms.localRows;
	localCols = ms.localCols;
	if(idxEle!=NULL) delete [] idxEle;
	idxEle = new A*[rows];	
	for(int i = 0; i < rows; i++){
		idxEle[i] = &ele[i*cols];
	}
}

template< typename A >
matrixStructure
dist_matrix< A >::createStructure(int rnk) const
{
	matrixStructure ms;
	int pY = rnk / bic;
	int pX = rnk % bic;
	ms.idxRow = pY * rows;
	ms.idxCol = pX * cols;
	
	int restn = (n + bir - 1) % bir;
	int restm = (m + bic - 1) % bic;
	
	if(pY <= restn){
			ms.localRows = rows;
	} else {
		ms.idxRow -= (pY - restn - 1);
		ms.localRows = rows - 1;
	}
	if(pX <= restm){
		ms.localCols = cols;
	} else {
		ms.idxCol -= (pX - restm - 1);
		ms.localCols = cols - 1;
	}
	return ms;
}

template< typename A >template< typename B >
unsigned char *
dist_matrix< A >::sendSerializingData(int target, B *dat, int num, dist_matrix_tag_t tag1, dist_matrix_tag_t tag2, SKEL_MPI_Request &reqs) 
{
	int bufSize = 0;
	for(int i = 0; i < num; i++){
		bufSize+=Serializer<B>::getSerializedSize(dat[i]);
	}
	//cout << "rank = " << myrank << ": Send " << bufSize  << " to " << target<< endl;
	SKEL_MPI_Send(&bufSize, 1, MPI_INT, target, tag1, SKEL_MPI_COMM_WORLD);
	unsigned char *buf = new unsigned char[bufSize];
	unsigned char *sbuf = buf;
	for(int i = 0; i < num; i++){
		buf=Serializer<B>::serialize(dat[i], buf);
	}
	//cout << "rank = " << myrank << ": Send data" << " to " << target << endl;
	SKEL_MPI_Isend(sbuf, bufSize, BYTE_TYPE, target, tag2, SKEL_MPI_COMM_WORLD, reqs);
	return sbuf;
}

template< typename A >
unsigned char *
dist_matrix< A >::sendSerializingData(int target, dist_matrix_tag_t tag1, dist_matrix_tag_t tag2, SKEL_MPI_Request &reqs) const
{
	int bufSize = 0;
	for(int i = 0; i < localRows; i++){
		for(int j = 0; j < localCols; j++){
			bufSize += Serializer<A>::getSerializedSize(idxEle[i][j]);
		}
	}
	SKEL_MPI_Send(&bufSize, 1, MPI_INT, target, tag1, SKEL_MPI_COMM_WORLD);
	unsigned char *buf = new unsigned char[bufSize];
	unsigned char *sbuf = buf;
	for(int i = 0; i < localRows; i++){
		for(int j = 0; j < localCols; j++){
			buf = Serializer<A>::serialize(idxEle[i][j], buf);
		}
	}
	SKEL_MPI_Isend(sbuf, bufSize, BYTE_TYPE, target, tag2, SKEL_MPI_COMM_WORLD, reqs);
	return sbuf;
}
template< typename A >template< typename B >
void
dist_matrix< A >::recvSerializingData(int target, B *dat, int num, dist_matrix_tag_t tag1, dist_matrix_tag_t tag2) 
{
	int bufSize;
	//cout << "rank = " << myrank << ": rcv from " << target << endl;
	SKEL_MPI_Recv(&bufSize, 1, MPI_INT, target, tag1, SKEL_MPI_COMM_WORLD);
	//cout << "rank = " << myrank << ": Recv " << bufSize << " from " << target << endl;
	unsigned char *buf = new unsigned char[bufSize];
	unsigned char *sbuf = buf;
	SKEL_MPI_Request reqr;
	SKEL_MPI_Irecv(buf, bufSize, BYTE_TYPE, target, tag2, SKEL_MPI_COMM_WORLD, reqr);
	SKEL_MPI_Wait(reqr);
	//cout << "rank = " << myrank << ": Recv data" << " from " << target << endl;
	for(int i = 0; i < num; i++){
		buf = Serializer<B>::construct(&dat[i], buf);
	}
	delete [] sbuf;
}

template< typename A >
void
dist_matrix< A >::recvSerializingData(int target, dist_matrix_tag_t tag1, dist_matrix_tag_t tag2)
{
	int bufSize;
	SKEL_MPI_Recv(&bufSize, 1, MPI_INT, target, tag1, SKEL_MPI_COMM_WORLD);
	unsigned char *buf = new unsigned char[bufSize];
	unsigned char *sbuf = buf;
	SKEL_MPI_Request reqr;
	SKEL_MPI_Irecv(buf, bufSize, BYTE_TYPE, target, tag2, SKEL_MPI_COMM_WORLD, reqr);
	SKEL_MPI_Wait(reqr);
	for(int i = 0; i < localRows; i++){
		for(int j = 0; j < localCols; j++){
			buf = Serializer<A>::construct(&idxEle[i][j], buf);
		}
	}
	delete [] sbuf;
}

////////////////////////////////////////////////////////////////////////
// Public Utilities

template< typename A >template< typename F >
void
dist_matrix< A >::generate(const F &f)
{
	init2();
	if(myrank >= blocks) return;
	A *p = ele;
	for(int i = 0; i < localRows; i++){
		p = &ele[i*cols];
		for(int j = 0; j < localCols; j++){
			f(i + idxRow, j + idxCol, p++);
		}
	}
}

template< typename A >
void
dist_matrix< A >::distribute(const matrix<A> *mat)
{
	int root = 0;
	n = mat->getRows();
	m = mat->getCols();
	if(ele!=NULL) delete [] ele;
	init(bir, bic);
	init2();
	A *p = ele;
	int size = rows*cols;
	if(myrank >= blocks) return;
	if(myrank==root){
		A *buf = new A[size];
		SKEL_MPI_Request reqs;
		for(int k = 0; k < blocks; k++){
			if(k==root) continue;
			matrixStructure ms = createStructure(k);
			if(Serializer<A>::needsSerializing()){
				int bufSize = 0;
				for(int i = 0; i < ms.localRows; i++){
					for(int j = 0; j < ms.localCols; j++){
						bufSize += Serializer<A>::getSerializedSize(mat->at(i + ms.idxRow,j + ms.idxCol));
					}
				}
				
				SKEL_MPI_Send(&bufSize, 1, MPI_INT, k, TAG_DISTRIBUTE2, SKEL_MPI_COMM_WORLD);
				unsigned char *buf = new unsigned char[bufSize];
				unsigned char *sbuf = buf;
				for(int i = 0; i < ms.localRows; i++){
					for(int j = 0; j < ms.localCols; j++){
						buf = Serializer<A>::serialize(mat->at(i + ms.idxRow,j + ms.idxCol), buf);
					}
				}
				SKEL_MPI_Isend(sbuf, bufSize, BYTE_TYPE, k, TAG_DISTRIBUTE, SKEL_MPI_COMM_WORLD, reqs);
				SKEL_MPI_Wait(reqs);
				delete [] sbuf;
			} else {
				for(int i = 0; i < ms.localRows; i++){
					p = &buf[i*cols];
					for(int j = 0; j < ms.localCols; j++){
						*p++ = mat->at(i + ms.idxRow, j + ms.idxCol);
					}
				}
				SKEL_MPI_Isend(buf, size, mpitype, k, TAG_DISTRIBUTE, SKEL_MPI_COMM_WORLD, reqs);
				SKEL_MPI_Wait(reqs);
			}
		}
		for(int i = 0; i < localRows; i++){
			p = &ele[i*cols];
			for(int j = 0; j < localCols; j++){
				*p++ = mat->at(i + idxRow, j + idxCol);
			}
		}
	} else {
		SKEL_MPI_Request reqr;
		
		if(Serializer<A>::needsSerializing()){
			recvSerializingData(root, TAG_DISTRIBUTE2, TAG_DISTRIBUTE);
		} else {
			SKEL_MPI_Irecv(ele, size, mpitype, root, TAG_DISTRIBUTE, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqr);
		}
	}	
}
template< typename A >
void
dist_matrix< A >::gatherB(matrix<A> *mat) const
{
	gather(mat);
	bcast(mat, this);
}
template< typename A >
void
dist_matrix< A >::gather(matrix<A> *mat) const
{
	mat->setSize(getRows(), getCols());
	if(myrank >= blocks) return;
	if(myrank==0){
		//cout << "gather\n";
		matrixStructure *mss = new matrixStructure[blocks];
		for(int i = 1; i < blocks; i++){
			//cout << "mss[" << i << "] receive\n";
			SKEL_MPI_Recv((int*)&mss[i], 4, MPI_INT, i, TAG_GATHER1, SKEL_MPI_COMM_WORLD);
		}
		int size = rows*cols;
		//cout << "local copy\n";
		for(int i = 0; i < localRows; i++){
			A *p = idxEle[i];
			for(int j = 0; j < localCols; j++){
				mat->at(i+idxRow, idxCol+j) = *p++;
			}
		}
		if(Serializer<A>::needsSerializing()){
			for(int k = 1; k < blocks; k++){
				int bufSize;
				SKEL_MPI_Recv(&bufSize, 1, MPI_INT, k, TAG_GATHER3, SKEL_MPI_COMM_WORLD);
				unsigned char *buf = new unsigned char[bufSize];
				unsigned char *sbuf = buf;
				SKEL_MPI_Request reqr;
				SKEL_MPI_Irecv(buf, bufSize, BYTE_TYPE, k, TAG_GATHER2, SKEL_MPI_COMM_WORLD, reqr);
				SKEL_MPI_Wait(reqr);
				int lr = mss[k].localRows;
				int lc = mss[k].localCols;
				int ir = mss[k].idxRow;
				int ic = mss[k].idxCol;
				for(int i = 0; i < lr; i++){
					for(int j = 0; j < lc; j++){
						//cout << "construct " << i << " " << j << endl;
						
						buf = Serializer<A>::construct(&mat->at(i+ir, ic+j), buf);
						//cout << *(q-1) << endl;
					}
				}
				delete [] sbuf;
			}	
		} else {
			A *buf = new A[size];
			for(int k = 1; k < blocks; k++){
				SKEL_MPI_Request reqr;
				SKEL_MPI_Irecv(buf, size, mpitype, k, TAG_GATHER2, SKEL_MPI_COMM_WORLD, reqr);
				SKEL_MPI_Wait(reqr);
				int lr = mss[k].localRows;
				int lc = mss[k].localCols;
				int ir = mss[k].idxRow;
				int ic = mss[k].idxCol;
				for(int i = 0; i < lr; i++){
					A *p = &buf[i*cols];
					for(int j = 0; j < lc; j++){
						mat->at(i+ir, ic+j) = *p++;
					}
				}
			}	
			delete [] buf;
		}
		delete [] mss;
	} else {
		matrixStructure ms;
		ms.idxRow = idxRow;
		ms.idxCol = idxCol;
		ms.localRows = localRows;
		ms.localCols = localCols;
		SKEL_MPI_Send(&ms, 4, MPI_INT, 0, TAG_GATHER1, SKEL_MPI_COMM_WORLD);
		if(Serializer<A>::needsSerializing()){
			SKEL_MPI_Request reqs;	
			unsigned char *sbuf = sendSerializingData(0, TAG_GATHER3, TAG_GATHER2, reqs);
			SKEL_MPI_Wait(reqs);
			delete [] sbuf;
		} else {
			SKEL_MPI_Request reqs;
			SKEL_MPI_Isend(ele, rows*cols, mpitype, 0, TAG_GATHER2, SKEL_MPI_COMM_WORLD, reqs);
			SKEL_MPI_Wait(reqs);
		}
	}
}

template< typename A >
void
dist_matrix< A >::print(std::ostream& os) const
{
	matrix <A> mat(getRows(), getCols());
	gather(&mat); 
	if(myrank==0){
		os << mat;
	}
}

template< typename A >	template <typename B>
void
dist_matrix< A >::shapeCopy(const dist_matrix<B> *mat)
{
	n = mat->getRows();
	m = mat->getCols();
	if(ele!=NULL) delete [] ele;
	init(mat->getBlocksInRow(), mat->getBlocksInCol());
	init2();
	idxRow = mat->getIdxRow();
	idxCol = mat->getIdxCol();
	localRows = mat->getLocalRows();
	localCols = mat->getLocalCols();	
}


template< typename A >
A*
dist_matrix< A >::globalGet(int i, int j) const
{
	if(myrank >= blocks) return NULL;
	A* res = new A;
	int ii = i - getIdxRow();
	int jj = j - getIdxCol();
	int flag = 0;
	if(0<=ii && ii < getLocalRows() && 0 <= jj && jj < getLocalCols()){
		flag = 1;
	} else {
		flag = 0;
	}
	if(myrank==0){
		int *flags = new int[blocks];
		flags[0] = flag;
		for(int i = 1; i < blocks; i++){
			SKEL_MPI_Recv((int*)&flags[i], 1, MPI_INT, i, TAG_GLOBAL_GET1, SKEL_MPI_COMM_WORLD);
		}
		if(flag==1){
			*res = idxEle[ii][jj];
		} else {
			for(int i = 1; i < blocks;i++){
				if(flags[i]==1){
					if(Serializer<A>::needsSerializing()){
						recvSerializingData(i, res, 1, TAG_GLOBAL_GET2, TAG_GLOBAL_GET22);
					} else {
						SKEL_MPI_Request reqr;
						SKEL_MPI_Irecv(res, 1, mpitype, i, TAG_GLOBAL_GET2, SKEL_MPI_COMM_WORLD, reqr);
						SKEL_MPI_Wait(reqr);
					}
					//break;
				}
			}

		}
		for(int i = 1; i < blocks; i++){
			if(Serializer<A>::needsSerializing()){
				SKEL_MPI_Request reqs;
				unsigned char* sbuf = sendSerializingData(i, res, 1, TAG_GLOBAL_GET3, TAG_GLOBAL_GET32, reqs);
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;				
			} else {
				SKEL_MPI_Request reqs;
				SKEL_MPI_Isend(res, i, mpitype, 0, TAG_GLOBAL_GET3, SKEL_MPI_COMM_WORLD, reqs);
				SKEL_MPI_Wait(reqs);
			}
		}
		delete [] flags;
	} else {
		SKEL_MPI_Send(&flag, 1, MPI_INT, 0, TAG_GLOBAL_GET1, SKEL_MPI_COMM_WORLD);
		if(flag==1){
			*res = idxEle[ii][jj];
			if(Serializer<A>::needsSerializing()){
				SKEL_MPI_Request reqs;
				unsigned char* sbuf = sendSerializingData(0, res, 1, TAG_GLOBAL_GET2, TAG_GLOBAL_GET22, reqs);
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;				
			} else {	
				SKEL_MPI_Request reqs;
				SKEL_MPI_Isend(res, 1, mpitype, 0, TAG_GLOBAL_GET2, SKEL_MPI_COMM_WORLD, reqs);
				SKEL_MPI_Wait(reqs);
			}
		}
		if(Serializer<A>::needsSerializing()){
			recvSerializingData(0, res, 1, TAG_GLOBAL_GET3, TAG_GLOBAL_GET32);
		} else {
			SKEL_MPI_Request reqr;
			SKEL_MPI_Irecv(res, 1, mpitype, 0, TAG_GLOBAL_GET3, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqr);
		}
	}
	return res;
}


template< typename A >
void
dist_matrix< A >::globalSet(int i, int j, const A &val) const
{
	if(myrank >= blocks) return;
	int ii = i - getIdxRow();
	int jj = j - getIdxCol();
	int flag = 0;
	if(0<=ii && ii < getLocalRows() && 0 <= jj && jj < getLocalCols()){
		flag = 1;
	} else {
		flag = 0;
	}
	if(myrank==0){
		int *flags = new int[blocks];
		flags[0] = flag;
		for(int i = 1; i < blocks; i++){
			SKEL_MPI_Recv((int*)&flags[i], 1, MPI_INT, i, TAG_GLOBAL_SET1, SKEL_MPI_COMM_WORLD);
		}
		if(flag==1){
			idxEle[ii][jj] = val;
		} else {
			for(int i = 1; i < blocks;i++){
				if(flags[i]==1){
					if(Serializer<A>::needsSerializing()){
						SKEL_MPI_Request reqs;
						unsigned char* sbuf = sendSerializingData(i, &val, 1, TAG_GLOBAL_SET2, TAG_GLOBAL_SET22, reqs);
						SKEL_MPI_Wait(reqs);	
						delete [] sbuf;				
					} else {	
						SKEL_MPI_Request reqs;
						SKEL_MPI_Isend(&val, 1, mpitype, i, TAG_GLOBAL_SET2, SKEL_MPI_COMM_WORLD, reqs);
						SKEL_MPI_Wait(reqs);
					}
					//break;
				}
			}

		}
		delete [] flags;
	} else {
		SKEL_MPI_Send(&flag, 1, MPI_INT, 0, TAG_GLOBAL_SET1, SKEL_MPI_COMM_WORLD);
		if(flag==1){
			if(Serializer<A>::needsSerializing()){
				recvSerializingData(0, &(idxEle[ii][jj]), 1, TAG_GLOBAL_SET2, TAG_GLOBAL_SET22);
			} else {
				SKEL_MPI_Request reqr;
				SKEL_MPI_Irecv(&(idxEle[ii][jj]), 1, mpitype, 0, TAG_GLOBAL_SET2, SKEL_MPI_COMM_WORLD, reqr);
				SKEL_MPI_Wait(reqr);
			}
		}
	}
}

template<typename A> template < typename B>
void
dist_matrix< A >::bcast(B *res, const dist_matrix<A> *mat)
{
	if(mat->myrank >= mat->blocks) return;
	if(mat->myrank==0){
		for(int i = 1; i < mat->blocks; i++){
			// send res to i-th process
			SKEL_MPI_Request reqs;
			if(Serializer<A>::needsSerializing()){
				unsigned char* sbuf = dist_matrix<A>::sendSerializingData(i, res, 1, TAG_REDUCE3, TAG_REDUCE1, reqs);
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;
			} else {
				SKEL_MPI_Isend(res, 1, mat->mpitype, i, TAG_REDUCE1, SKEL_MPI_COMM_WORLD, reqs);
				SKEL_MPI_Wait(reqs);	
			}
		}
	} else {
		// receive res from root
		if(Serializer<A>::needsSerializing()){
			dist_matrix<A>::recvSerializingData(0, res, 1, TAG_REDUCE3, TAG_REDUCE1);
		} else {
			SKEL_MPI_Request reqr;
			SKEL_MPI_Irecv(res, 1, mat->mpitype, 0, TAG_REDUCE1, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqr);
		}
	}
	
}




#endif // __DIST_MATRIX_TPP__
#endif // __DIST_MATRIX_H__

