/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file matrix_skeletons.tpp
 * @brief The implementation of class @a matrix_skeletons,
 *        consists of skeletons on matrices(two-dimensional arrays)
 *        that manipulating the distributed matrix.
 * @author Kento EMOTO
 */

#ifdef __MATRIX_SKELETONS_H__
#ifndef __MATRIX_SKELETONS_TPP__
#define __MATRIX_SKELETONS_TPP__


////////////////////////////////////////////////////////////////////////
// Didtribution Policies for dist_matrix
void
matrix_skeletons::calcDivision(int &r, int &c)
{
	int np = skeleton::procs;
	switch(distributionPolicy){
		case DPOLICY_SQUARE:{
			int k;
			for(k = 1; k*k <= np; k++);
			k--;
			r = k;
			c = k;}
			break;
		case DPOLICY_ROW:
			r = np;
			c = 1;
			break;
		case DPOLICY_COLUMN:
			r = 1;
			c = np;
			break;
		case DPOLICY_RECTANGLE:{
			int k;
			for(k = 1; k*k <= np; k++);
			k--;
			int l;
			for(l = k; l > 0; l--){
				if((np % l) == 0){
					break;
				}
			}
			r = np / l;
			c = l;}
			break;
		default:
		break;
	}
}

int
matrix_skeletons::setDistributionPolicy(int p)
{
	int o = distributionPolicy;
	distributionPolicy = p;
	return o;
}

int
matrix_skeletons::getDistributionPolicy()
{
	return distributionPolicy;
}

////////////////////////////////////////////////////////////////////////
// Data parallel Skeletons
template <typename A, typename F>
void
matrix_skeletons::map_i(const F &f, dist_matrix<A> *mat)
{
	if(mat->myrank >= mat->blocks) return;
	for(int i = 0; i < mat->localRows; i++){
		for(int j = 0; j < mat->localCols; j++){
			f(&mat->at(i, j));
		}
	}
}

template< typename B, typename A, typename F>
void
matrix_skeletons::map(const F &f, const dist_matrix<A> *mat, dist_matrix<B> *dmb)
{
	if(mat->myrank >= mat->blocks) return;
	for(int i = 0; i < mat->localRows; i++){
		for(int j = 0; j < mat->localCols; j++){
			f(mat->at(i, j), &dmb->at(i, j));
		}
	}
}
	
template<typename A, typename F>
dist_matrix<typename F::result_type>*
matrix_skeletons::map(const F &f, const dist_matrix<A> *mat)
{
	dist_matrix<typename F::result_type>* dmb = new dist_matrix<typename F::result_type>();
	dmb->shapeCopy(mat);
	map(f, mat, dmb);
	return dmb;
}


template< typename C, typename A, typename B, typename F>
void
matrix_skeletons::zipwith(const F &f, const dist_matrix<A>* dma, const dist_matrix<B>* dmb, dist_matrix<C>* dmc )
{
	if(dma->myrank >= dma->blocks) return;
	for(int i = 0; i < dma->localRows; i++){
		for(int j = 0; j < dma->localCols; j++){
			f(dma->at(i, j), dmb->at(i, j), &dmc->at(i, j));
		}
	}
}

template< typename A, typename B, typename F>
dist_matrix<typename F::result_type>*
matrix_skeletons::zipwith(const F &f, const dist_matrix<A>* dma, const dist_matrix<B>* dmb)
{
	dist_matrix<typename F::result_type>* dmc = new dist_matrix<typename F::result_type>();
	dmc->shapeCopy(dma);
	zipwith(f, dma, dmb, dmc);
	return dmc;
}
	
template< typename A, typename B >
dist_matrix < std::pair<A,B> > *
matrix_skeletons::zip(const dist_matrix<A>* dma, const dist_matrix<B>* dmb)
{
	return zipwith(Mkpair<A,B>(), dma, dmb);
}


template <typename A, typename F, typename G>
A*
matrix_skeletons::reduceB(const F &oplus, const G &otimes, const dist_matrix<A>*mat)
{
	A* res = new A;
	reduceB(oplus, otimes, mat, res);
	return res;		
}

template <typename A, typename F, typename G>
void
matrix_skeletons::reduceB(const F &oplus, const G &otimes, const dist_matrix<A>*mat, A *res)
{
	reduce(oplus, otimes, mat, res);
	dist_matrix<A>::bcast(res, mat);
}
template <typename A, typename F, typename G>
void
matrix_skeletons::reduce(const F &oplus, const G &otimes, const dist_matrix<A>* mat, A *res)
{
	if(mat->myrank >= mat->blocks) return;
	int bicfloor = (mat->myrank/mat->bic)*mat->bic;
	A *vals = new A[mat->localRows];
	for(int i = 0; i < mat->localRows; i++){
		A val = mat->at(i, 0);
		for(int j = 1; j < mat->localCols; j++){
			A tmp;
			otimes(val, mat->at(i, j), &tmp);
			val = tmp;
		}
		vals[i] = val;
	}
	A val = vals[0];
	for(int i = 1; i < mat->localRows; i++){
		A tmp;
		oplus(val, vals[i], &tmp);
		val = tmp;
	}
	//if(myrank==0) cout << "reduce 1\n";
	// tree communication (row)
	for(int stage = 1; stage < mat->bic; stage <<= 1){
		int target = ((mat->myrank%mat->bic)^stage) + bicfloor;
		//cout << "rank = " << myrank << " target : " << target << endl;
		if( target < mat->myrank ) {
			SKEL_MPI_Request reqs;
			if(Serializer<A>::needsSerializing()){
				//unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE3, TAG_REDUCE1, reqs);
				unsigned char* sbuf = dist_matrix<A>::sendSerializingData(target, &val, 1, TAG_REDUCE3, TAG_REDUCE1, reqs);
				SKEL_MPI_Wait(reqs);	
				delete [] sbuf;
				//cout << "rank = " << myrank << " sent" << endl;
			} else {
				SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE1, SKEL_MPI_COMM_WORLD, reqs);
				SKEL_MPI_Wait(reqs);	
			}
			break;
		} else if (((mat->myrank%mat->bic)^stage) < mat->bic ) {
			A rval;
			if(Serializer<A>::needsSerializing()){
				//cout << "rank = " << myrank << " ready to receive" << endl;
				//mat->recvSerializingData(target, &rval, 1, TAG_REDUCE3, TAG_REDUCE1);
				dist_matrix<A>::recvSerializingData(target, &rval, 1, TAG_REDUCE3, TAG_REDUCE1);
				//cout << "rank = " << myrank << " received" << endl;
			} else {
				SKEL_MPI_Request reqr;
				SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE1, SKEL_MPI_COMM_WORLD, reqr);
				SKEL_MPI_Wait(reqr);
			}
			A tmp = val;
			otimes(tmp, rval, &val);
		}
	}
	//if(myrank==0) cout << "reduce 2\n";
	if(mat->myrank%mat->bic==0){
		for(int stage = 1; stage < mat->bir; stage <<= 1){
			int target = ((mat->myrank/mat->bic)^stage)*mat->bic;
			//cout << "rank = " << myrank << " target : " << target << endl;
			if( target < mat->myrank ) {	
				SKEL_MPI_Request reqs;
				if(Serializer<A>::needsSerializing()){
					//unsigned char* sbuf = mat->sendSerializingData(target, &val, 1, TAG_REDUCE4, TAG_REDUCE2, reqs);
					unsigned char* sbuf = dist_matrix<A>::sendSerializingData(target, &val, 1, TAG_REDUCE4, TAG_REDUCE2, reqs);
					SKEL_MPI_Wait(reqs);	
					delete [] sbuf;
					//cout << "rank = " << myrank << " sent" << endl;
				} else {
					SKEL_MPI_Isend(&val, 1, mat->mpitype, target, TAG_REDUCE2, SKEL_MPI_COMM_WORLD, reqs);
					SKEL_MPI_Wait(reqs);	
				}
				break;
				//} else if ( target < blocks ) {
			} else if (((mat->myrank/mat->bic)^stage) < mat->bir ) {
				A rval;
				if(Serializer<A>::needsSerializing()){
					//cout << "rank = " << myrank << " ready to receive" << endl;
					//mat->recvSerializingData(target, &rval, 1, TAG_REDUCE4, TAG_REDUCE2);
					dist_matrix<A>::recvSerializingData(target, &rval, 1, TAG_REDUCE4, TAG_REDUCE2);
					//cout << "rank = " << myrank << " received" << endl;
				} else {
					SKEL_MPI_Request reqr;
					SKEL_MPI_Irecv(&rval, 1, mat->mpitype, target, TAG_REDUCE2, SKEL_MPI_COMM_WORLD, reqr);
					SKEL_MPI_Wait(reqr);
				}
				A tmp = val;
				oplus(tmp, rval, &val);
			}
		}
	}
	*res = val; 
}


template <typename A, typename F, typename G>
A*
matrix_skeletons::reduce(const F &oplus, const G &otimes, const dist_matrix<A>* mat)
{
	A* res = new A;
	reduce(oplus, otimes, mat, res);
	return res;		
}


template <typename A, typename F, typename G>
void
matrix_skeletons::scan(const F &oplus, const G &otimes, const dist_matrix<A>* dma, dist_matrix<A>* dmb)
{
	/***
	 * this function consists of two sets of three steps:
	 * 1. row-wise
	 *  1.1. local reduce
	 *  1.2. communication
	 *  1.3. local scan
	 * 2. col-wise
	 *  2.1. local reduce
	 *  2.2. communication
	 *  2.3. local scan
	 */
	if(dma->myrank >= dma->blocks) return;

	int bicfloor = (dma->myrank/dma->bic)*dma->bic;
	A *t_vals;
	A *f_vals;
	A *r_vals;
	
	/* 		for(int i = 0; i < localRows; i++){ */
	/* 			for (int j = 0; j < localCols; j++){ */
	/* 				dmb->at(i, j) = at(i, j); */
	/* 			} */
	/* 		} */
	/* 		goto LLABEL; */
	/***
	 * 1.1. local reduce(row)
	 */
	//if(myrank==0) cout << "scan 1.1\n";
	t_vals = new A[dma->localRows];
	for(int i = 0; i < dma->localRows; i++){
		A t_val = dma->at(i, 0);
		for(int j = 1; j < dma->localCols; j++){
			A tmp;
			otimes(t_val, dma->at(i, j), &tmp);
			t_val = tmp;
		}
		t_vals[i] = t_val;
	}
	/***
	 * 1.2. communication (row) (h-cube)
	 */
	//if(myrank==0) cout << "scan 1.2\n";
	f_vals = new A[dma->localRows];
	for(int i = 0; i < dma->localRows; i++){
		f_vals[i] = dma->at(i, 0);
	}
	r_vals = new A[dma->localRows];
	for(int stage = 1; stage < dma->bic; stage <<= 1){
		int target = ((dma->myrank%dma->bic)^stage) + bicfloor;
		if (((dma->myrank%dma->bic)^stage) >= dma->bic) continue;
		//if (target >= blocks ) continue;
		SKEL_MPI_Request reqs, reqr;
		
		if(Serializer<A>::needsSerializing()){
			//unsigned char* sbuf = dma->sendSerializingData(target, t_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12, reqs);
			//dma->recvSerializingData(target, r_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12);
			unsigned char* sbuf = dist_matrix<A>::sendSerializingData(target, t_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12, reqs);
			dist_matrix<A>::recvSerializingData(target, r_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, dma->localRows, dma->mpitype, target, TAG_SCAN12, SKEL_MPI_COMM_WORLD, reqs);
			SKEL_MPI_Irecv(r_vals, dma->localRows, dma->mpitype, target, TAG_SCAN12, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}
		if( target > dma->myrank ) {
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(t_vals[i], r_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		} else {
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(r_vals[i], f_vals[i], &tmp);
				f_vals[i] = tmp;
			}
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(r_vals[i], t_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		}
	}
	/***
	 * 1.3. local scan(row)
	 */
	//if(myrank==0) cout << "scan 1.3\n";
	for(int i = 0; i < dma->localRows; i++){
		dmb->at(i, 0) = f_vals[i];
		for (int j = 1; j < dma->localCols; j++){
			otimes(dmb->at(i, j-1), dma->at(i, j), &dmb->at(i, j));
		}
	}
	delete [] r_vals;
	delete [] f_vals;
	delete [] t_vals;
	/* 		return; */
	/* 	LLABEL: */
	  
	/***
	 * 2.1. local reduce(col)
	 */
	//if(myrank==0) cout << "scan 2.1\n";
	t_vals = new A[dma->localCols];
	for(int j = 0; j < dma->localCols; j++){
		A t_val = dmb->at(0, j);
		for(int i = 1; i < dma->localRows; i++){
			A tmp;
			oplus(t_val, dmb->at(i, j), &tmp);
			t_val = tmp;
		}
		t_vals[j] = t_val;
	}
	/***
	 * 2.2. communication (row) (h-cube)
	 */
	//	if(myrank==0) cout << "scan 2.2\n";
	f_vals = new A[dma->localCols];
	for(int j = 0; j < dma->localCols; j++){
		f_vals[j] = dmb->at(0, j);
	}
	r_vals = new A[dma->localCols];
	for(int stage = 1; stage < dma->bir; stage <<= 1){
		int target = ((dma->myrank/dma->bic)^stage)*dma->bic + dma->myrank%dma->bic;
		if (((dma->myrank/dma->bic)^stage) >= dma->bir) continue;
		//	if (target >= blocks) continue;
		SKEL_MPI_Request reqs, reqr;
		if(Serializer<A>::needsSerializing()){
			//unsigned char* sbuf = dma->sendSerializingData(target, t_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22, reqs);
			//dma->recvSerializingData(target, r_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22);
			unsigned char* sbuf = dist_matrix<A>::sendSerializingData(target, t_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22, reqs);
			dist_matrix<A>::recvSerializingData(target, r_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, dma->localCols, dma->mpitype, target, TAG_SCAN22, SKEL_MPI_COMM_WORLD, reqs);
			SKEL_MPI_Irecv(r_vals, dma->localCols, dma->mpitype, target, TAG_SCAN22, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}

		if( target > dma->myrank ) {
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(t_vals[j], r_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		} else {
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(r_vals[j], f_vals[j], &tmp);
				f_vals[j] = tmp;
			}
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(r_vals[j], t_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		}
	}
	/***
	 * 2.3. local scan(row)
	 */
	//if(myrank==0) cout << "scan 2.3\n";
	for (int j = 0; j < dma->localCols; j++){
		dmb->at(0, j) = f_vals[j];
		for(int i = 1; i < dma->localRows; i++){
			A tmp;
			oplus(dmb->at(i-1, j), dmb->at(i, j), &tmp);
			dmb->at(i, j) = tmp;
		}
	}
		
		
}

template <typename A, typename F, typename G>
dist_matrix<A>*
matrix_skeletons::scan(const F &oplus, const G &otimes, const dist_matrix<A>* dma)
{
	dist_matrix<A>* dmb = new dist_matrix<A>();
	dmb->shapeCopy(dma);
	scan(oplus, otimes, dma, dmb);
	return dmb;
}


template <typename A, typename F, typename G>
void
matrix_skeletons::scanr(const F &oplus, const G &otimes, const dist_matrix<A>* dma, dist_matrix<A>* dmb)
{
	/***
	 * this function consists of two sets of three steps:
	 * 1. row-wise
	 *  1.1. local reduce
	 *  1.2. communication
	 *  1.3. local scan
	 * 2. col-wise
	 *  2.1. local reduce
	 *  2.2. communication
	 *  2.3. local scan
	 */
	if(dma->myrank >= dma->blocks) return;

	int bicfloor = (dma->myrank/dma->bic)*dma->bic;
	A *t_vals;
	A *f_vals;
	A *r_vals;
		
	/***
	 * 1.1. local reduce(row)
	 */
	//if(myrank==0) cout << "scan 1.1\n";
	t_vals = new A[dma->localRows];
	for(int i = 0; i < dma->localRows; i++){
		A t_val = dma->at(i, dma->localCols-1);
		for(int j = dma->localCols-2; j >=0; j--){
			A tmp;
			otimes(dma->at(i, j), t_val, &tmp);
			t_val = tmp;
		}
		t_vals[i] = t_val;
	}
	/***
	 * 1.2. communication (row) (h-cube)
	 */
	//if(myrank==0) cout << "scan 1.2\n";
	f_vals = new A[dma->localRows];
	for(int i = 0; i < dma->localRows; i++){
		f_vals[i] = dma->at(i, dma->localCols-1);
	}
	r_vals = new A[dma->localRows];
	int vmyrank = dma->bic - (dma->myrank%dma->bic) - 1;
	for(int stage = 1; stage < dma->bic; stage <<= 1){
		int vtarget = (vmyrank^stage);
		int target = dma->bic - vtarget -1;
		if(target < 0) continue;
		target += bicfloor;
		SKEL_MPI_Request reqs, reqr;
			
		if(Serializer<A>::needsSerializing()){
			//unsigned char* sbuf = dma->sendSerializingData(target, t_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12, reqs);
			//dma->recvSerializingData(target, r_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12);
			unsigned char* sbuf = dist_matrix<A>::sendSerializingData(target, t_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12, reqs);
			dist_matrix<A>::recvSerializingData(target, r_vals, dma->localRows, TAG_SCAN12s, TAG_SCAN12);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, dma->localRows, dma->mpitype, target, TAG_SCAN12, SKEL_MPI_COMM_WORLD, reqs);
			SKEL_MPI_Irecv(r_vals, dma->localRows, dma->mpitype, target, TAG_SCAN12, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}
		if( vtarget > vmyrank ) {
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(r_vals[i], t_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		} else {
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(f_vals[i], r_vals[i], &tmp);
				f_vals[i] = tmp;
			}
			for(int i = 0; i < dma->localRows; i++){
				A tmp;
				otimes(t_vals[i], r_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		}
	}
	/***
	 * 1.3. local scan(row)
	 */
	//if(myrank==0) cout << "scan 1.3\n";
	for(int i = 0; i < dma->localRows; i++){
		dmb->at(i, dma->localCols-1) = f_vals[i];
		for (int j = dma->localCols-2; j >= 0; j--){
			otimes(dma->at(i, j), dmb->at(i, j+1), &dmb->at(i, j));
		}
	}
	delete [] r_vals;
	delete [] f_vals;
	delete [] t_vals;
	/* 		return; */
	/* 	LLABEL: */
	  
	/***
	 * 2.1. local reduce(col)
	 */
	//if(myrank==0) cout << "scan 2.1\n";
	t_vals = new A[dma->localCols];
	for(int j = 0; j < dma->localCols; j++){
		A t_val = dmb->at(dma->localRows-1, j);
		for(int i = dma->localRows-2; i >= 0; i--){
			A tmp;
			oplus(dmb->at(i, j), t_val, &tmp);
			t_val = tmp;
		}
		t_vals[j] = t_val;
	}
	/***
	 * 2.2. communication (row) (h-cube)
	 */
	//	if(myrank==0) cout << "scan 2.2\n";
	f_vals = new A[dma->localCols];
	for(int j = 0; j < dma->localCols; j++){
		f_vals[j] = dmb->at(dma->localRows-1, j);
	}
	vmyrank = dma->bir - dma->myrank/dma->bic - 1;
	r_vals = new A[dma->localCols];
	for(int stage = 1; stage < dma->bir; stage <<= 1){
		int vtarget = (vmyrank^stage);
		int target = dma->bir - vtarget -1;
		if(target < 0) continue;
		target = target * dma->bic + dma->myrank%dma->bic;
			
		//	if (target >= blocks) continue;
		SKEL_MPI_Request reqs, reqr;
		if(Serializer<A>::needsSerializing()){
			//unsigned char* sbuf = dma->sendSerializingData(target, t_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22, reqs);
			//dma->recvSerializingData(target, r_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22);
			unsigned char* sbuf = dist_matrix<A>::sendSerializingData(target, t_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22, reqs);
			dist_matrix<A>::recvSerializingData(target, r_vals, dma->localCols, TAG_SCAN22s, TAG_SCAN22);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, dma->localCols, dma->mpitype, target, TAG_SCAN22, SKEL_MPI_COMM_WORLD, reqs);
			SKEL_MPI_Irecv(r_vals, dma->localCols, dma->mpitype, target, TAG_SCAN22, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}

		if( vtarget > vmyrank ) {
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(r_vals[j], t_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		} else {
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(f_vals[j], r_vals[j], &tmp);
				f_vals[j] = tmp;
			}
			for(int j = 0; j < dma->localCols; j++){
				A tmp;
				oplus(t_vals[j], r_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		}
	}
	/***
	 * 2.3. local scan(row)
	 */
	//if(myrank==0) cout << "scan 2.3\n";
	for (int j = 0; j < dma->localCols; j++){
		dmb->at(dma->localRows-1, j) = f_vals[j];
		for(int i = dma->localRows-2; i>=0; i--){
			A tmp;
			oplus(dmb->at(i, j), dmb->at(i+1, j), &tmp);
			dmb->at(i, j) = tmp;
		}
	}
		
		
	
}
template <typename A, typename F, typename G>
dist_matrix<A>*
matrix_skeletons::scanr(const F &oplus, const G &otimes, const dist_matrix<A>* dma)
{
	dist_matrix<A>* dmb = new dist_matrix<A>();
	dmb->shapeCopy(dma);
	scanr(oplus, otimes, dma, dmb);
	return dmb;
}




////////////////////////////////////////////////////////////////////////
// Data parallel Skeletons (shifting)

template<typename A>
void
matrix_skeletons::rotateRowsConst(int s, dist_matrix<A> *mat)
{
	if(mat->myrank >= mat->blocks) return;
	int receiver = rotRow(s, mat->posY, mat) * mat->bic + mat->posX;
	int sender = rotRow(-s, mat->posY, mat) * mat->bic + mat->posX;
	permute(sender, receiver, mat);
}

template<typename A, typename F>
void
matrix_skeletons::rotateRows(const F &f, dist_matrix<A> *mat)
{
	rotateRowsConst(f(mat->posX), mat);
}

template<typename A>
void
matrix_skeletons::rotateColsConst(int s, dist_matrix<A> *mat)
{
	if(mat->myrank >= mat->blocks) return;
	int receiver = mat->posY * mat->bic + rotCol(s, mat->posX, mat);
	int sender = mat->posY * mat->bic + rotCol(-s, mat->posX, mat);
	permute(sender, receiver, mat);
}

template<typename A, typename F>
void
matrix_skeletons::rotateCols(const F &f, dist_matrix<A> *mat)
{
	rotateColsConst(f(mat->posY), mat);
}



////////////////////////////////////////////////////////////////////////
// Private Utilities
template<typename A>
int
matrix_skeletons::rotRow(int s, int y, dist_matrix<A> *mat)
{
	int newY = y+s;
	while(newY >= mat->bir) newY-=mat->bir;
	while(newY < 0) newY+=mat->bir;
	return newY;
}

template<typename A>
int
matrix_skeletons::rotCol(int s, int x, dist_matrix<A> *mat)
{
	int newX = x+s;
	while(newX >= mat->bic) newX-=mat->bic;
	while(newX < 0) newX+=mat->bic;
	return newX;
}

template<typename A>
void
matrix_skeletons::permute(int sender, int receiver, dist_matrix<A> *mat)
{
	if(receiver==mat->myrank && mat->myrank==sender) return;
	matrixStructure ms, msr;
	ms.idxRow = mat->idxRow;
	ms.idxCol = mat->idxCol;
	ms.localRows = mat->localRows;
	ms.localCols = mat->localCols;
	SKEL_MPI_Send(&ms, 4, MPI_INT, receiver, TAG_PERMUTE1, SKEL_MPI_COMM_WORLD);
	SKEL_MPI_Recv(&msr, 4, MPI_INT, sender, TAG_PERMUTE1, SKEL_MPI_COMM_WORLD);
	SKEL_MPI_Request reqs, reqr;
	if(Serializer<A>::needsSerializing()){
		SKEL_MPI_Request reqs;	
		unsigned char *sbuf = mat->sendSerializingData(receiver, TAG_PERMUTE3, TAG_PERMUTE2, reqs);
		mat->idxRow = msr.idxRow;
		mat->idxCol = msr.idxCol;
		mat->localRows = msr.localRows;
		mat->localCols = msr.localCols;
		mat->recvSerializingData(sender, TAG_PERMUTE3, TAG_PERMUTE2);
		SKEL_MPI_Wait(reqs);
		delete [] sbuf;
	} else {
		int size = mat->rows*mat->cols;
		A *buf = new A[size];
		SKEL_MPI_Isend(mat->ele, size, mat->mpitype, receiver, TAG_PERMUTE2, SKEL_MPI_COMM_WORLD, reqs);
		mat->idxRow = msr.idxRow;
		mat->idxCol = msr.idxCol;
		mat->localRows = msr.localRows;
		mat->localCols = msr.localCols;
		SKEL_MPI_Irecv(buf, size, mat->mpitype, sender, TAG_PERMUTE2, SKEL_MPI_COMM_WORLD, reqr);
		SKEL_MPI_Wait(reqs);
		SKEL_MPI_Wait(reqr);
		A *p = buf;
		A *q = mat->ele;
		for(int i = 0; i < size; i++){
			*q++ = *p++;
		}
		delete [] buf;
	}
}

template<typename A>
dist_matrix< matrix< A > >*
matrix_skeletons::all_rows(const dist_matrix<A> *mat){
	dist_matrix< matrix<A> >* dmb = new dist_matrix< matrix<A> >();
	dmb->shapeCopy(mat);
	if(mat->myrank >= mat->blocks) return dmb;
	dist_matrix<A>* dmc = new dist_matrix<A>();
	dmc->shapeCopy(mat);
	for(int i = 0; i < mat->localRows; i++){
		dmb->at(i, 0).setSize(1, mat->getCols());
		for(int j = 0; j < mat->localCols; j++){
			dmc->at(i, j) = mat->at(i, j);
		}
	}
	for(int l = 0; l < mat->bic; l++){
		for(int i = 0; i < dmb->getLocalRows(); i++){
			int idxc = dmc->getIdxCol();
			for(int k = 0; k < dmc->getLocalCols(); k++){
				dmb->at(i,0).at(0, idxc++) = dmc->at(i, k);
			}
		}
		matrix_skeletons::rotateColsConst(1, dmc);
	}
	for(int i = 0; i < mat->localRows; i++){
		for (int j = 1; j < mat->localCols; j++){
			Shadower<matrix<A > >::shadow(dmb->at(i, 0),&dmb->at(i, j));
		}
	}
	delete dmc;
	return dmb;
}
template<typename A>
dist_matrix< matrix< A > >*
matrix_skeletons::all_cols(const dist_matrix<A> *mat){
	dist_matrix< matrix<A> >* dmb = new dist_matrix< matrix<A> >();
	dmb->shapeCopy(mat);
	if(mat->myrank >= mat->blocks) return dmb;
	dist_matrix<A>* dmc = new dist_matrix<A>();
	dmc->shapeCopy(mat);
	for(int i = 0; i < mat->localRows; i++){
		for(int j = 0; j < mat->localCols; j++){
			dmc->at(i, j) = mat->at(i, j);
		}
	}
	for(int j = 0; j < mat->localCols; j++){
		dmb->at(0, j).setSize(mat->getRows(), 1);
	}
	for(int l = 0; l < mat->bir; l++){
		for(int j = 0; j < dmb->getLocalCols(); j++){
			int idxc = dmc->getIdxRow();
			for(int k = 0; k < dmc->getLocalRows(); k++){
				dmb->at(0, j).at(idxc++, 0) = dmc->at(k, j);
			}
		}
		matrix_skeletons::rotateRowsConst(1, dmc);
	}
	for(int i = 1; i < mat->localRows; i++){
		for (int j = 0; j < mat->localCols; j++){
			Shadower<matrix<A> >::shadow(dmb->at(0, j),&dmb->at(i, j));
		}
	}
	delete dmc;
	return dmb;
}

template <typename A, typename F>
dist_matrix< A >*
matrix_skeletons::allreduce_row(const F &otimes, const dist_matrix<A> *mat){
	dist_matrix< A >* dmb = new dist_matrix< A >();
	dmb->shapeCopy(mat);
	if(mat->myrank >= mat->blocks) return dmb;

	const int localCols = mat->localCols;
	const int localRows = mat->localRows;
	//	const int bir = mat->bir;
	const int bic = mat->bic;
	const int myrank = mat->myrank;

	int bicfloor = (myrank/bic)*bic;
	A *t_vals;
	A *r_vals;
	
	t_vals = new A[localRows];
	for(int i = 0; i < localRows; i++){
		A t_val = mat->at(i, 0);
		for(int j = 1; j < localCols; j++){
			A tmp;
			otimes(t_val, mat->at(i, j), &tmp);
			t_val = tmp;
		}
		t_vals[i] = t_val;
	}
	r_vals = new A[localRows];
	/**
	 * butterfly (up)
	 */
	int stage;
	int cnt = 0;
	for(stage = 1; stage < bic; stage <<= 1){
		if(bic&stage) cnt++;
		int target = ((myrank%bic)^stage) + bicfloor;
		if (((myrank%bic)^stage) >= bic) continue;
		SKEL_MPI_Request reqs, reqr;
		
		if(Serializer<A>::needsSerializing()){
			unsigned char* sbuf = dist_matrix<A>::sendSerializingData(target, t_vals, localRows, TAG_ALLREDUCEROW2, TAG_ALLREDUCEROW1, reqs);
			dist_matrix<A>::recvSerializingData(target, r_vals, localRows, TAG_ALLREDUCEROW2, TAG_ALLREDUCEROW1);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, localRows, mat->mpitype, target, TAG_ALLREDUCEROW1, SKEL_MPI_COMM_WORLD, reqs);
			SKEL_MPI_Irecv(r_vals, localRows, mat->mpitype, target, TAG_ALLREDUCEROW1, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}
		if( target > myrank ) {
			for(int i = 0; i < localRows; i++){
				A tmp;
				otimes(t_vals[i], r_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		} else {
			for(int i = 0; i < localRows; i++){
				A tmp;
				otimes(r_vals[i], t_vals[i], &tmp);
				t_vals[i] = tmp;
			}
		}
	}
	delete [] r_vals;
	if(cnt > 1){
		/**
		 * tree (down), omitted when bic = 2^k
		 */
		for(; cnt > 0; cnt--) stage>>=1;
		int base = 1;
		for(;base&bic==0;base<<=1);
		unsigned char* sbuf=NULL;
		int bufSize = 0;
		for(; stage >= base; stage >>= 1){
			if((myrank%bic)&((stage-1)&(~(base-1)))!=0) continue;
			int target = ((myrank%bic)^stage) + bicfloor;
			if (((myrank%bic)^stage) >= bic) continue;
			SKEL_MPI_Request reqs, reqr;
			if( target > myrank ) {
				if(Serializer<A>::needsSerializing()){
					if(sbuf==NULL){
						bufSize = 0;
						for(int i = 0; i < localRows; i++){
							bufSize+=Serializer<A>::getSerializedSize(t_vals[i]);
						}
					}
					SKEL_MPI_Send(&bufSize, 1, MPI_INT, target, TAG_ALLREDUCEROW4, SKEL_MPI_COMM_WORLD);
					if(sbuf==NULL){
						unsigned char *buf = new unsigned char[bufSize];
						sbuf = buf;
						for(int i = 0; i < localRows; i++){
							buf=Serializer<A>::serialize(t_vals[i], buf);
						}
					}
					SKEL_MPI_Isend(sbuf, bufSize, BYTE_TYPE, target, TAG_ALLREDUCEROW3, SKEL_MPI_COMM_WORLD, reqs);
					SKEL_MPI_Wait(reqs);	
				} else {
					SKEL_MPI_Isend(t_vals, localRows, mat->mpitype, target, TAG_ALLREDUCEROW3, SKEL_MPI_COMM_WORLD, reqs);
					SKEL_MPI_Wait(reqs);
				}
			} else {
				if(Serializer<A>::needsSerializing()){
					SKEL_MPI_Recv(&bufSize, 1, MPI_INT, target, TAG_ALLREDUCEROW4, SKEL_MPI_COMM_WORLD);
					unsigned char *buf = new unsigned char[bufSize];
					sbuf = buf;
					SKEL_MPI_Irecv(buf, bufSize, BYTE_TYPE, target, TAG_ALLREDUCEROW3, SKEL_MPI_COMM_WORLD, reqr);
					SKEL_MPI_Wait(reqr);
					for(int i = 0; i < localRows; i++){
						buf = Serializer<A>::construct(&t_vals[i], buf);
					}
				} else {
					SKEL_MPI_Irecv(t_vals, localRows, mat->mpitype, target, TAG_ALLREDUCEROW3, SKEL_MPI_COMM_WORLD, reqr);
					SKEL_MPI_Wait(reqr);
				}		
			}
		}
		if(sbuf!=NULL) delete [] sbuf;
	}
	/**
	 * shadowing
	 */
	for(int i = 0; i < localRows; i++){
		dmb->at(i, 0) = t_vals[i];
		for (int j = 1; j < localCols; j++){
			Shadower<A>::shadow(dmb->at(i, 0),&dmb->at(i, j));
		}
	}
	delete [] t_vals;
	return dmb;
}

template <typename A, typename F>
dist_matrix< A >*
matrix_skeletons::allreduce_col(const F &oplus, const dist_matrix<A> *mat){
	dist_matrix< A >* dmb = new dist_matrix< A >();
	dmb->shapeCopy(mat);
	if(mat->myrank >= mat->blocks) return dmb;
	
	A* t_vals;
	A* r_vals;

	const int localCols = mat->localCols;
	const int localRows = mat->localRows;
	const int bir = mat->bir;
	const int bic = mat->bic;
	const int myrank = mat->myrank;
	
	t_vals = new A[localCols];
	for(int j = 0; j < localCols; j++){
		A t_val = mat->at(0, j);
		for(int i = 1; i < localRows; i++){
			A tmp;
			oplus(t_val, mat->at(i, j), &tmp);
			t_val = tmp;
		}
		t_vals[j] = t_val;
	}
	r_vals = new A[localCols];
	/**
	 * butterfly (up)
	 */
	int stage;
	int cnt = 0;
	for(stage = 1; stage < bir; stage <<= 1){
		if(bir&stage) cnt++;
		int target = ((myrank/bic)^stage)*bic + myrank%bic;
		if (((myrank/bic)^stage) >= bir) continue;
		SKEL_MPI_Request reqs, reqr;
		if(Serializer<A>::needsSerializing()){
			unsigned char* sbuf = dist_matrix<A>::sendSerializingData(target, t_vals, localCols, TAG_ALLREDUCECOL2, TAG_ALLREDUCECOL1, reqs);
			dist_matrix<A>::recvSerializingData(target, r_vals, localCols, TAG_ALLREDUCECOL2, TAG_ALLREDUCECOL1);
			SKEL_MPI_Wait(reqs);	
			delete [] sbuf;
		} else {
			SKEL_MPI_Isend(t_vals, localCols, mat->mpitype, target, TAG_ALLREDUCECOL1, SKEL_MPI_COMM_WORLD, reqs);
			SKEL_MPI_Irecv(r_vals, localCols, mat->mpitype, target, TAG_ALLREDUCECOL1, SKEL_MPI_COMM_WORLD, reqr);
			SKEL_MPI_Wait(reqs);
			SKEL_MPI_Wait(reqr);
		}
		if( target > myrank ) {
			for(int j = 0; j < localCols; j++){
				A tmp;
				oplus(t_vals[j], r_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		} else {
			for(int j = 0; j < localCols; j++){
				A tmp;
				oplus(r_vals[j], t_vals[j], &tmp);
				t_vals[j] = tmp;
			}
		}
	}
	delete [] r_vals;
	if(cnt > 1){
		/**
		 * tree (down), omitted when bir = 2^k
		 */
		for(; cnt > 0; cnt--) stage>>=1;
		int base = 1;
		for(;base&bir==0;base<<=1);
		unsigned char* sbuf=NULL;
		int bufSize = 0;
		for(; stage >= base; stage >>= 1){
			if((myrank/bic)&((stage-1)&(~(base-1)))!=0) continue;
			int target = ((myrank/bic)^stage)*bic + myrank%bic;
			if (((myrank/bic)^stage) >= bir) continue;
			SKEL_MPI_Request reqs, reqr;
			if( target > myrank ) {
				if(Serializer<A>::needsSerializing()){
					if(sbuf==NULL){
						bufSize = 0;
						for(int i = 0; i < localCols; i++){
							bufSize+=Serializer<A>::getSerializedSize(t_vals[i]);
						}
					}
					SKEL_MPI_Send(&bufSize, 1, MPI_INT, target, TAG_ALLREDUCECOL4, SKEL_MPI_COMM_WORLD);
					if(sbuf==NULL){						
						unsigned char *buf = new unsigned char[bufSize];
						sbuf = buf;
						for(int i = 0; i < localCols; i++){
							buf=Serializer<A>::serialize(t_vals[i], buf);
						}
					}
					SKEL_MPI_Isend(sbuf, bufSize, BYTE_TYPE, target, TAG_ALLREDUCECOL3, SKEL_MPI_COMM_WORLD, reqs);
					
					SKEL_MPI_Wait(reqs);	
				} else {
					SKEL_MPI_Isend(t_vals, localCols, mat->mpitype, target, TAG_ALLREDUCECOL3, SKEL_MPI_COMM_WORLD, reqs);
					SKEL_MPI_Wait(reqs);
				}
			} else {
				if(Serializer<A>::needsSerializing()){
					//std::cout << "rank = " << myrank << ", Rcv target = " << target << etd::endl;
					SKEL_MPI_Recv(&bufSize, 1, MPI_INT, target, TAG_ALLREDUCECOL4, SKEL_MPI_COMM_WORLD);
					//std::cout << "rank = " << myrank << ", Rcv bufSize = " << bufSize << etd::endl;
					unsigned char *buf = new unsigned char[bufSize];
					sbuf = buf;
					SKEL_MPI_Irecv(buf, bufSize, BYTE_TYPE, target, TAG_ALLREDUCECOL3, SKEL_MPI_COMM_WORLD, reqr);
					SKEL_MPI_Wait(reqr);
					//std::cout << "rank = " << myrank << ", Rcv data" << etd::endl;
					for(int i = 0; i < localCols; i++){
						buf = Serializer<A>::construct(&t_vals[i], buf);
					}
				} else {
						SKEL_MPI_Irecv(t_vals, localCols, mat->mpitype, target, TAG_ALLREDUCECOL3, SKEL_MPI_COMM_WORLD, reqr);
						SKEL_MPI_Wait(reqr);
				}
			}
		}
		if(sbuf!=NULL) delete [] sbuf;
	}
	/**
	 * shadowing
	 */
	for (int j = 0; j < localCols; j++){
		dmb->at(0, j) = t_vals[j];
		for(int i = 1; i < localRows; i++){
			Shadower<A>::shadow(dmb->at(0, j),&dmb->at(i, j));
			}
	}
	delete [] t_vals;
	return dmb;
}


#endif // __MATRIX_SKELETONS_TPP__
#endif // __MATRIX_SKELETONS_H__
