/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file primitive_functions.h
 * @brief Collection of some useful function objects.
 * @author Kento EMOTO
 */
#ifndef __PRIMITIVE_FUNCTIONS_H__
#define __PRIMITIVE_FUNCTIONS_H__

#include "matrix.h"
#include <ctime>
#include <cstdlib>

/**
 * @brief Collection of some useful function objects.
 */
namespace primitive_functions {

/**
 * @brief Singleton constructor.
 */
template<typename A>
class Singleton : public unary_function< A, matrix <A> >{
public:
	void operator()(const A &a, matrix <A> *mat) const {
		mat->singleton(a);
	}
};


/**
 * @brief Inverse of singleton constructor.
 */
template <typename A>
class The {
public:
	void operator()(const matrix <A> &mat, A *r) const{
		*r = mat.at(0,0);
	}
};

/**
 * @brief Nested Singleton constructor.
 */
template<typename A>
class Singleton2 {
public:
	void operator()(const A &a, matrix <matrix <A> > *mat) const {
		matrix <A> singl;
		singl.singleton(a);
		mat->singleton(singl);
	}
};

/**
 * @brief Above constructor.
 */
template <typename A>
class Abv : public binary_function<matrix <A>, matrix <A>, matrix <A> >{
public:
	void operator()(const matrix <A> &a, const matrix <A> &b, matrix <A> *mat) const {
		m_skeletons::abv(&a, &b, mat);
	}
};

/**
 * @brief Beside constructor.
 */
template <typename A>
class Bsd  : public binary_function<matrix <A>, matrix <A>, matrix <A> >{
public:
	void operator()(const matrix <A> &a, const matrix <A> &b, matrix <A> *mat) const {
		m_skeletons::bsd(&a, &b, mat);
	}
};


/**
 * @brief zipwith by beside constructor.
 */
template <typename A>
class Zipwbsd {
public:
	void operator()(const matrix <matrix <A> >&a, const matrix <matrix <A> > &b, matrix <matrix <A> > *mat) const {
		m_skeletons::zipwith(Bsd<A>(), &a, &b, mat);
	}
};

/**
 * @brief zipwith by above constructor.
 */
template <typename A>
class Zipwabv {
public:
	void operator()(const matrix <matrix <A> > &a, const matrix <matrix <A> > &b, matrix <matrix <A> > *mat) const {
		m_skeletons::zipwith(Abv<A>(), &a, &b, mat);
	}
};

	
/**
 * @brief Transpose.
 */
template <typename A>
class Tr : public unary_function<A , A >{
public:
	void operator()(const A &a, A *r) const {
		*r = a;
		r->tr();
	}
};

/**
 * @brief Left Operator.
 */
template <typename A>
class Left : public binary_function< A, A, A > {
public:
	void operator()(const A &a, const A &, A *r) const {
		*r = a;
	}
};
/**
 * @brief Right Operator.
 */
template <typename A>
class Right : public binary_function< A, A, A > {
public:
	void operator()(const A &, const A &b, A*r) const{
		*r = b;
	}
};

/**
 * @brief Square (inplace).
 */
template<typename A>
class SquareI {
public:
    void operator()(A *r) const { 
    	*r = *r * *r;
    }
};

/**
 * @brief Square.
 */
template<typename A>
class Square : public unary_function<A , A >{
public:
    void operator()(const A&a, A *r) const { 
    	*r = a * a;
    }
};

/**
 * @brief Addition.
 */
template<typename A>
class Add : public binary_function< A, A, A > {
public:
	void operator()(const A &i, const A &j, A *r) const {
		*r = i + j;
	}
};

/**
 * @brief Subtraction.
 */
template<typename A>
class Sub : public binary_function< A, A, A > {
public:
	void operator()(const A &i, const A &j, A *r) const {
		*r = i - j;
	}
};

/**
 * @brief Multiplication.
 */
template<typename A>
class Mul : public binary_function< A, A, A > {
public:
	void operator()(const A &i, const A &j, A *r) const {
		*r = i * j;
	}
};

/**
 * @brief Division.
 */
template<typename A>
class Div : public binary_function< A, A, A > {
public:
	void operator()(const A &i, const A &j, A *r) const {
		*r = i / j;
	}
};

/**
 * @brief Modulo.
 */
template<typename A>
class Mod : public binary_function< A, A, A > {
public:
	void operator()(const A &i, const A &j, A *r) const {
		*r = i % j;
	}
};

/**
 * @brief Max.
 */
template<typename A>
class Max : public binary_function< A, A, A > {
public:
	void operator()(const A &i, const A &j, A *r) const {
		*r = (i >= j) ? i : j;
	}
};

/**
 * @brief Min.
 */
template<typename A>
class Min : public binary_function< A, A, A > {
public:
	void operator()(const A &i, const A &j, A *r) const {
		*r = (i <= j) ? i : j;
	}
};


/**
 * @brief Generator function : Zero
 */
template<typename A>
class GenZero {
public:
	void operator()(int , int , A *r) const {
		*r = 0;
	}
};

/**
 * @brief Generator function : Addition
 */
template<typename A>
class GenAdd {
public:
	void operator()(int i, int j, A *r) const {
		*r = i + j;
	}
};

/**
 * @brief Generator function : Subtraction
 */
template<typename A>
class GenSub {
public:
	void operator()(int i, int j, A *r) const {
		*r = i - j;
	}
};

/**
 * @brief Generator function : Multiplication
 */
template<typename A>
class GenMul {
public:
	void operator()(int i, int j, A *r) const {
		*r = i * j;
	}
};

/**
 * @brief Generator function : Random double in [0, 1)
 */
class GenRand1 {
public:
	GenRand1(){
		srand((unsigned int)time(NULL));
	}
	void operator()(int i, int j, double *r) const {
		i = j = 0;
		*r = rand()/(double)RAND_MAX;
	}
};
/**
 * @brief Generator function : Random Integer in [0, n)
 */
class GenRandN {
	int n;
public:
	GenRandN(int nn){
		n = nn;
		srand((unsigned int)time(NULL));
	}
	void operator()(int i, int j, int *r) const {
		i = j = 0;
		*r = (int)((rand()/(double)RAND_MAX)*n);
	}
};


/**
 * @brief Divide by a number
 */
template <typename A>
class DivBy : public unary_function < A, A >{
	A d;
public:
	DivBy(A dd){
		d = dd;
	}
	void operator()(A i, A *r) const {
		*r = i / d;
	}
};
	
}
#endif // __PRIMITIVE_FUNCTIONS_H__
