/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file alltest.cpp
 * @brief Tests of the skeletons.
 * @author Kento EMOTO
 */


#include "matrix_skeletons.h"
#include "primitive_functions.h"

using namespace primitive_functions;

/**
 * @brief Total number of processors.
 */
int size;
/**
 * @brief This rank.
 */
int rank;

void print_test(char *str, bool t){
	if(rank==0){
		if(t){
			std::cout << str << ":  OK\n";
		} else {
			std::cout << str << ":  failed\n";
		}
	}
}

int matSizeR;
int matSizeC;



/**
 * @brief Print string for policy @a p.
 *
 * @param p policy.
 */
void printPolicy(int p)
{
	switch(p){
	case matrix_skeletons::DPOLICY_RECTANGLE:
		std::cout << "DPOLICY_RECTANGLE";
		break;
	case matrix_skeletons::DPOLICY_SQUARE:
		std::cout << "DPOLICY_SQUARE";
		break;
	case matrix_skeletons::DPOLICY_ROW:
		std::cout << "DPOLICY_ROW";
		break;
	case matrix_skeletons::DPOLICY_COLUMN:
		std::cout << "DPOLICY_COLUMN";
		break;
	default:
		std::cout << "OTHER";
		break;
	}
}

/**
 * @brief Set policy and display it.
 */
void setPolicy(int p)
{
	if(rank==0){
		std::cout << "set policy to ";
		printPolicy(p);
		std::cout << std::endl;
	}
	matrix_skeletons::setDistributionPolicy(p);
}

/**
 * @brief Check policy.
 */
void policySetTest(int p)
{
	
	setPolicy(p);
	dist_matrix<int> mat(matSizeR, matSizeC, GenAdd<int>());
	int r = mat.getBlocksInRow();
	int c = mat.getBlocksInCol();
	if(rank==0){
		std::cout << "Division is (" << r << ", " << c << ")" << std::endl;
	}
	switch(p){
	case matrix_skeletons::DPOLICY_RECTANGLE:
		print_test("Policy test(RECTANGLE)", r*c == size);
		break;
	case matrix_skeletons::DPOLICY_SQUARE:
		print_test("Policy test(SQUARE)", r==c && r*c <= size);
		break;
	case matrix_skeletons::DPOLICY_ROW:
		print_test("Policy test(ROW)", r==size && c==1);
		break;
	case matrix_skeletons::DPOLICY_COLUMN:
		print_test("Policy test(COLUMN)", c==size && r==1);
		break;
	default:
		std::cout << "OTHER";
		break;
	}
}

/**
 * @brief Map check.
 */
void mapTest()
{
	matrix<int> mat(matSizeR, matSizeC);
	mat.generate(GenAdd<int>());
	dist_matrix<int> dmat(&mat);
	
	matrix<int> *mat2 = m_skeletons::map(Square<int>(), &mat);
	dist_matrix<int> *dmat2 = matrix_skeletons::map(Square<int>(), &dmat);
	
	m_skeletons::map_i(SquareI<int>(), &mat);
	matrix_skeletons::map_i(SquareI<int>(), &dmat);
	
	matrix<int> res, resi;
	
	dmat2->gather(&res);
	dmat.gather(&resi);
	
	print_test("Map test", res==*mat2 && resi==mat);
	
	delete mat2;
	delete dmat2;
}

/**
 * @brief Scan check.
 */
void scanTest()
{
	matrix<int> mat(matSizeR, matSizeC);
	mat.generate(GenAdd<int>());
	dist_matrix<int> dmat(&mat);
	
	matrix<int> *mat2 = m_skeletons::scan(Max<int>(), Max<int>(), &mat);
	dist_matrix<int> *dmat2 = matrix_skeletons::scan(Max<int>(), Max<int>(), &dmat);
	matrix<int> *mat3 = m_skeletons::scanr(Max<int>(), Max<int>(), &mat);
	dist_matrix<int> *dmat3 = matrix_skeletons::scanr(Max<int>(), Max<int>(), &dmat);
	
	matrix<int> res;
	matrix<int> resr;
	
	dmat2->gather(&res);
	dmat3->gather(&resr);
	
	print_test("Scan test", res==*mat2);
	print_test("Scanr test", resr==*mat3);

	
	delete mat2;
	delete dmat2;
	delete mat3;
	delete dmat3;
}

/**
 * @brief Allreduce check.
 */
void allreduceTest()
{
	matrix<int> mat(matSizeR, matSizeC);
	mat.generate(GenAdd<int>());
	dist_matrix<int> dmat(&mat);
	
	matrix<int> *mat2 = m_skeletons::allreduce_col(Max<int>(), &mat);
	dist_matrix<int> *dmat2 = matrix_skeletons::allreduce_col(Max<int>(), &dmat);
	matrix<int> *mat3 = m_skeletons::allreduce_row(Max<int>(), &mat);
	dist_matrix<int > *dmat3 = matrix_skeletons::allreduce_row(Max<int>(), &dmat);
	
	
	matrix<matrix<int> > *mats = m_skeletons::map(Singleton<int>(), &mat);
	
	matrix<matrix<int> > *mat4 = m_skeletons::allreduce_col(Abv<int>(), mats);
	dist_matrix<matrix<int> > *dmat4 = matrix_skeletons::all_cols(&dmat);
	matrix<matrix<int> > *mat5 = m_skeletons::allreduce_row(Bsd<int>(), mats);
	dist_matrix<matrix<int> > *dmat5 = matrix_skeletons::all_rows(&dmat);

	matrix<int> res, resr;
	matrix<matrix<int> > res2, resr2;
	
	dmat2->gather(&res);
	dmat3->gather(&resr);
	dmat4->gather(&res2);
	dmat5->gather(&resr2);
	
	print_test("Allreduce test", res==*mat2 && resr==*mat3);
	print_test("Allcols/rows test", res2==*mat4 && resr2==*mat5);
	
	delete mats;
	delete mat2;
	delete dmat2;
	delete mat3;
	delete dmat3;
	delete mat4;
	delete dmat4;
	delete mat5;
	delete dmat5;
}


/**
 * @brief Reduce check.
 */
void reduceTest()
{
	matrix<int> mat(matSizeR, matSizeC);
	mat.generate(GenAdd<int>());
	dist_matrix<int> dmat(&mat);
	
	int *r = m_skeletons::reduce(Max<int>(), Max<int>(), &mat);
	int *dr = matrix_skeletons::reduce(Max<int>(), Max<int>(), &dmat);
	
	print_test("Reduce test", *r==*dr);
	
	delete r;
	delete dr;
}

/**
 * @brief Zip check.
 */
void zipTest()
{
	matrix<int> mat(matSizeR, matSizeC);
	mat.generate(GenAdd<int>());
	matrix<int> mat2(matSizeR, matSizeC);
	mat2.generate(GenSub<int>());
	dist_matrix<int> dmat(&mat);
	dist_matrix<int> dmat2(&mat2);
	
	matrix<int> *mat3 = m_skeletons::zipwith(Mul<int>(), &mat, &mat2);
	dist_matrix<int> *dmat3 = matrix_skeletons::zipwith(Mul<int>(), &dmat, &dmat2);
	
	matrix<std::pair<int, int> > *mat4 = m_skeletons::zip(&mat, &mat2);
	dist_matrix<std::pair<int, int> > *dmat4 = matrix_skeletons::zip(&dmat, &dmat2);
	
	matrix<int> res;
	matrix<std::pair<int, int> >resz;
	
	dmat3->gather(&res);
	dmat4->gather(&resz);
	
	print_test("Zip test", res==*mat3 && resz==*mat4);
	
	delete mat3;
	delete dmat3;
	delete mat4;
	delete dmat4;
}







class Suba {
public:
	void operator()(const matrix <int> &a, const matrix <int> &b, matrix <int> *mat) const {
		m_skeletons::zipwith(Sub<int>(), &a, &b, mat);
	}
};



class Add2 {
public:
	void operator()(const std::pair<int, int> &a, const std::pair<int, int> &b, std::pair<int, int>  *r) const {
		r->first  = a.first  + b.first;
		r->second = a.second + b.second;
	}
};

class GenA {
public:
	void operator()(int i, int j, matrix <int> *r) const {
		r->singleton(i + j);
	}
};

template <typename A>
A **create2DArray(int n, int m){
	A **dat = new A*[n];
	for(int i = 0; i < n; i++){
		dat[i] = new A[m];
	}
	return dat;
}
template <typename A>
void free2DArray(A** dat, int n, int m){
	m = 0;
	for(int i = 0; i < n; i++){
		delete [] dat[i];
	}
	delete [] dat;
}

void complexTest()
{
	if(rank==0) std::cout << "Complex Tests" << std::endl;
	int n = matSizeR;
	int m = matSizeC;
	dist_matrix< matrix <int> > A(n, m, GenA());
	matrix <int> **dat = create2DArray< matrix <int> >(n, m);

	matrix < matrix < int> > mat(n, m);
	mat.generate(GenA());
	dist_matrix< matrix <int> > B(&mat);
	
	matrix < matrix < int> > mat2(n, m);
	B.gather(&mat2);

	if(rank==0){
		print_test("distribute & gather", (mat==mat2));
	}
	
	matrix< int > *rmat = matrix_skeletons::reduce(Abv<int>(), Bsd<int>(), &B);
		
	if(rank==0){
		matrix < int > *rmat2 = m_skeletons::reduce(Abv<int>(), Bsd<int>(), &mat2);
		print_test("reduce & abv & bsd", (*rmat==*rmat2));
		delete rmat2;
	}

	dist_matrix < matrix< int > > *smat = matrix_skeletons::scan(Abv<int>(), Bsd<int>(), &B);
	
	matrix < int >  **dats = create2DArray< matrix <int> >(n, m);
	matrix < matrix < int > > ssmat(n, m);
	smat->gather(&ssmat);
	if(rank==0){
		matrix < matrix < int > > *smat2 = m_skeletons::scan(Abv<int>(), Bsd<int>(), &mat2);
		print_test("scan & abv & bsd", (ssmat==*smat2));
		delete smat2;
	}
	ssmat.setSize(0,0);


	matrix < int > *srmat = matrix_skeletons::reduce(Right<matrix <int> >(), Right<matrix <int > >(), smat);
	
	if(rank==0){
		print_test("reduce & scan", (*srmat==*rmat));
	}
	ssmat.setSize(0,0);
	delete smat;
	delete srmat;
	delete rmat;

	dist_matrix < matrix< int > > *smatr = matrix_skeletons::scanr(Abv<int>(), Bsd<int>(), &B);
	
	matrix < matrix < int > > ssmatr(n, m);
	smatr->gather(&ssmatr);
	if(rank==0){
		matrix < matrix < int > > *smatr2 = m_skeletons::scanr(Abv<int>(), Bsd<int>(), &mat2);

		print_test("scanr & abv & bsd", (ssmatr==*smatr2));
		delete smatr2;
	}
	free2DArray(dats, n, m);
	free2DArray(dat, n, m);

	int **data = create2DArray< int >(n, m);
	matrix < int > mata(n, m);
	mata.generate(GenAdd<int>());
	dist_matrix< int > Ba(&mata);
	free2DArray(data, n, m);
	
	std::pair<int, int> **data2 = create2DArray< std::pair<int, int> >(n, m);
	dist_matrix< std::pair< int, int > > *Ba2 = matrix_skeletons::zip(&Ba, &Ba);
	matrix < std::pair < int, int> > mata2(n, m);
	Ba2->gather(&mata2);
	if(rank==0){
		matrix < std::pair < int, int> >  *mataz = m_skeletons::zip(&mata, &mata);
		print_test("std::pair & zip", (*mataz==mata2));
		if(rank==0) delete mataz;
	}

	
	dist_matrix< int > *Ba3 = matrix_skeletons::allreduce_row(Add<int>(), &Ba);
	matrix < int > mata3(n, m);
	Ba3->gather(&mata3);
	matrix < int >  *mataz2 = m_skeletons::scan(Right<int>(), Add<int>(), &mata);
	matrix < int >  *mataz3 = m_skeletons::scanr(Left<int>(), Right<int>(), mataz2);
	
	if(rank==0){
		print_test("allreduce_row", (*mataz3==mata3));
	}

	delete mataz2;
	delete mataz3;
	delete Ba3;

 	dist_matrix< int > *Ba4 = matrix_skeletons::allreduce_col(Add<int>(), &Ba);
 	matrix < int > mata4(n, m);
	Ba4->gather(&mata4);
	matrix < int >  *mataz4 = m_skeletons::scan(Add<int>(), Right<int>(), &mata);
	matrix < int >  *mataz5 = m_skeletons::scanr(Right<int>(), Left<int>(), mataz4);
	
	if(rank==0){
		print_test("allreduce_col", (*mataz5==mata4));
	}

	delete mataz4;
	delete mataz5;
	delete Ba4;
	
	dist_matrix< std::pair<int,int> > *Ba5 = matrix_skeletons::zip(&Ba, &Ba);
	dist_matrix< std::pair<int,int> > *Ba6 = matrix_skeletons::allreduce_col(Add2(), Ba5);
	matrix <  std::pair<int,int> > mata6(n, m);
	Ba6->gather(&mata6);
	matrix < std::pair<int,int> >  *mataz6 = m_skeletons::zip(&mata, &mata);
	matrix < std::pair<int,int> >  *mataz7 = m_skeletons::scan(Add2(), Right<std::pair<int, int> >(), mataz6);
	matrix < std::pair<int,int> >  *mataz8 = m_skeletons::scanr(Right<std::pair<int, int> >(), Left<std::pair<int, int> >(), mataz7);
	
	if(rank==0){
		print_test("allreduce_col & zip", (*mataz8==mata6));
	}

	delete mataz6;
	delete mataz7;
	delete mataz8;
	delete Ba5;
	delete Ba6;

	bool fl = true;
	for(int i = 0; i < n; i++){
		for(int j = 0; j < m; j++){
			std::pair< int, int > *rr3 = Ba2->globalGet(i, j);
			if(rank==0){
				bool kk = (*rr3==mata2.get(i, j));
				fl = fl && kk;
			}
			if(rank==0 && rr3!=NULL) delete rr3;
		}
	}
	if(rank==0){
		print_test("globalGet", fl);
	}

	std::pair<int, int> pr(0, 0);
	matrix < std::pair<int, int> > mataxnn(n, m);
	matrix < std::pair<int, int> > mataxnn2(n, m);
	for(int i = 0; i < n; i++){
		for(int j = 0; j < m; j++){
			pr.first = j*i + 3;
			pr.second = j + i - 3;
			Ba2->globalSet(i, j, pr);
			mataxnn2.at(i, j) = pr;
		}
	}
	Ba2->gather(&mataxnn);
	if(rank==0){
		print_test("globalSet", mataxnn2==mataxnn);
	}

	matrix < int > matax(n, m);
	matrix < int > matax2(n, m);
	for(int i = 0; i < n; i++){
		for(int j = 0; j < m; j++){
			matax.at(i, j) = i * j;
		}
	}
	dist_matrix< int > Bax(&matax);
	int dd = 0;
	matrix_skeletons::reduceB(Add<int>(), Add<int>(), &Bax, &dd);
	dist_matrix < int >  *Bax2 = matrix_skeletons::map(DivBy<int>(dd), &Bax);
	Bax2->gather(&matax2);
	if(rank==0){
		int d = 0;
		m_skeletons::reduce(Add<int>(), Add<int>(), &matax, &d);
		matrix < int >  *mataz = m_skeletons::map(DivBy<int>(d), &matax);
		print_test("reduceB (bcast)", (matax2==*mataz));
		delete mataz;
	}
	delete Bax2;
	delete Ba2;
	free2DArray(data2, n, m);
}







int SketoMain( int argc, char **argv ) {
	rank = skeleton::rank;
	size = skeleton::procs;

	matSizeR = size*2;
	matSizeC = size*3;
	if(argc > 2){
		matSizeR = atoi(argv[1]);
		matSizeC = atoi(argv[2]);
	}
	
	if(rank==0){
		std::cout << "#proc = " << size <<std::endl;
		int p = matrix_skeletons::getDistributionPolicy();
		std::cout << "default policy = ";
		printPolicy(p);
		std::cout << std::endl;
	}
	
	if(rank==0) std::cout << std::endl;
	policySetTest(matrix_skeletons::DPOLICY_RECTANGLE);

	if(rank==0) std::cout << std::endl;
	policySetTest(matrix_skeletons::DPOLICY_SQUARE);

	if(rank==0) std::cout << std::endl;
	policySetTest(matrix_skeletons::DPOLICY_ROW);

	if(rank==0) std::cout << std::endl;
	policySetTest(matrix_skeletons::DPOLICY_COLUMN);

	if(rank==0) std::cout << std::endl;
	setPolicy(matrix_skeletons::DPOLICY_RECTANGLE);

	if(rank==0) std::cout << std::endl;
	mapTest();

 	if(rank==0) std::cout << std::endl;
 	zipTest();
	
 	if(rank==0) std::cout << std::endl;
 	reduceTest();
	
 	if(rank==0) std::cout << std::endl;
 	scanTest();
	
 	if(rank==0) std::cout << std::endl;
 	allreduceTest();
	

 	complexTest();

	return 0;
}

