/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file serialize_test.cpp
 * @brief Tests of serialization of complex structures.
 * @author Kento EMOTO
 */

 

#include <iostream>
#include <iomanip>
#include <mpi.h>
#include <cmath>
#include <cstdlib>

#include "matrix_skeletons.h"

#include "primitive_functions.h"

using namespace primitive_functions;

class Suba {
public:
	void operator()(const matrix <int> &a, const matrix <int> &b, matrix <int> *mat) const {
// 	std::cout << "suba\n";
// 	std::cout << "a = \n" << a << "\n";
// 	std::cout << "b = \n" << b << "\n";
		m_skeletons::zipwith(Sub<int>(), &a, &b, mat);
	}
};

class Add2 {
public:
	void operator()(const std::pair<int, int> &a, const std::pair<int, int> &b, std::pair<int, int>  *r) const {
		r->first = a.first + b.first;
		r->second = a.second + b.second;
	}
};

void print_test(char *str, bool t){
	if(t){
		std::cout << str << ":  OK\n";
	} else {
		std::cout << str << ":  failed\n";
	}
}
class GenA {
public:
	void operator()(int i, int j, matrix <int> *r) const {
		r->singleton(i + j);
	}
};

template <typename A>
A **create2DArray(int n, int m){
	A **dat = new A*[n];
	for(int i = 0; i < n; i++){
		dat[i] = new A[m];
	}
	return dat;
}
template <typename A>
void free2DArray(A** dat, int n, int){
	for(int i = 0; i < n; i++){
		delete [] dat[i];
	}
	delete [] dat;
}


int SketoMain( int argc, char **argv )
{
	
	int n = 5;
	int m = 4;
	int debug = 0;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	if(argc > 2){
		m = atoi(argv[2]);
	}

	
	if(argc > 1){
		n = atoi(argv[1]);
	}
	if(argc > 2){
		m = atoi(argv[2]);
	}
	if(argc > 3){
		debug = atoi(argv[3]);
	}
	

	int myrank = skeleton::rank;
	int nproc = skeleton::procs;
	
	int q = (int)(sqrt((double)nproc)+0.1);
	int r = q;

	if(argc > 4){
		q = atoi(argv[4]);
	}
	if(argc > 5){
		r = atoi(argv[5]);
	}
	if(nproc < r * q){
		std::cout << "too few procs.\n";
		return 0;
	}
	
	dist_matrix< matrix <int> > A(n, m, GenA(), q, r);
	if(myrank==0){
		std::cout << "A : (" << n << " x " << m << ")\n";
		std::cout << " total prosess = " << nproc << "\n";
		std::cout << " using prosess = " << q << "*" << r << "\n";
	}
	if(myrank==0)	std::cout << "A = \n";
	if(debug){
		A.print(std::cout);
	}
	matrix <int> **dat = create2DArray< matrix <int> >(n, m);

	matrix < matrix < int> > mat(n, m);
	mat.generate(GenA());
	dist_matrix< matrix <int> > B(&mat, q, r);
	if(myrank==0)	std::cout << "B = \n";
	if(debug){
		B.print(std::cout);
	}
	
	matrix < matrix < int> > mat2(n, m);
	B.gather(&mat2);

	if(myrank==0){
		print_test("distribute & gather", (mat==mat2));
	}
	
	matrix< int > *rmat = matrix_skeletons::reduce(Abv<int>(), Bsd<int>(), &B);
	if(myrank==0)	std::cout << "rmat = \n";
	if(myrank==0 && debug) std::cout << *rmat << std::endl;
		
	if(myrank==0){
		matrix < int > *rmat2 = m_skeletons::reduce(Abv<int>(), Bsd<int>(), &mat2);
		std::cout << "rmat2 = \n";
		if(debug) std::cout << *rmat2 << std::endl;
		
		print_test("reduce & abv & bsd", (*rmat==*rmat2));
		delete rmat2;
	}

	dist_matrix < matrix< int > > *smat = matrix_skeletons::scan(Abv<int>(), Bsd<int>(), &B);
	if(myrank==0)	std::cout << "smat = \n";
	if(debug) smat->print(std::cout);
	
	matrix < int >  **dats = create2DArray< matrix <int> >(n, m);
	matrix < matrix < int > > ssmat(n, m);
	smat->gather(&ssmat);
	if(myrank==0){
		matrix < matrix < int > > *smat2 = m_skeletons::scan(Abv<int>(), Bsd<int>(), &mat2);
		std::cout << "smat2 = \n";
		if(debug) std::cout << *smat2 << std::endl;

		print_test("scan & abv & bsd", (ssmat==*smat2));
		delete smat2;
	}
	ssmat.setSize(0,0);


	matrix < int > *srmat = matrix_skeletons::reduce(Right<matrix <int> >(), Right<matrix <int > >(), smat);
	if(myrank==0)	std::cout << "srmat = \n";
	if(debug && myrank==0) srmat->print(std::cout);
	
	if(myrank==0){
		print_test("reduce & scan", (*srmat==*rmat));
	}
	ssmat.setSize(0,0);
	delete smat;
	delete srmat;
	delete rmat;

	dist_matrix < matrix< int > > *smatr = matrix_skeletons::scanr(Abv<int>(), Bsd<int>(), &B);
	if(myrank==0)	std::cout << "smatr = \n";
	if(debug) smatr->print(std::cout);
	
	matrix < matrix < int > > ssmatr(n, m);
	smatr->gather(&ssmatr);
	if(myrank==0){
		matrix < matrix < int > > *smatr2 = m_skeletons::scanr(Abv<int>(), Bsd<int>(), &mat2);
		std::cout << "smat2r = \n";
		if(debug) std::cout << *smatr2 << std::endl;

		print_test("scanr & abv & bsd", (ssmatr==*smatr2));
		delete smatr2;
	}
	free2DArray(dats, n, m);
	free2DArray(dat, n, m);

	int **data = create2DArray< int >(n, m);
	matrix < int > mata(n, m);
	mata.generate(GenAdd<int>());
	dist_matrix< int > Ba(&mata, q, r);
	if(myrank==0)	std::cout << "Ba = \n";
	if(debug){
		Ba.print(std::cout);
	}
	free2DArray(data, n, m);
	
	std::pair<int, int> **data2 = create2DArray< std::pair<int, int> >(n, m);
	dist_matrix< std::pair< int, int > > *Ba2 = matrix_skeletons::zip(&Ba, &Ba);
	matrix < std::pair < int, int> > mata2(n, m);
	Ba2->gather(&mata2);
	if(myrank==0){
		matrix < std::pair < int, int> >  *mataz = m_skeletons::zip(&mata, &mata);
		std::cout << "mataz = \n";
		//if(debug) std::cout << *mataz << std::endl;
		print_test("std::pair & zip", (*mataz==mata2));
		if(myrank==0) delete mataz;
	}

	
	dist_matrix< int > *Ba3 = matrix_skeletons::allreduce_row(Add<int>(), &Ba);
	matrix < int > mata3(n, m);
	Ba3->gather(&mata3);
	matrix < int >  *mataz2 = m_skeletons::scan(Right<int>(), Add<int>(), &mata);
	matrix < int >  *mataz3 = m_skeletons::scanr(Left<int>(), Right<int>(), mataz2);
	
	if(myrank==0)	std::cout << "mataz3 = \n";
	if(myrank==0 && debug) std::cout << *mataz3 << std::endl;
	if(myrank==0)	std::cout << "mata3 = \n";
	if(myrank==0 && debug) std::cout << mata3 << std::endl;
	if(myrank==0){
		print_test("allreduce_row", (*mataz3==mata3));
	}

	delete mataz2;
	delete mataz3;
	delete Ba3;

 	dist_matrix< int > *Ba4 = matrix_skeletons::allreduce_col(Add<int>(), &Ba);
 	matrix < int > mata4(n, m);
	Ba4->gather(&mata4);
	matrix < int >  *mataz4 = m_skeletons::scan(Add<int>(), Right<int>(), &mata);
	matrix < int >  *mataz5 = m_skeletons::scanr(Right<int>(), Left<int>(), mataz4);
	
	if(myrank==0)	std::cout << "mataz5 = \n";
	if(myrank==0 && debug) std::cout << *mataz5 << std::endl;
	if(myrank==0)	std::cout << "mata4 = \n";
	if(myrank==0 && debug) std::cout << mata4 << std::endl;
	if(myrank==0){
		print_test("allreduce_col", (*mataz5==mata4));
	}

	delete mataz4;
	delete mataz5;
	delete Ba4;
	
	dist_matrix< std::pair<int,int> > *Ba5 = matrix_skeletons::zip(&Ba, &Ba);
	if(myrank==0)	std::cout << "Ba5 = \n";
	//if(debug)	Ba5->print(std::cout);
	dist_matrix< std::pair<int,int> > *Ba6 = matrix_skeletons::allreduce_col(Add2(), Ba5);
	if(myrank==0)	std::cout << "Ba6 = \n";
	//if(debug)	Ba6->print(std::cout);
	matrix <  std::pair<int,int> > mata6(n, m);
	Ba6->gather(&mata6);
	matrix < std::pair<int,int> >  *mataz6 = m_skeletons::zip(&mata, &mata);
	matrix < std::pair<int,int> >  *mataz7 = m_skeletons::scan(Add2(), Right<std::pair<int, int> >(), mataz6);
	matrix < std::pair<int,int> >  *mataz8 = m_skeletons::scanr(Right<std::pair<int, int> >(), Left<std::pair<int, int> >(), mataz7);
	
	if(myrank==0)	std::cout << "mataz8 = \n";
	//if(myrank==0 && debug) std::cout << *mataz8 << std::endl;
	if(myrank==0)	std::cout << "mata6 = \n";
	//if(myrank==0 && debug) std::cout << mata6 << std::endl;
	if(myrank==0){
		print_test("allreduce_col & zip", (*mataz8==mata6));
	}

	delete mataz6;
	delete mataz7;
	delete mataz8;
	delete Ba5;
	delete Ba6;

	bool fl = true;
	for(int i = 0; i < n; i++){
		for(int j = 0; j < m; j++){
			std::pair< int, int > *rr3 = Ba2->globalGet(i, j);
			if(myrank==0){
				bool kk = (*rr3==mata2.get(i, j));
				fl = fl && kk;
			}
			if(myrank==0 && rr3!=NULL) delete rr3;
		}
	}
	if(myrank==0){
		print_test("globalGet", fl);
	}

	std::pair<int, int> pr(0, 0);
	matrix < std::pair<int, int> > mataxnn(n, m);
	matrix < std::pair<int, int> > mataxnn2(n, m);
	for(int i = 0; i < n; i++){
		for(int j = 0; j < m; j++){
			pr.first = j*i + 3;
			pr.second = j + i - 3;
			Ba2->globalSet(i, j, pr);
			mataxnn2.at(i, j) = pr;
		}
	}
	Ba2->gather(&mataxnn);
	if(myrank==0){
		std::cout << "mataxnn = " << std::endl;
		//if(debug) std::cout << mataxnn << std::endl;
		print_test("globalSet", mataxnn2==mataxnn);
	}

	matrix < int > matax(n, m);
	matrix < int > matax2(n, m);
	for(int i = 0; i < n; i++){
		for(int j = 0; j < m; j++){
			matax.at(i, j) = i * j;
		}
	}
	dist_matrix< int > Bax(&matax, q, r);
	int dd = 0;
	matrix_skeletons::reduceB(Add<int>(), Add<int>(), &Bax, &dd);
	dist_matrix < int >  *Bax2 = matrix_skeletons::map(DivBy<int>(dd), &Bax);
	Bax2->gather(&matax2);
	if(myrank==0){
		int d = 0;
		m_skeletons::reduce(Add<int>(), Add<int>(), &matax, &d);
		matrix < int >  *mataz = m_skeletons::map(DivBy<int>(d), &matax);
		std::cout << "mataz = \n";
		if(debug) std::cout << *mataz << std::endl;
		print_test("reduceB (bcast)", (matax2==*mataz));
		delete mataz;
	}
	delete Bax2;
	delete Ba2;
	free2DArray(data2, n, m);
	
	goto FINAL;
{
}	FINAL:
	
	return 0;
}
