/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file dist_tree.h
 * @brief The Header of class @a dist_tree, the data for distributed
 *        trees and read/write functions. 
 * @author Kiminori MATSUZAKI
 */

#ifndef __DIST_TREE_H__
#define __DIST_TREE_H__

#include "node.h"
#include <fstream>

class tree_skeletons;

/**
 * @brief A template class for handling tree distributed on processors.
 */
template< typename A >
class dist_tree
{
  friend class tree_skeletons;

  ////////////////////////////////////////////////////////////////////////
  // Member variables
private:
  /** The root of local block associated to the process */
  node< A >* local_root;
  /** The root of global tree structure */
  node< int >* global_root;
  /** The block node in the global tree structure */
  node< int >* global_node;

  /** The parent block's processor */
  int proc_parent;
  /** The left-child block's processor */
  int proc_leftchild;
  /** The right-child block's processor */
  int proc_rightchild;

  ////////////////////////////////////////////////////////////////////////
  // constructor and destructor
  /**
   * @brief Default Constructor.
   *
   * To restrict to construct this data, we do not permit users to use
   * any constructors. 
   */
private:
  dist_tree( );

  /**
   * @brief Constructor copying already distributed tree's structure.
   *
   * To restrict to construct this data, we do not permit users to use
   * any constructors. 
   */
private:
  template< typename B >
  dist_tree( const dist_tree< B >* tree );
  
  /**
   * @brief Destructor
   */
public:
  ~dist_tree( );

  ////////////////////////////////////////////////////////////////////////
  // access methods
public:
  inline const node< A >*
  get_local_root( ) const;

  inline const node< int >*
  get_global_root( ) const;

  inline const node< int >*
  get_global_node( ) const;

  inline int
  get_proc_parent( ) const;
  
  inline int
  get_proc_leftchild( ) const;
  
  inline int
  get_proc_rightchild( ) const;
  
  ////////////////////////////////////////////////////////////////////////
  // Private Utilities

  /**
   * @brief Copy the shape from an already constructed tree. 
   *
   * This function construct by setting the global shape with the same shape
   * as the given tree. 
   *
   * @param tree : A distributed tree to copy the shape of.
   */
private:
  template< typename B >
  void copy_shape( const dist_tree< B >* tree );

private:
  node< int >* copy_shape_rec( const node< int >* root,
			       const node< int >* orgPointer );

  ////////////////////////////////////////////////////////////////////////
  // Input and Output

  /**
   * @brief Construct a tree from a data file
   *
   * The file format of the distributed tree is as follows.
   *  - 1st line: An integer expressing the number of distributed parts (blocks)
   *        @c p.
   *  - 2nd - (p + 1)th line: Each line expresses flag (0 = internal block,
   *        1 = leaf block) and the number of nodes in the block in a pre-fix
   *        order traversal. 
   *  - rest: The following each line expresses a node with flag (0 = internal
   *        node, 1 = leaf, 2 = cut node) and value. The nodes are output
   *        in a pre-fix order.
   *
   * @param filename: Input data file
   */
public:
  static dist_tree< A >* read_from_file( const char* filename );

private:
  void read_from_file_create_global_tree( const int* flags,
					  const int* localsizes );

  /**
   * @param flags : Pointer to the array of flags. This is used for
   *                representing a node is a leaf, a internal node. 
   * @param localsizes : Pointer to the array of sizes for each blocks.
   * @param count : Pointer used to count the index. 
   * @param parentflag : 0 = Do nothing, 1 = Set as left child,
   *                     2 = Set as right child
   * @param parentcount : count of the parent node
   */
private:
  node< int >* read_from_file_create_global_tree_rec( const int** flags,
						      const int** localsizes,
						      int *count,
						      int parentflag,
						      int parentcount );

private:
  void read_from_file_create_subtree( const int* flags,
				      const A* values );

private:
  static node< A >* read_from_file_create_subtree_rec( const int** flags,
						       const A** values );

  /**
   * @brief Output the tree into a data file
   *
   * About the output format, please refer the explanation of
   * @a readFromFile function. 
   *
   * @param filename: Output data file
   * @retval true : The tree is correctly written to the file.
   * @retval false : Something wrong occurred.
   */
public:
  bool write_to_file( const char* filename ) const;

private:
  void write_to_file_global_rec( const node< int >* tree,
				 std::fstream &fout ) const;
  
private:
  void write_to_file_other_subtree( node< int > *root,
				    int *proc,
				    std::fstream &fout,
				    const int TAG ) const;

  /**
   * @brief Simple Print function.
   *
   * This function is only for debugging, because the order of output
   * is not guaranteed. 
   */
public:
  void print( ) const;

private:
  template< typename B >
  void print_rec( node< B >* root ) const;

};

#include "dist_tree.tpp"

#endif /* __DIST_TREE_H__ */
