/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file dist_tree.tpp
 * @brief The implementation of class @a dist_tree, the data for distributed
 *        treed and read/write functions. 
 * @author Kiminori MATSUZAKI
 */

#ifdef __DIST_TREE_H__
#ifndef __DIST_TREE_TPP__
#define __DIST_TREE_TPP__

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <assert.h>

#include <mpi.h>

#include "../util/skeleton.h"

#include "node.h"
// Functions in this file uses static functions in tree_util.
#include "tree_util.h" 

////////////////////////////////////////////////////////////////////////
// constructor and destructor

/* private */
template< typename A >
dist_tree< A >::dist_tree( )
  : local_root( NULL ),
    global_root( NULL ),
    global_node( NULL ),
    proc_parent( -1 ),
    proc_leftchild( -1 ),
    proc_rightchild( -1 )
{
}

/* private */
template< typename A > template< typename B >
dist_tree< A >::dist_tree( const dist_tree< B >* tree )
{
  copy_shape( tree );
}

/* public */
template< typename A >
dist_tree< A >::~dist_tree( )
{
  global_node = NULL;
  tree_util::free_all_tree( local_root ); local_root = NULL;
  tree_util::free_all_tree( global_root ); global_root = NULL;
}

////////////////////////////////////////////////////////////////////////
// access methods

/* public */
template< typename A >
const node< A >*
dist_tree< A >::get_local_root( ) const
{
  return this->local_root;
}

/* public */
template< typename A >
const node< int >*
dist_tree< A >::get_global_root( ) const
{
  return this->global_root;
}

/* public */
template< typename A >
const node< int >*
dist_tree< A >::get_global_node( ) const
{
  return this->global_node;
}

/* public */
template< typename A >
int
dist_tree< A >::get_proc_parent( ) const
{
  return this->proc_parent;
}
  
/* public */
template< typename A >
int
dist_tree< A >::get_proc_leftchild( ) const
{
  return this->proc_leftchild;
}
  
/* public */
template< typename A >
int
dist_tree< A >::get_proc_rightchild( ) const
{
  return this->proc_rightchild;
}

////////////////////////////////////////////////////////////////////////
// Private Utilities

/* private */
template< typename A > template< typename B >
void
dist_tree< A >::copy_shape( const dist_tree< B >* tree )
{
  proc_parent = tree->get_proc_parent( );
  proc_leftchild = tree->get_proc_leftchild( );
  proc_rightchild = tree->get_proc_rightchild( );

  local_root = NULL;
  
  global_node = NULL;
  global_root = copy_shape_rec( tree->get_global_root( ),
				tree->get_global_node( ) );
//   node< int > *ret = new node< int >( );
//   delete ret;
}

/* private */
template< typename A >
node< int > *
dist_tree< A >::copy_shape_rec( const node< int >* root,
				const node< int >* orgPointer )
{
  node< int > *ret = new node< int >( root->value, node< int >::NORMAL );
  if ( root == orgPointer ) {
    this->global_node = ret;
  }

  if ( !root->isLeaf( ) ) {
    ret->left = copy_shape_rec( root->left, orgPointer );
    ret->left->parent = ret;
    
    ret->right = copy_shape_rec( root->right, orgPointer );
    ret->right->parent = ret;
  }

  return ret;
}

////////////////////////////////////////////////////////////////////////
// Input and Output

/* public, static */
template< typename A >
dist_tree< A >*
dist_tree< A >::read_from_file( const char* filename )
{
  const int TAG = 0;
  dist_tree< A >* ret_tree = new dist_tree< A >( );

  if ( skeleton::rank == 0 ) {
    std::fstream fin( filename, std::ios::in );
    if ( fin.fail( ) ) { return NULL; }
      
    // number of blocks
    int np; fin >> np;
    if ( np > skeleton::procs ) {
      std::cerr << "nprocs must be larger than the number of distribution"
		<< std::endl;
      std::exit( 2 );
    }
    MPI_Bcast( &np, 1, MPI_INT, 0, MPI_COMM_WORLD );
    
    // global trees
    int *localsizes = new int[ np ];
    {
      int *flags = new int[ np ];
      
      for ( int p = 0; p < np; p++ ) {
	fin >> flags[ p ] >> localsizes[ p ];
      }
      MPI_Bcast( flags, np, MPI_INT, 0, MPI_COMM_WORLD );
      MPI_Bcast( localsizes, np, MPI_INT, 0, MPI_COMM_WORLD );
    
      ret_tree->read_from_file_create_global_tree( flags, localsizes );
      delete[] flags;
    }
    
    // root subtree
    {
      int *flags = new int[ localsizes[ 0 ] ];
      A *values = new A[ localsizes[ 0 ] ];

      for ( int i = 0; i < localsizes[ 0 ]; i++ ) {
	fin >> flags[ i ] >> values[ i ];
      }
      ret_tree->read_from_file_create_subtree( flags, values );

      delete[] values;
      delete[] flags;
    }

    // other subtrees
    for ( int p = 1; p < np; p++ ) {
      int *flags = new int[ localsizes[ p ] ];
      A *values = new A[ localsizes[ p ] ];

      for ( int i = 0; i < localsizes[ p ]; i++ ) {
	fin >> flags[ i ] >> values[ i ];
      }

      MPI_Request req1, req2;
      MPI_Isend( flags, localsizes[ p ], MPI_INT,
		 p, TAG, MPI_COMM_WORLD, &req1 );
      MPI_Isend( values, localsizes[ p ] * sizeof( A ), MPI_BYTE,
		 p, TAG, MPI_COMM_WORLD, &req2 );
      MPI_Wait( &req1, MPI_STATUS_IGNORE );
      MPI_Wait( &req2, MPI_STATUS_IGNORE );

      delete[] values;
      delete[] flags;
    }
    delete [] localsizes;
  }
  else {
    // number of blocks
    int np;
    MPI_Bcast( &np, 1, MPI_INT, 0, MPI_COMM_WORLD );

    // global trees
    {
      int *localsizes = new int[ np ];
      int *flags = new int[ np ];

      MPI_Bcast( flags, np, MPI_INT, 0, MPI_COMM_WORLD );
      MPI_Bcast( localsizes, np, MPI_INT, 0, MPI_COMM_WORLD );

      ret_tree->read_from_file_create_global_tree( flags, localsizes );
      
      delete[] flags;
      delete[] localsizes;
    }

    // subtree
    if ( skeleton::rank < np ) {
      const int localsize = ret_tree->global_node->value;
      int *flags = new int[ localsize ];
      A *values = new A[ localsize ];

      MPI_Request req1, req2;
      MPI_Irecv( flags, localsize, MPI_INT,
		 0, TAG, MPI_COMM_WORLD, &req1 );
      MPI_Irecv( values, localsize * sizeof( A ), MPI_BYTE,
		 0, TAG, MPI_COMM_WORLD, &req2 );
      MPI_Wait( &req1, MPI_STATUS_IGNORE );
      MPI_Wait( &req2, MPI_STATUS_IGNORE );

      ret_tree->read_from_file_create_subtree( flags, values );

      delete[] values;
      delete[] flags;
    }
  }
  return ret_tree;
}

/* private */
template< typename A >
void
dist_tree< A >::read_from_file_create_global_tree( const int* flags,
						   const int* localsizes )
{
  int count = 0;
  global_root
    = read_from_file_create_global_tree_rec( &flags, &localsizes,
					     &count, 0, -1 );
}

/* private */
template< typename A >
node< int >*
dist_tree< A >::read_from_file_create_global_tree_rec( const int** flags,
						       const int** localsizes,
						       int *count,
						       int parentflag,
						       int parentcount )
{
  const int& flag = *((*flags)++);
  const int& localsize = *((*localsizes)++);

  int nodecount = (*count)++;
  int leftflag = 0, rightflag = 0;
  
  node< int >* new_node = new node< int >( localsize, node< int >::NORMAL );
  if ( nodecount == skeleton::rank ) {
    global_node = new_node;
    proc_parent = parentcount;
    leftflag = 1;
    rightflag = 2;
  }
  if ( parentflag == 1 ) {
    proc_leftchild = nodecount;
  }
  if ( parentflag == 2 ) {
    proc_rightchild = nodecount;
  }
      
  if ( flag == node< int >::SER_NODE ) {
    new_node->left
      = read_from_file_create_global_tree_rec( flags, localsizes,
					       count, leftflag, nodecount );
    new_node->left->parent = new_node;
    
    new_node->right
      = read_from_file_create_global_tree_rec( flags, localsizes,
					       count, rightflag, nodecount );
    new_node->right->parent = new_node;
  }

  return new_node;
}

/* private */
template< typename A >
void
dist_tree< A >::read_from_file_create_subtree( const int* flags,
					       const A* values )
{
  local_root = read_from_file_create_subtree_rec( &flags, &values );
}

/* private, static */
template< typename A >
node< A >*
dist_tree< A >::read_from_file_create_subtree_rec( const int** flags,
						   const A** values )
{
  const int& flag = *((*flags)++);
  const A& value = *((*values)++);

  node< A >* new_node = new node< A >( value, 
				       ( flag == node< A >::SER_CUTNODE ) ? node< A >::CUTNODE : node< A >::NORMAL );

  if ( flag == node< A >::SER_NODE ) {
    new_node->left = read_from_file_create_subtree_rec( flags, values );
    new_node->left->parent = new_node;
    
    new_node->right = read_from_file_create_subtree_rec( flags, values );
    new_node->right->parent = new_node;
  }

  return new_node;
}

/* public */
template< typename A >
bool
dist_tree< A >::write_to_file( const char *filename ) const
{
  const int TAG = 120;
  
  if ( skeleton::rank == 0 ) {
    // master process
    std::fstream fout( filename, std::ios::out );
    if ( fout.fail( ) ) std::exit( 1 );

    // number of processors
    int np = tree_util::get_size_rec( global_root );
    fout << np << std::endl;

    // global trees
    write_to_file_global_rec( global_root, fout );

    // local subtrees
    {
      int *flags    = new int[ global_node->value ];
      A   *values   = new A[ global_node->value ];

      int *p_flags  = flags;
      A   *p_values = values;

      tree_util::serialize_subtree( local_root, &p_flags, &p_values );
      for ( int i = 0; i < global_node->value; i++ ) {
	fout << flags[ i ] << " " << values[ i ] << std::endl;
      }
      delete [] values;
      delete [] flags;
    }

    // other subtrees
    int proc = 1;
    write_to_file_other_subtree( global_root->left,  &proc, fout, TAG );
    write_to_file_other_subtree( global_root->right, &proc, fout, TAG );
  }
  else if ( global_node ) {
    // slave process
    
    int *flags  = new int[ global_node->value ];
    A   *values = new A[ global_node->value ];

    int *p_flags = flags;
    A   *p_values = values;

    tree_util::serialize_subtree( local_root, &p_flags, &p_values );

    MPI_Request req1, req2;
    MPI_Isend( flags, global_node->value, MPI_INT,
	       0, TAG, MPI_COMM_WORLD, &req1 );
    MPI_Isend( values, global_node->value * sizeof( A ), MPI_BYTE,
	       0, TAG , MPI_COMM_WORLD, &req2 );
    MPI_Wait( &req1, MPI_STATUS_IGNORE );
    MPI_Wait( &req2, MPI_STATUS_IGNORE );

    delete [] values;
    delete [] flags;
  } else {
    // idle process
  }
  
  return true;
}

/* private */
template< typename A >
void
dist_tree< A >::write_to_file_global_rec( const node< int >* tree,
					  std::fstream &fout ) const
{
  if ( tree->isLeaf( ) ) {
    fout << node< int >::SER_LEAF << " " << tree->value << std::endl;
  } else {
    fout << node< int >::SER_NODE << " " << tree->value << std::endl;

    write_to_file_global_rec( tree->left,  fout );
    write_to_file_global_rec( tree->right, fout );
  }
}


/* private */
template< typename A >
void
dist_tree< A >::write_to_file_other_subtree( node< int > *tree,
					     int *proc,
					     std::fstream &fout,
					     const int TAG ) const
{
  if ( tree == NULL ) return;
  const int p = (*proc)++;
  const int size = tree->value;
  int *flags = new int[ size ];
  A *values = new A[ size ];

  MPI_Request req1, req2;
  MPI_Irecv( flags, size, MPI_INT,
	     p, TAG, MPI_COMM_WORLD, &req1 );
  MPI_Irecv( values, size * sizeof( A ), MPI_BYTE,
	     p, TAG, MPI_COMM_WORLD, &req2 );
  MPI_Wait( &req1, MPI_STATUS_IGNORE );
  MPI_Wait( &req2, MPI_STATUS_IGNORE );
  
  for ( int i = 0; i < size; i++ ) {
    fout << flags[ i ] << " " << values[ i ] << std::endl;
  }

  delete [] values;
  delete [] flags;

  if ( !tree->isLeaf( ) ) {
    write_to_file_other_subtree( tree->left,  proc, fout, TAG );
    write_to_file_other_subtree( tree->right, proc, fout, TAG );
  }
}



/* public */
template< typename A >
void
dist_tree< A >::print( ) const
{
  if ( !local_root ) return;
  
  if ( skeleton::rank == 0 ) {
    print_rec( global_root );
  }

  print_rec( local_root );

  // for debug
  //   std::cout << "[" << skeleton::rank << "]"
  // 	    << " parent = " << proc_parent << ","
  // 	    << " left = " << proc_leftchild << ","
  // 	    << " right = " << proc_rightchild << std::endl;
}

/* private */
template< typename A > template< typename B >
void
dist_tree< A >::print_rec( node< B >* root ) const
{
  std::cout << "[" << skeleton::rank << "] "
	    << ( ( root->nodetype == node< B >::CUTNODE ) ? "C" :
		 ( ( root->isLeaf( ) ) ? "L" : "N" ) )
	    << ": " << root->value << std::endl;

  if ( !root->isLeaf( ) ) {
    print_rec( root->left );
    print_rec( root->right );
  }
}


#endif /* __DIST_TREE_TPP__ */
#endif /* __DIST_TREE_H__ */

