/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file rose_tree_skeletons.h
 * @brief The header file of class @a rose_tree_skeletons;
 *        a collection of skeletons on rose trees. 
 * @author Kiminori MATSUZAKI
 */

#ifndef __ROSE_TREE_SKELETONS_H__
#define __ROSE_TREE_SKELETONS_H__

#include "dist_rose_tree.h"
#include "tree_skeletons.h"

/**
 * @brief A collection of skeletons on rose trees.
 *
 * Currently, we show the parallelizable form only for @a map, @a zip,
 * and reduce.
 */
class rose_tree_skeletons
{

  /**
   * @brief Map skeleton on rose trees.
   *
   * This skeleton accepts two functions, @a k1 and @a k2, and applies @a k1
   * to every leaf and applies @a k2 for every node on rose trees.
   *
   * Informal definition (in Haskell-like-notation):
   * @code
   *   map :: (a -> c) -> (a -> c) -> RTree a a -> RTree c c
   *   map k1 k2 (RLeaf a) = RLeaf (k1 a)
   *   map k1 k2 (RNode b [t1, t2, ..., tn]
   *       = RNode (k2 b) [map k1 k2 t1, map k1 k2 t2, ..., map k1 k2 tn]
   * @endcode
   *
   * Formal definition (using BMF notation):
   * @code
   *   map k1 k2 (RLeaf a)    = RLeaf (k1 a)
   *   map k1 k2 (RNode b ts) = RNode (k2 b) (mapL (map k1 k2) ts)
   * @endcode
   * 
   * To call this skeleton, you need to specify the types below.
   *  - K1 = type of function object for k1 (option), 
   *  - K2 = type of function object for k2 (option).
   *
   * Sample: \n
   * < Haskell code >
   * @code
   *   t2 = map k1 k2 t1
   * @endcode
   * < C++ code >
   * @code
   *   t2 = rose_tree_skeletons::map< c >( k1, k2, t1 );
   * @endcode
   *
   * @param k1 : The function object applied to every leaf,
   *             and this function needs to inherit @c unary_function
   *             with suitable types. 
   * @param k2 : The function object applied to every internal node, 
   *             and this function needs to inherit @c unary_function
   *             with suitable types. 
   * @param t : A distributed rose tree for input. 
   */
public:
  template< typename K1, typename K2 >
  static dist_rose_tree< typename K1::result_type > *
  map( const K1 &k1,
       const K2 &k2,
       const dist_rose_tree< typename K1::argument_type > *t );

private:
  template< typename C, typename A, typename K1, typename K2 >
  static dist_rose_tree< C > *map_adapter( const K1 &k1,
					   const K2 &k2,
					   const dist_rose_tree< A > *t );
  
private:
  template< typename C, typename A, typename K1 >
  class map_k1_t : public unary_function< rose_value< A >,
					  rose_value< C > >
  {
  private:
    const K1 &k1;
  public:
    map_k1_t( const K1 &k1 );
    rose_value< C > operator()( const rose_value< A > &a ) const;
  };

private:
  template< typename C, typename A, typename K2 >
  class map_k2_t : public unary_function< rose_value< A >,
					  rose_value< C > >
  {
  private:
    const K2 &k2;
  public:
    map_k2_t( const K2 &k2 );
    rose_value< C > operator()( const rose_value< A > &b ) const;
  };

  /**
   * @brief Zipwith skeleton on rose trees.
   *
   * This skeleton accepts two rose trees @a at and @a bt, and zips them up
   * into a rose tree by applying @a k1 for every pair of leaves and
   * applying @a k2 for every pair of internal nodes.
   *
   * Informal definition (in Haskell-like notation):
   * @code
   *   zipwith :: (a -> c -> e) -> (a -> c -> e) -> RTree a a -> RTree c c -> RTree e e
   *   zipwith k1 k2 (RLeaf a) (RLeaf a') = RLeaf (k1 a a')
   *   zipwith k1 k2 (RNode b [t1, t2, ..., tn]) (RNode b' [t1', t2', ..., tn'])
   *     = RNode (k2 b b') [zipwith k1 k2 t1 t1', zipwith k1 k2 t2 t2', ..., zipwith k1 k2 tn tn']
   * @endcode
   *
   * Formal definition (using BMF notation):
   * @code
   *   zipwith k1 k2 (RLeaf a) (RLeaf a')        = RLeaf (k1 a a')
   *   zipwith k1 k2 (RNode b ts) (RNode b' ts') = RNode (k2 b b') (zipwithL (zipwith k1 k2) ts ts')
   * @endcode
   *
   * To call this skeleton, you need to specify the types below.
   *  - K1 = type of function object for k1 (option),
   *  - K2 = type of function object for k2 (option).
   *
   * Sample: \n
   * < Haskell code >
   * @code
   *   t = zipwith k1 k2 at bt
   * @endcode
   * < C++ code >
   * @code
   *   t = rose_tree_skeletons::zipwith< e >( k1, k2, at, bt );
   * @endcode
   * 
   * @param k1 : The function object applied to every pair of leaves.
   *             This function needs to inherit
   *             @code binary_function< A, C, E > @endcode
   * @param k2 : The function object applied to every pair of internal nodes. 
   *             This function needs to inherit
   *             @code binary_function< A, C, E > @endcode
   * @param at : The first rose tree to zip up.
   * @param bt : The second rose tree to zip up.
   */
public:
  template< typename K1, typename K2 >
  static dist_rose_tree< typename K1::result_type > *
  zipwith( const K1 &k1,
	   const K2 &k2,
	   const dist_rose_tree< typename K1::first_argument_type >* at, 
	   const dist_rose_tree< typename K1::second_argument_type >* bt );

private:
  template< typename E, typename A, typename C, typename K1, typename K2 >
  static dist_rose_tree< E > *zipwith_adapter( const K1 &k1,
					       const K2 &k2,
					       const dist_rose_tree< A > *at,
					       const dist_rose_tree< C > *bt );
				      
private:
  template< typename E, typename A, typename C, typename K1 >
  class zip_k1_t : public binary_function< rose_value< A >,
					   rose_value< C >,
					   rose_value< E > >
  {
  private:
    const K1 &k1;
  public:
    zip_k1_t( const K1 &k1 );
    rose_value< E > operator()( const rose_value< A > &a,
				const rose_value< C > &c ) const;
  };

private:
  template< typename E, typename A, typename C, typename K2 >
  class zip_k2_t : public binary_function< rose_value< A >,
					   rose_value< C >,
					   rose_value< E > >
  {
  private:
    const K2 &k2;
  public:
    zip_k2_t( const K2 &k2 );
    rose_value< E > operator()( const rose_value< A > &a,
				const rose_value< C > &c ) const;
  };

  /**
   * @brief Reduce skeleton on rose trees. 
   *
   * This skeleton first applies @a k1 for every leaf and @a k2 for every
   * internal node, followed by collapsing with operator @a otimes among
   * children and @a oplus between children and parent.
   *
   * For parallel implementation, @a otimes must be associative and
   * be extended-distributive over @a oplus. See "Conditions" for more
   * detail about the conditions on @a otimes, @a oplus, and definitions
   * of @a p1, @a p2, and @a p3.
   *
   * Informal Definition (in Haskell-like notation):
   * @code
   *   reduce :: (a -> c) -> (a -> d) -> (d -> c -> c) -> (c -> c -> c) <assoc>
   *             -> BTree a a -> c
   *   reduce k1 k2 (oplus) (otimes) (RLeaf a) = k1 a
   *   reduce k1 k2 (oplus) (otimes) (RNode b [t1, t2, ..., tn])
   *       = (k2 b) oplus ((reduce k1 k2 t1) otimes (reduce k1 k2 t2) otimes ... otimes (reduce k1 k2 tn))
   * @endcode
   *
   * Formal Definition (using BMF notation):
   * @code
   *   reduce k1 k2 (oplus) (otimes) (RLeaf a) = k1 a
   *   reduce k1 k2 (oplus) (otimes) (RNode b ts)
   *       = (k2 b) oplus (reduceL (otimes) (mapL (reduce k1 k2) ts))
   * @endcode
   *
   * Conditions:\n
   * To automatically derive a parallel implementation, you need to specify
   * @a e_oplus and @a e_otimes which are the (left-) unit of @a oplus and
   * the unit @a otimes, respectively. Furthermore, you need to derive three
   * characteristic functions, @a p1, @a p2, and @a p3, such that the
   * following equation holds for any values, a1, b1, c1, a2, b2, and c2.
   * @code
   *     (\x -> a1 oplus (b1 otimes x otimes c1)) . (\x -> a2 oplus (b2 otimes x otimes c2))
   *   = (\x -> A oplus (B otimes x otimes C))
   *     where A = p1 (a1, b1, c1, a2, b2, c2)
   *           B = p2 (a1, b1, c1, a2, b2, c2)
   *           C = p3 (a1, b1, c1, a2, b2, c2)
   * @endcode
   *
   * To call this skeleton, you don't need to specify template types below.
   *  - K1 : type of function object for k1 (option),
   *  - K2 : type of function object for k2 (option),
   *  - OPLUS : type of function object for oplus (option),
   *  - OTIMES : type of function object for otimes (option),
   *  - P1 : type of function object for otimes (option),
   *  - P2 : type of function object for otimes (option),
   *  - P3 : type of function object for otimes (option).
   *
   * Sample: \n
   * < Haskell code (using function in the definition) >
   * @code
   *   v = reduce k1 k2 (oplus) (otimes) t
   * @endcode
   * < C++ code (including all function in the definition) >
   * @code
   *   v = rose_tree_skeletons::reduce( k1, k2, oplus, e_oplus, otimes o_otimes,
   *                                    p1, p2, p3, t );
   * @endcode
   *
   * @param k1: Function object applied to every leaf, and
   *            this function needs to inherit
   *            @code unary_function< A, C > @endcode
   * @param k2: Function object applied to every internal node, and
   *            this function needs to inherit
   *            @code unary_function< A, D > @endcode
   * @param oplus: Function object for composing parent and result of children,
   *            and this function needs to inherit
   *            @code binary_function< D, C, C > @endcode
   * @param e_oplus: The (left-) unit of operator @a oplus. 
   * @param otimes: Function object for folding children, and 
   *            and this function needs to inherit
   *            @code binary_function< C, C, C > @endcode
   * @param e_otimes: The unit of operator @a otimes. 
   * @param p1: Function object for the first of characteristic functions, and
   *            this needs to implement the following operator()
   *            @code D operator()( const D&, const C&, const C&, const D&, const C&, const C& ) @endcode
   * @param p2: Function object for the second of characteristic functions, and
   *            this needs to implement the following operator()
   *            @code C operator()( const D&, const C&, const C&, const D&, const C&, const C& ) @endcode
   * @param p3: Function object for the third of characteristic functions, and
   *            this needs to implement the following operator()
   *            @code C operator()( const D&, const C&, const C&, const D&, const C&, const C& ) @endcode
   * @param t: The distributed rose tree for input
   */
public:
  template< typename K1, typename K2, typename OPLUS, typename OTIMES,
	    typename P1, typename P2, typename P3 >
  static typename K1::result_type
  reduce( const K1 &k1,
	  const K2 &k2,
	  const OPLUS &oplus,
	  const typename K2::result_type &e_oplus,
	  const OTIMES &otimes,
	  const typename K1::result_type &e_otimes,
	  const P1 &p1,
	  const P2 &p2,
	  const P3 &p3,
	  const dist_rose_tree< typename K1::argument_type >* t );

private:
  template< typename C, typename D, typename A,
	    typename K1, typename K2, typename OPLUS, typename OTIMES,
	    typename P1, typename P2, typename P3 >
  static C reduce_adapter( const K1 &k1,
	  const K2 &k2,
	  const OPLUS &oplus,
	  const D &e_oplus,
	  const OTIMES &otimes,
	  const C &e_otimes,
	  const P1 &p1,
	  const P2 &p2,
	  const P3 &p3,
	  const dist_rose_tree< A >* t );

private:
  template< typename C, typename D >
  struct reduce_tup_t
  {
    D a;
    C b;
    C c;
  };

private:
  template< typename C, typename A,
	    typename K1 >
  class reduce_k1_t : public unary_function< rose_value< A >, C >
  {
  private:
    const K1 &k1;
    const C &e_otimes;
  public:
    reduce_k1_t( const K1 &k1, const C &e_otimes );
    C operator()( const rose_value< A > &a ) const;
  };

private:
  template< typename C, typename A,
	    typename K2, typename OPLUS, typename OTIMES >
  class reduce_k2_t
  {
  private:
    const K2 &k2;
    const OPLUS &oplus;
    const OTIMES &otimes;
  public:
    reduce_k2_t( const K2 &k2, const OPLUS &oplus, const OTIMES &otimes);
    C operator()( const rose_value< A > &b, const C &l, const C &r ) const;
  };

private:
  template< typename C, typename D, typename A,
	    typename K2 >
  class reduce_psi_t : public unary_function< rose_value< A >,
					      rose_tree_skeletons::reduce_tup_t< C, D > >
  {
  private:
    const K2 &k2;
    const D &e_oplus;
    const C &e_otimes;
  public:
    reduce_psi_t( const K2 &k2, const D &e_oplus, const C &e_otimes );
    rose_tree_skeletons::reduce_tup_t< C, D >
    operator()( const rose_value< A > &b ) const;
  };

private:
  template< typename C, typename D,
	    typename OTIMES, typename P1, typename P2, typename P3 >
  class reduce_phiL_t : public ternary_function< rose_tree_skeletons::reduce_tup_t< C, D >,
						 C,
						 rose_tree_skeletons::reduce_tup_t< C, D >,
						 rose_tree_skeletons::reduce_tup_t< C, D > >
  {
  private:
    const OTIMES &otimes;
    const P1 &p1;
    const P2 &p2;
    const P3 &p3;
  public:
    reduce_phiL_t( const OTIMES &otiems,
		   const P1 &p1,
		   const P2 &p2,
		   const P3 &p3 );

    rose_tree_skeletons::reduce_tup_t< C, D >
    operator()( const rose_tree_skeletons::reduce_tup_t< C, D > &p,
		const C& l,
		const rose_tree_skeletons::reduce_tup_t< C, D > &r ) const;
  };

private:
  template< typename C, typename D,
	    typename OTIMES, typename P1, typename P2, typename P3 >
  class reduce_phiR_t : public ternary_function< rose_tree_skeletons::reduce_tup_t< C, D >,
						 C,
						 rose_tree_skeletons::reduce_tup_t< C, D >,
						 rose_tree_skeletons::reduce_tup_t< C, D > >
  {
  private:
    const OTIMES &otimes;
    const P1 &p1;
    const P2 &p2;
    const P3 &p3;
  public:
    reduce_phiR_t( const OTIMES &otiems,
		   const P1 &p1,
		   const P2 &p2,
		   const P3 &p3 );

    rose_tree_skeletons::reduce_tup_t< C, D >
    operator()( const rose_tree_skeletons::reduce_tup_t< C, D > &p,
		const C &r,
		const rose_tree_skeletons::reduce_tup_t< C, D > &l ) const;
  };

private:
  template< typename C, typename D,
	    typename OPLUS, typename OTIMES >
  class reduce_G_t : public ternary_function< rose_tree_skeletons::reduce_tup_t< C, D >,
					      C,
					      C,
					      C >
  {
  private:
    const OPLUS &oplus;
    const OTIMES &otimes;
  public:
    reduce_G_t( const OPLUS &oplus, const OTIMES &otimes );
    C operator()( const rose_tree_skeletons::reduce_tup_t< C, D > &a,
		  const C &x,
		  const C &y ) const;
  };

  /**
   * @brief A specialized version of reduce skeleton with commutative operator.
   *
   * This skeleton compute the same as @a reduce skeleton. The difference
   * from it is that guaranteeing the commutativity of @a otimes, not only
   * @a associativity, we can call this skeleton more simply and obtain
   * more efficient computation than normal reduce skeleton.
   * 
   * For parallel implementation, @a otimes must still be extended-distributive
   * over @a oplus. See "Conditions" for more detail about the conditions and
   * definitions of @a p1 and @a p2. 
   *
   * Conditions:\n
   * To automatically derive a parallel implementation, you need to specify
   * @a e_oplus and @a e_otimes which are the (left-) unit of @a oplus and
   * the unit @a otimes, respectively. Furthermore, you need to derive two
   * characteristic functions, @a p1 and @a p2, such that the following
   * equation holds for any values, a1, b1, a2, and b2.
   * @code
   *     (\x -> a1 oplus (b1 otimes x)) . (\x -> a2 oplus (b2 otimes x))
   *   = (\x -> A oplus (B otimes x))
   *     where A = p1 (a1, b1, a2, b2)
   *           B = p2 (a1, b1, a2, b2)
   * @endcode
   *
   * To call this skeleton, you don't need to specify template types below.
   *  - K1 : type of function object for k1 (option),
   *  - K2 : type of function object for k2 (option),
   *  - OPLUS : type of function object for oplus (option),
   *  - OTIMES : type of function object for otimes (option),
   *  - P1 : type of function object for otimes (option),
   *  - P2 : type of function object for otimes (option),
   *
   * Sample: \n
   * < Haskell code (using function in the definition) >
   * @code
   *   v = reduce k1 k2 (oplus) (otimes) t
   * @endcode
   * < C++ code (including all function in the definition) >
   * @code
   *   v = rose_tree_skeletons::reduce( k1, k2, oplus, e_oplus, otimes o_otimes,
   *                                    p1, p2, t );
   * @endcode
   *
   * @param k1: Function object applied to every leaf, and
   *            this function needs to inherit
   *            @code unary_function< A, C > @endcode
   * @param k2: Function object applied to every internal node, and
   *            this function needs to inherit
   *            @code unary_function< A, D > @endcode
   * @param oplus: Function object for composing parent and result of children,
   *            and this function needs to inherit
   *            @code binary_function< D, C, C > @endcode
   * @param e_oplus: The (left-) unit of operator @a oplus. 
   * @param otimes: Function object for folding children, and
   *            this function needs to inherit
   *            @code unary_function< C, C, C > @endcode
   * @param e_otimes: The unit of operator @a otimes. 
   * @param p1: Function object for the first of characteristic functions, and
   *            this needs to implement the following operator()
   *            @code D operator()( const D&, const C&, const D&, const C& ) @endcode
   * @param p2: Function object for the second of characteristic functions, and
   *            this needs to implement the following operator()
   *            @code C operator()( const D&, const C&, const D&, const C& ) @endcode
   * @param t: The distributed rose tree for input
   */
public:
  template< typename K1, typename K2, typename OPLUS, typename OTIMES,
	    typename P1, typename P2 >
  static typename K1::result_type
  reduce_c( const K1 &k1,
	    const K2 &k2, 
	    const OPLUS &oplus,
	    const typename K2::result_type &e_oplus,
	    const OTIMES &otimes,
	    const typename K1::result_type &e_otimes,
	    const P1 &p1,
	    const P2 &p2,
	    const dist_rose_tree< typename K1::argument_type > *t );

private:
  template< typename C, typename D, typename A,
	    typename K1, typename K2, typename OPLUS, typename OTIMES,
	    typename P1, typename P2 >
  static C
  reduce_c_adapter( const K1 &k1,
		    const K2 &k2, 
		    const OPLUS &oplus,
		    const D &e_oplus,
		    const OTIMES &otimes,
		    const C &e_otimes,
		    const P1 &p1,
		    const P2 &p2,
		    const dist_rose_tree< A > *t );

private:
  template< typename C, typename D >
  struct reduce_c_tup_t
  {
    D a;
    C b;
  };
  
private:
  template< typename C, typename A, typename K1 >
  class reduce_c_k1_t : public unary_function< rose_value< A >, C >
  {
  private:
    const K1 &k1;
    const C &e_otimes;
  public:
    reduce_c_k1_t( const K1 &k1, const C &e_otimes );
    C operator()( const rose_value< A > &a ) const;
  };

private:
  template< typename C, typename A,
	    typename K2, typename OPLUS, typename OTIMES >
  class reduce_c_k2_t : public ternary_function< rose_value< A >, C, C, C >
  {
  private:
    const K2 &k2;
    const OPLUS &oplus;
    const OTIMES &otimes;
  public:
    reduce_c_k2_t( const K2 &k2, const OPLUS &oplus, const OTIMES &otimes);
    C operator()( const rose_value< A > &b, const C &l, const C &r ) const;
  };

private:
  template< typename C, typename D, typename A, typename K2 >
  class reduce_c_psi_t : public unary_function< rose_value< A >, 
						rose_tree_skeletons::reduce_c_tup_t< C, D > >
  {
  private:
    const K2 &k2;
    const D &e_oplus;
    const C &e_otimes;
  public:
    reduce_c_psi_t( const K2 &k2, const D &e_oplus, const C &e_otimes );
    rose_tree_skeletons::reduce_c_tup_t< C, D >
    operator()( const rose_value< A >& b ) const;
  };

private:
  template< typename C, typename D,
	    typename OTIMES, typename P1, typename P2 >
  class reduce_c_phi_t : public ternary_function< rose_tree_skeletons::reduce_c_tup_t< C, D >,
						  C,
						  rose_tree_skeletons::reduce_c_tup_t< C, D >,
						  rose_tree_skeletons::reduce_c_tup_t< C, D > >
  {
  private:
    const OTIMES &otimes;
    const P1 &p1;
    const P2 &p2;
  public:
    reduce_c_phi_t( const OTIMES &otimes, const P1 &p1, const P2 &p2 );
    rose_tree_skeletons::reduce_c_tup_t< C, D >
    operator()( const rose_tree_skeletons::reduce_c_tup_t< C, D > &p,
		const C& l,
		const rose_tree_skeletons::reduce_c_tup_t< C, D > &o ) const;
  };

private:
  template< typename C, typename D, typename OPLUS, typename OTIMES >
  class reduce_c_G_t : public ternary_function< rose_tree_skeletons::reduce_c_tup_t< C, D >,
						C,
						C,
						C >
  {
  private:
    const OPLUS &oplus;
    const OTIMES &otimes;
  public:
    reduce_c_G_t( const OPLUS &oplus, const OTIMES &otimes );
    C operator()( const rose_tree_skeletons::reduce_c_tup_t< C, D > &a,
		  const C& x,
		  const C& y ) const;
  };

  /**
   * @brief Distribute the input tree over processors.
   *
   * This function distribute the input tree to the processors.
   * The input tree is needed to be a binary tree whose each value
   * is of type @a rose_value< A >. 
   * @param node: The pointer to the root node of the input tree.
   * @return The pointer to the instance of @a dist_rose_tree< A > which
   *         is distributed over the processors. 
   */
public:
  template< typename A >
  static dist_rose_tree< A > *distribute( node< rose_value< A > > *node );
  
};

#include "rose_tree_skeletons.tpp"

#endif /* __ROSE_TREE_SKELETONS_H__ */
