/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file rose_tree_skeletons.tpp
 * @brief The implementation of class @a rose_tree_skeletons;
 *        a collection of skeletons on rose trees.
 * @author Kiminori MATSUZAKI
 */

#ifdef __ROSE_TREE_SKELETONS_H__
#ifndef __ROSE_TREE_SKELETONS_TPP__
#define __ROSE_TREE_SKELETONS_TPP__

/* public, static */
template< typename K1, typename K2 >
inline dist_rose_tree< typename K1::result_type > *
rose_tree_skeletons::map( const K1 &k1,
			  const K2 &k2,
			  const dist_rose_tree< typename K1::argument_type > *t )
{
  return map_adapter<
    typename K1::result_type,
    typename K1::argument_type,
    K1, K2 >( k1, k2, t );
}
    
/* private, static */
template< typename C, typename A, typename K1, typename K2 >
inline dist_rose_tree< C > *
rose_tree_skeletons::map_adapter( const K1 &k1,
				  const K2 &k2,
				  const dist_rose_tree< A > *t )
{
  dist_tree< rose_value< C > > *new_btree
    = tree_skeletons::map( map_k1_t< C, A, K1 >( k1 ),
			   map_k2_t< C, A, K2 >( k2 ),
			   t->btree );

  return new dist_rose_tree< C >( new_btree );
}

/* private::public */
template< typename C, typename A, typename K1 >
inline rose_tree_skeletons::map_k1_t< C, A, K1 >::map_k1_t( const K1 &k1_ )
  : k1( k1_ )
{
}

/* private::public */
template< typename C, typename A, typename K1 >
inline rose_value< C >
rose_tree_skeletons::map_k1_t< C, A, K1 >::
operator()( const rose_value< A > &a ) const
{
  rose_value< C > ret;
  if ( a.valuetype == rose_value< A >::ORGLEAF ) {
    ret.valuetype = rose_value< C >::ORGLEAF;
    ret.value     = k1( a.value );
  } else {
    ret.valuetype = rose_value< C >::ORGNIL;
  }
  return ret;
}

/* private::public */
template< typename C, typename A, typename K2 >
inline rose_tree_skeletons::map_k2_t< C, A, K2 >::map_k2_t( const K2 &k2_ )
  : k2( k2_ )
{
}

/* private::public */
template< typename C, typename A, typename K2 >
inline rose_value< C >
rose_tree_skeletons::map_k2_t< C, A, K2 >::
operator()( const rose_value< A > &b ) const
{
  rose_value< C > ret;
  if ( b.valuetype == rose_value< A >::ORGNODE ) {
    ret.valuetype = rose_value< C >::ORGNODE;
    ret.value     = k2( b.value );
  } else {
    ret.valuetype = rose_value< C >::DUMMY;
  }
  return ret;
}

/* public, static */
template< typename K1, typename K2 >
inline dist_rose_tree< typename K1::result_type > *
rose_tree_skeletons::zipwith( const K1 &k1,
			      const K2 &k2,
			      const dist_rose_tree< typename K1::first_argument_type >* at, 
			      const dist_rose_tree< typename K1::second_argument_type >* bt )
{
  return zipwith_adapter<
    typename K1::result_type,
    typename K1::first_argument_type,
    typename K1::second_argument_type,
    K1, K2 >( k1, k2, at, bt );
}
  
/* private, static */
template< typename E, typename A, typename C, typename K1, typename K2 >
inline dist_rose_tree< E >*
rose_tree_skeletons::zipwith_adapter( const K1 &k1,
				      const K2 &k2,
				      const dist_rose_tree< A >* at, 
				      const dist_rose_tree< C >* bt )
{
  dist_tree< rose_value< E > > *new_btree
    = tree_skeletons::zipwith( zip_k1_t< E, A, C, K1 >( k1 ),
			       zip_k2_t< E, A, C, K2 >( k2 ),
			       at->btree,
			       bt->btree );

  return new dist_rose_tree< E >( new_btree );
}

/* private::public */
template< typename E, typename A, typename C, typename K1 >
rose_tree_skeletons::zip_k1_t< E, A, C, K1 >::zip_k1_t( const K1 &k1_ )
  : k1( k1_ )
{
}

/* private::public */
template< typename E, typename A, typename C, typename K1 >
rose_value< E >
rose_tree_skeletons::zip_k1_t< E, A, C, K1 >::
operator()( const rose_value< A > &a,
	    const rose_value< C > &c ) const
{
  rose_value< E > ret;
  if ( a.valuetype == rose_value< A >::ORGLEAF ) {
    ret.valuetype = rose_value< E >::ORGLEAF;
    ret.value = k1( a.value, c.value );
  } else {
    ret.valuetype = rose_value< E >::ORGNIL;
  }
  return ret;
}

/* private::public */
template< typename E, typename A, typename C, typename K2 >
rose_tree_skeletons::zip_k2_t< E, A, C, K2 >::zip_k2_t( const K2 &k2_ )
  : k2( k2_ )
{
}

/* private::public */
template< typename E, typename A, typename C, typename K2 >
rose_value< E >
rose_tree_skeletons::zip_k2_t< E, A, C, K2 >::
operator()( const rose_value< A > &a,
	    const rose_value< C > &c ) const
{
  rose_value< E > ret;
  if ( a.valuetype == rose_value< A >::ORGNODE ) {
    ret.valuetype = rose_value< E >::ORGNODE;
    ret.value = k2( a.value, c.value );
  } else {
    ret.valuetype = rose_value< E >::DUMMY;
  }
  return ret;
}

/* public, static */
template< typename K1, typename K2, typename OPLUS, typename OTIMES,
	  typename P1, typename P2, typename P3 >
inline typename K1::result_type
rose_tree_skeletons::reduce( const K1 &k1,
			     const K2 &k2,
			     const OPLUS &oplus,
			     const typename K2::result_type &e_oplus,
			     const OTIMES &otimes,
			     const typename K1::result_type &e_otimes,
			     const P1 &p1,
			     const P2 &p2,
			     const P3 &p3,
			     const dist_rose_tree< typename K1::argument_type >* t )
{
  return reduce_adapter<
    typename K1::result_type,
    typename K2::result_type,
    typename K1::argument_type,
    K1, K2, OPLUS, OTIMES, P1, P2, P3 >
    ( k1, k2, oplus, e_oplus, otimes, e_otimes, p1, p2, p3, t );
}

/* private, static */
template< typename C, typename D, typename A,
	  typename K1, typename K2, typename OPLUS, typename OTIMES,
	  typename P1, typename P2, typename P3 >
inline C
rose_tree_skeletons::reduce_adapter( const K1 &k1,
			     const K2 &k2,
			     const OPLUS &oplus,
			     const D &e_oplus,
			     const OTIMES &otimes,
			     const C &e_otimes,
			     const P1 &p1,
			     const P2 &p2,
			     const P3 &p3,
			     const dist_rose_tree< A >* t )
{
  return tree_skeletons::reduce( reduce_k1_t< C, A, K1 >( k1, e_otimes ),
				 reduce_k2_t< C, A, K2, OPLUS, OTIMES >( k2, oplus, otimes ),
				 reduce_psi_t< C, D, A, K2 >( k2, e_oplus, e_otimes ),
				 reduce_phiL_t< C, D, OTIMES, P1, P2, P3 >( otimes, p1, p2, p3 ),
				 reduce_phiR_t< C, D, OTIMES, P1, P2, P3 >( otimes, p1, p2, p3 ),
				 reduce_G_t< C, D, OPLUS, OTIMES >( oplus, otimes ),
				 t->btree );
}

/* private::public */
template< typename C, typename A, typename K1 >
rose_tree_skeletons::reduce_k1_t< C, A, K1 >::
reduce_k1_t( const K1 &k1_,
	     const C &e_otimes_ )
  : k1( k1_ ),
    e_otimes( e_otimes_ )
{
}

/* private::public */
template< typename C, typename A, typename K1 >
C
rose_tree_skeletons::reduce_k1_t< C, A, K1 >::
operator()( const rose_value< A > &a ) const
{
  C ret;
  if ( a.valuetype == rose_value< A >::ORGLEAF ) {
    ret = k1( a.value );
  } else {
    ret = e_otimes;
  }
  return ret;
}

/* private::public */
template< typename C, typename A,
	  typename K2, typename OPLUS, typename OTIMES >
rose_tree_skeletons::reduce_k2_t< C, A, K2, OPLUS, OTIMES >::
reduce_k2_t( const K2 &k2_,
	     const OPLUS &oplus_,
	     const OTIMES &otimes_)
  : k2( k2_ ),
    oplus( oplus_ ),
    otimes( otimes_ )
{
}

/* private::public */
template< typename C, typename A,
	  typename K2, typename OPLUS, typename OTIMES >
C
rose_tree_skeletons::reduce_k2_t< C, A, K2, OPLUS, OTIMES >::
operator()( const rose_value< A > &b,
	    const C &l,
	    const C &r ) const
{
  C ret;
  if ( b.valuetype == rose_value< A >::ORGNODE ) {
    ret = oplus( k2( b.value ), otimes( l, r ) );
  } else {
    ret = otimes( l, r );
  }
  return ret;
}

/* private::public */
template< typename C, typename D, typename A, typename K2 >
rose_tree_skeletons::reduce_psi_t< C, D, A, K2 >::
reduce_psi_t( const K2 &k2_,
	      const D &e_oplus_,
	      const C &e_otimes_ )
  : k2( k2_ ),
    e_oplus( e_oplus_ ),
    e_otimes( e_otimes_ )
{
}

/* private::public */
template< typename C, typename D, typename A, typename K2 >
rose_tree_skeletons::reduce_tup_t< C, D >
rose_tree_skeletons::reduce_psi_t< C, D, A, K2 >::
operator()( const rose_value< A > &b ) const
{
  rose_tree_skeletons::reduce_tup_t< C, D > ret;
  if ( b.valuetype == rose_value< A >::ORGNODE ) {
    ret.a = k2( b.value );
    ret.b = e_otimes;
    ret.c = e_otimes;
  } else {
    ret.a = e_oplus;
    ret.b = e_otimes;
    ret.c = e_otimes;
  }
  return ret;
}

/* private::public */
template< typename C, typename D,
	  typename OTIMES, typename P1, typename P2, typename P3 >
rose_tree_skeletons::reduce_phiL_t< C, D, OTIMES, P1, P2, P3 >::
reduce_phiL_t( const OTIMES &otimes_,
	       const P1 &p1_,
	       const P2 &p2_,
	       const P3 &p3_ )
  : otimes( otimes_ ),
    p1( p1_ ),
    p2( p2_ ),
    p3( p3_ )
{
}

/* private::public */
template< typename C, typename D,
	  typename OTIMES, typename P1, typename P2, typename P3 >
rose_tree_skeletons::reduce_tup_t< C, D >
rose_tree_skeletons::reduce_phiL_t< C, D, OTIMES, P1, P2, P3 >::
operator()( const rose_tree_skeletons::reduce_tup_t< C, D > &p,
	    const C& l,
	    const rose_tree_skeletons::reduce_tup_t< C, D > &r ) const
{
  rose_tree_skeletons::reduce_tup_t< C, D > ret;
  C b1_otimes_l = otimes( p.b, l );
  ret.a = p1( p.a, b1_otimes_l, p.c, r.a, r.b, r.c );
  ret.b = p2( p.a, b1_otimes_l, p.c, r.a, r.b, r.c );
  ret.c = p3( p.a, b1_otimes_l, p.c, r.a, r.b, r.c );
  
  return ret;
}   

/* private::public */
template< typename C, typename D,
	  typename OTIMES, typename P1, typename P2, typename P3 >
rose_tree_skeletons::reduce_phiR_t< C, D, OTIMES, P1, P2, P3 >::
reduce_phiR_t( const OTIMES &otimes_,
	       const P1 &p1_,
	       const P2 &p2_,
	       const P3 &p3_ )
  : otimes( otimes_ ),
    p1( p1_ ),
    p2( p2_ ),
    p3( p3_ )
{
}

/* private::public */
template< typename C, typename D,
	  typename OTIMES, typename P1, typename P2, typename P3 >
rose_tree_skeletons::reduce_tup_t< C, D >
rose_tree_skeletons::reduce_phiR_t< C, D, OTIMES, P1, P2, P3 >::
operator()( const rose_tree_skeletons::reduce_tup_t< C, D > &p,
	    const C &r,
	    const rose_tree_skeletons::reduce_tup_t< C, D > &l ) const
{
  rose_tree_skeletons::reduce_tup_t< C, D > ret;
  C r_otimes_c1 = otimes( r, p.c );
  ret.a = p1( p.a, p.b, r_otimes_c1, l.a, l.b, l.c );
  ret.b = p2( p.a, p.b, r_otimes_c1, l.a, l.b, l.c );
  ret.c = p3( p.a, p.b, r_otimes_c1, l.a, l.b, l.c );

  return ret;
}   

/* private::public */
template< typename C, typename D, typename OPLUS, typename OTIMES >
rose_tree_skeletons::reduce_G_t< C, D, OPLUS, OTIMES >::
reduce_G_t( const OPLUS &oplus_,
	    const OTIMES &otimes_ )
  : oplus( oplus_ ),
    otimes( otimes_ )
{
}

/* private::public */
template< typename C, typename D, typename OPLUS, typename OTIMES >
C
rose_tree_skeletons::reduce_G_t< C, D, OPLUS, OTIMES >::
operator()( const rose_tree_skeletons::reduce_tup_t< C, D > &abc,
	    const C &x,
	    const C &y ) const
{
  return oplus( abc.a, otimes( abc.b, otimes( x, otimes( y, abc.c ) ) ) );
}

/* public, static */
template< typename K1, typename K2, typename OPLUS, typename OTIMES,
	  typename P1, typename P2 >
inline typename K1::result_type
rose_tree_skeletons::reduce_c( const K1 &k1,
			       const K2 &k2, 
			       const OPLUS &oplus,
			       const typename K2::result_type &e_oplus,
			       const OTIMES &otimes,
			       const typename K1::result_type &e_otimes,
			       const P1 &p1,
			       const P2 &p2,
			       const dist_rose_tree< typename K1::argument_type > *t )
{
  return reduce_c_adapter<
    typename K1::result_type,
    typename K2::result_type,
    typename K1::argument_type,
    K1, K2, OPLUS, OTIMES, P1, P2 >
    ( k1, k2, oplus, e_oplus, otimes, e_otimes, p1, p2, t );
}

/* private, static */
template< typename C, typename D, typename A,
	  typename K1, typename K2, typename OPLUS, typename OTIMES,
	  typename P1, typename P2 >
C
rose_tree_skeletons::reduce_c_adapter( const K1 &k1,
				       const K2 &k2, 
				       const OPLUS &oplus,
				       const D &e_oplus,
				       const OTIMES &otimes,
				       const C &e_otimes,
				       const P1 &p1,
				       const P2 &p2,
				       const dist_rose_tree< A > *t )
{
  return tree_skeletons::reduce( reduce_c_k1_t< C, A, K1 >( k1, e_otimes ),
				 reduce_c_k2_t< C, A, K2, OPLUS, OTIMES >( k2, oplus, otimes ),
				 reduce_c_psi_t< C, D, A, K2 >( k2, e_oplus, e_otimes ),
				 reduce_c_phi_t< C, D, OTIMES, P1, P2 >( otimes, p1, p2 ),
				 reduce_c_phi_t< C, D, OTIMES, P1, P2 >( otimes, p1, p2 ),
				 reduce_c_G_t< C, D, OPLUS, OTIMES >( oplus, otimes ),
				 t->btree);
}

/* private::public */
template< typename C, typename A, typename K1 >
rose_tree_skeletons::reduce_c_k1_t< C, A, K1 >::
reduce_c_k1_t( const K1 &k1_,
	       const C &e_otimes_ )
  : k1( k1_ ),
    e_otimes( e_otimes_ )
{ }

/* private::public */
template< typename C, typename A, typename K1 >
C
rose_tree_skeletons::reduce_c_k1_t< C, A, K1 >::
operator()( const rose_value< A > &a ) const
{
  C ret;
  if ( a.valuetype == rose_value< A >::ORGLEAF ) {
    ret = k1( a.value );
  } else {
    ret = e_otimes;
  }
  return ret;
}

/* private::public */
template< typename C, typename A,
	  typename K2, typename OPLUS, typename OTIMES >
rose_tree_skeletons::reduce_c_k2_t< C, A, K2, OPLUS, OTIMES >::
reduce_c_k2_t( const K2 &k2_,
	       const OPLUS &oplus_,
	       const OTIMES &otimes_ )
  : k2( k2_ ),
    oplus( oplus_ ),
    otimes( otimes_ )
{
}

/* private::public */
template< typename C, typename A,
	  typename K2, typename OPLUS, typename OTIMES >
C
rose_tree_skeletons::reduce_c_k2_t< C, A, K2, OPLUS, OTIMES >::
operator()( const rose_value< A > &b,
	    const C &l,
	    const C &r ) const
{
  C ret;
  if ( b.valuetype == rose_value< A >::ORGNODE ) {
    ret = oplus( k2( b.value ), otimes( l, r ) );
  } else {
    ret = otimes( l, r );
  }
  return ret;
}

/* private::public */
template< typename C, typename D, typename A, typename K2 >
rose_tree_skeletons::reduce_c_psi_t< C, D, A, K2 >::
reduce_c_psi_t( const K2 &k2_,
		const D &e_oplus_,
		const C &e_otimes_ )
  : k2( k2_ ),
    e_oplus( e_oplus_ ),
    e_otimes( e_otimes_ )
{
}

/* private::public */
template< typename C, typename D, typename A, typename K2 >
rose_tree_skeletons::reduce_c_tup_t< C, D >
rose_tree_skeletons::reduce_c_psi_t< C, D, A, K2 >::
operator()( const rose_value< A >& b ) const
{
  rose_tree_skeletons::reduce_c_tup_t< C, D > ret;
  if ( b.valuetype == rose_value< A >::ORGNODE ) {
    ret.a = k2( b.value );
    ret.b = e_otimes;
  } else {
    ret.a = e_oplus;
    ret.b = e_otimes;
  }
  return ret;
}

/* private::public */
template< typename C, typename D,
	  typename OTIMES, typename P1, typename P2 >
rose_tree_skeletons::reduce_c_phi_t< C, D, OTIMES, P1, P2 >::
reduce_c_phi_t( const OTIMES &otimes_,
		const P1 &p1_,
		const P2 &p2_ )
  : otimes( otimes_ ),
    p1( p1_ ),
    p2( p2_ )
{
}

/* private::public */
template< typename C, typename D,
	  typename OTIMES, typename P1, typename P2 >
rose_tree_skeletons::reduce_c_tup_t< C, D >
rose_tree_skeletons::reduce_c_phi_t< C, D, OTIMES, P1, P2 >::
operator()( const rose_tree_skeletons::reduce_c_tup_t< C, D > &p,
	    const C &l,
	    const rose_tree_skeletons::reduce_c_tup_t< C, D > &o ) const
{
  rose_tree_skeletons::reduce_c_tup_t< C, D > ret;
  
  C pb_otimes_l = otimes( p.b, l );
  ret.a = p1( p.a, pb_otimes_l, o.a, o.b );
  ret.b = p2( p.a, pb_otimes_l, o.a, o.b );
  
  return ret;
}

/* private::public */
template< typename C, typename D, typename OPLUS, typename OTIMES >
rose_tree_skeletons::reduce_c_G_t< C, D, OPLUS, OTIMES >::
reduce_c_G_t( const OPLUS &oplus_,
	      const OTIMES &otimes_ )
  : oplus( oplus_ ), otimes( otimes_ )
{
}

/* private::public */
template< typename C, typename D, typename OPLUS, typename OTIMES >
C
rose_tree_skeletons::reduce_c_G_t< C, D, OPLUS, OTIMES >::
operator()( const rose_tree_skeletons::reduce_c_tup_t< C, D > &ab,
	    const C& x,
	    const C& y ) const
{
  return oplus( ab.a, otimes( ab.b, otimes( x, y ) ) );
}

/* public, static */
template< typename A >
dist_rose_tree< A > *
rose_tree_skeletons::distribute( node< rose_value< A > > *root )
{
  dist_tree< rose_value< A > >* btree
    = tree_skeletons::distribute2< rose_value< A > >( root );

  return new dist_rose_tree< A >( btree );
}


#endif /* __ROSE_TREE_SKELETONS_TPP__ */
#endif /* __ROSE_TREE_SKELETONS_H__ */
