/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file test_rose_tree_skeletons.cpp
 * @brief Test code for rose_tree_skeletons.h and rose_tree_skeletons.tpp.
 * @author Kiminori MATSUZAKI
 */

#include <assert.h>
#include <iostream>

#include "rose_tree_skeletons.h"
#include "matrix2.h"

////////////////////////////////////////////////////////////////////////
// Definitions of Function Objects

// function objects for testing map. 
class FuncDB : unary_function< double, int > {
public:
  int operator()( const double& x ) const {
    return static_cast< int >( 2.0 * x );
  }
};

class FuncSQ : unary_function< double, int > {
public:
  int operator()( const double& x ) const {
    return static_cast< int >( x * x );
  }
};

// function objects for testing zip.
class Func_zip_k1 : binary_function< double, int, char > {
public:
  char operator()( const double& a, const int& b ) const {
    return ( a == ( double )b ) ? 'T' : 'F';
  }
};

class Func_zip_k2 : binary_function< double, int, char > {
public:
  char  operator()( const double& a, const int& b ) const {
    return ( a <= ( double )b ) ? 'S' : 'L';
  }
};

// function objects and constants for testing reduce. 
// This test uses matrix multiplication for otimes, and
// scalar multiplication for oplus.
class Func_reduce_k1 : unary_function< double, Matrix2< int > > {
public:
  Matrix2< int > operator()( const double &a ) const
  {
    return Matrix2< int >( a / 2 + 1, a / 3 + 0, a / 4 + 0, a / 5 + 2 );
  }
};

class Func_reduce_k2 : unary_function< double, int > {
public:
  int operator()( const double& b ) const { return static_cast< int >( (15 - b) / 5 + 1 ); }
};

class Func_reduce_oplus : binary_function< int, Matrix2< int >, Matrix2< int > > {
public:
  Matrix2< int > operator()( const int &lhs,
			     const Matrix2< int > &rhs ) const {
    return rhs * lhs;
  }
};

class Func_reduce_otimes : binary_function< Matrix2< int >, Matrix2< int >, Matrix2< int > > {
public:
  Matrix2< int > operator()( const Matrix2< int > &lhs,
			     const Matrix2< int > &rhs ) const {
    return lhs * rhs;
  }
};

class Func_reduce_p1 {
public:
  int operator()( const int &a1,
		  const Matrix2< int > &,
		  const Matrix2< int > &,
		  const int &a2,
		  const Matrix2< int > &,
		  const Matrix2< int > & ) const {
    return a1 * a2;
  }
};

class Func_reduce_p2 {
public:
  Matrix2< int > operator()( const int &,
			     const Matrix2< int > &b1,
			     const Matrix2< int > &,
			     const int &,
			     const Matrix2< int > &b2,
			     const Matrix2< int > & ) const {
    return b1 * b2;
  }
};

class Func_reduce_p3 {
public:
  Matrix2< int > operator()( const int &,
			     const Matrix2< int > &,
			     const Matrix2< int > &c1,
			     const int &,
			     const Matrix2< int > &,
			     const Matrix2< int > &c2 ) const {
    return c2 * c1;
  }
};

const int reduce_e_oplus = 1;
const Matrix2< int > reduce_e_otimes( 1, 0, 0, 1 );

class Func_reduce_c_k1 : unary_function< double, int > {
public:
  int operator()( const double& ) const { return 2; }
};

class Func_reduce_c_k2 : unary_function< double, char > {
public:
  char operator()( const double& ) const { return '1'; };
};

class Func_reduce_c_oplus : binary_function< char, int, int > {
public:
  int operator()( const char& lhs, const int& rhs ) const { return ( lhs - '0' ) + rhs; }
};

class Func_reduce_c_otimes : binary_function< int, int, int > {
public:
  int operator()( const int &lhs, const int& rhs ) const { return lhs + rhs; }
};

class Func_reduce_c_p1 {
public:
  char operator()( const char &a1, const int&, const char &a2, const int& ) const { return static_cast< char >( ( a1 - '0' ) + ( a2 - '0' ) + '0' ); }
};

class Func_reduce_c_p2 {
public:
  int operator()( const char &, const int &b1, const char &, const int &b2 ) const { return b1 + b2; }
};

const char reduce_c_e_oplus = '0';
const int reduce_c_e_otimes = 0;

////////////////////////////////////////////////////////////////////////
// Test code

void test_rose_tree_skeletons_map( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_map.txt";

  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< int > *rtree2
    = rose_tree_skeletons::map( FuncDB( ), FuncSQ( ), rtree1 );
  assert( rtree2 );

  assert( rtree2->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of map on rose tree: " << filename_out
	      << std::endl;
  }

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_zip( )
{
  const char filename_in[]  = "dist_rose_tree_data.txt";
  const char filename_out[] = "result_test_rose_tree_skeletons_zip.txt";

  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename_in );
  assert( rtree1 );

  dist_rose_tree< int > *rtree2
    = rose_tree_skeletons::map( FuncDB( ), FuncSQ( ), rtree1 );
  assert( rtree2 );

  dist_rose_tree< char > *rtree3
    = rose_tree_skeletons::zipwith( Func_zip_k1( ),
				    Func_zip_k2( ),
				    rtree1,
				    rtree2 );
  assert( rtree3 );
  
  assert( rtree3->write_to_file( filename_out ) );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Check the result of zip on rose tree: " << filename_out
	      << std::endl;
  }

  if ( rtree3 ) delete rtree3;
  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_reduce( )
{
  const char filename[]  = "dist_rose_tree_data.txt";

  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename );
  assert( rtree1 );

  Matrix2< int > result
    = rose_tree_skeletons::reduce( Func_reduce_k1( ),
				   Func_reduce_k2( ),
				   Func_reduce_oplus( ),
				   reduce_e_oplus,
				   Func_reduce_otimes( ),
				   reduce_e_otimes,
				   Func_reduce_p1( ),
				   Func_reduce_p2( ),
				   Func_reduce_p3( ),
				   rtree1 );
  
  std::cout << "result of reduce on proc[" << skeleton::rank
	    << "] is " << result << std::endl;

  if ( rtree1 ) delete rtree1;
}

void test_rose_tree_skeletons_reduce_c( )
{
  const char filename[]  = "dist_rose_tree_data.txt";
  
  dist_rose_tree< double > *rtree1
    = dist_rose_tree< double >::read_from_file( filename );
  assert( rtree1 );

  int result
    = rose_tree_skeletons::reduce_c( Func_reduce_c_k1( ),
				     Func_reduce_c_k2( ),
				     Func_reduce_c_oplus( ),
				     reduce_c_e_oplus,
				     Func_reduce_c_otimes( ),
				     reduce_c_e_otimes,
				     Func_reduce_c_p1( ),
				     Func_reduce_c_p2( ),
				     rtree1 );
  
  std::cout << "result of reduce_c on proc[" << skeleton::rank
	    << "] is " << result << std::endl;
  
  if ( rtree1 ) delete rtree1;
}

int SketoMain( int, char ** )
{
  if ( skeleton::rank == 0 ) {
    cout << "Testing " __FILE__ << endl;
  }
  
  test_rose_tree_skeletons_map( );
  test_rose_tree_skeletons_zip( );
  test_rose_tree_skeletons_reduce( );
  test_rose_tree_skeletons_reduce_c( );
  
  if ( skeleton::rank == 0 ) {
    cout << "   ---> OK " __FILE__ << endl;
  }
  return 0;
}

////////////////////////////////////////////////////////////////////////
#if 0 // The following functions are not used (tested).
////////////////////////////////////////////////////////////////////////

template< typename A > inline A max( const A &x, const A &y )
   { return ( x > y ) ? x : y; }

struct mis_pair {
  double r, s;
  mis_pair( double r_ = 0, double s_ = 0 )
    : r( r_ ), s( s_ ) { }
  mis_pair( bool )
    : r( 0.0 ), s( 0.0 ) { }
};

struct mis_4tup {
  double a, b, c, d;
  mis_4tup( double a_ = 0, double b_ = 0, double c_ = 0, double d_ = 0 )
    : a( a_ ), b( b_ ), c( c_ ), d( d_ ) {}
  mis_4tup( bool )
    : a( 0 ), b( 0 ), c( 0 ), d( 0 ) {}
};

class Func_mis_k1 {
public:
  mis_pair operator()( const double &a ) const {
    //     mis_pair ret; ret.r = 0; ret.s = max( a, 0.0 );
    //     return ret;
    return mis_pair( 0, max( a, 0.0 ) );
  }
};

class Func_mis_k2 {
public:
  mis_4tup operator()( const double &b ) const {
    //     mis_4tup ret; ret.a = -10000.0; ret.b = b; ret.c = 0; ret.d = 0;
    //     return ret;
    return mis_4tup( -10000.0, b, 0, 0 );
  }
};

class Func_mis_oplus {
public:
  mis_pair operator()( const mis_4tup &abcd, const mis_pair &rs ) const {
    const double ar = abcd.a + rs.r;
    const double br = abcd.b + rs.r;
    const double cs = abcd.c + rs.s;
    const double ds = abcd.d + rs.s;
    //     mis_pair ret; ret.r = max( ar, cs ); ret.s = max( br, ds );
    //     return ret;
    return mis_pair( max( ar, cs ), max( br, ds ) );
  }
};

class Func_mis_otimes {
public:
  mis_pair operator()( const mis_pair &rs1, const mis_pair &rs2 ) const {
    //     mis_pair ret; ret.r = rs1.r + rs2.r; ret.s = rs1.s + rs2.s;
    //     return ret;
    return mis_pair( rs1.r + rs2.r, rs1.s + rs2.s );
  }
};

class Func_mis_p1 {
public:
  mis_4tup operator()( const mis_4tup &abcd1, const mis_pair &rs1,
		       const mis_4tup &abcd2, const mis_pair &rs2 ) const {
    //     mis_4tup ret;
    //     ret.a = max( abcd1.a + rs1.r + abcd2.a + rs2.r,
    // 		 abcd1.c + rs1.s + abcd2.b + rs2.r );
    //     ret.b = max( abcd1.b + rs1.r + abcd2.a + rs2.r,
    // 		 abcd1.d + rs1.s + abcd2.b + rs2.r );
    //     ret.c = max( abcd1.a + rs1.r + abcd2.c + rs2.s,
    // 		 abcd1.c + rs1.s + abcd2.d + rs2.s );
    //     ret.d = max( abcd1.b + rs1.r + abcd2.c + rs2.s,
    // 		 abcd1.d + rs1.s + abcd2.d + rs2.s );
    //     return ret;
    return mis_4tup( max( abcd1.a + rs1.r + abcd2.a + rs2.r,
			  abcd1.c + rs1.s + abcd2.b + rs2.r ),
		     max( abcd1.b + rs1.r + abcd2.a + rs2.r,
			  abcd1.d + rs1.s + abcd2.b + rs2.r ),
		     max( abcd1.a + rs1.r + abcd2.c + rs2.s,
			  abcd1.c + rs1.s + abcd2.d + rs2.s ),
		     max( abcd1.b + rs1.r + abcd2.c + rs2.s,
			  abcd1.d + rs1.s + abcd2.d + rs2.s ) );
  }
};

class Func_mis_p2 {
public:
  mis_pair operator()( const mis_4tup &, const mis_pair &,
		       const mis_4tup &, const mis_pair & ) const {
    //     mis_pair ret; ret.r = 0; ret.s = 0;
    //     return ret;
    return mis_pair( 0, 0 );
  }
};

// const mis_4tup mis_e_oplus = { 0, -10000, -10000, 0 };
// const mis_pair mis_e_otimes = { 0, 0 };
const mis_4tup mis_e_oplus( 0, -10000, -10000, 0 );
const mis_pair mis_e_otimes( 0, 0 );


// // functions for testing uAcc. 
// double k1( const double& a ) { return a; }
// double k2( const double& b, const double& l, const double& r ) { return b+l+r; }
// double psi( const double& b ) { return b; }
// double phiL( const double& p, const double& l, const double& r ){return p+l+r;}
// double phiR( const double& p, const double& r, const double& l ){return p+r+l;}
// double G( const double& n, const double& l, const double& r ){ return n+l+r; }

// // functions for testing dAcc.
// int plus( const int& a, const int& b ) { return a + b; }
// int dAcc_gl( const double& a ) { return ( int )( -a ); }
// int dAcc_gr( const double& a ) { return ( int )( 2.0 * a ); }

void test_Dist_Rose_Tree_Map( )
{
  const char filename[] = "dist_rose_tree_data.txt";
  CDistRoseTree< double >* rtree1
    = new CDistRoseTree< double >( filename );

  FuncDB func_db;
  FuncSQ func_sq;
  
  CDistRoseTree< int >* rtree2 = rtree1->map< int, FuncDB, FuncSQ >( func_db, func_sq );

  //  rtree2->print( );

  if ( rtree2 ) delete rtree2;
  if ( rtree1 ) delete rtree1;
}

void test_Dist_Rose_Tree_Reduce_Normal( )
{
  const char filename[] = "dist_rose_tree_data_int.txt";
  CDistRoseTree< int >* rtree
    = new CDistRoseTree< int >( filename );

  Func_Reduce_Normal_k1 func_k1;
  Func_Reduce_Normal_k2 func_k2;
  Func_Reduce_Normal_oplus func_oplus;
  Func_Reduce_Normal_otimes func_otimes;
  Func_Reduce_Normal_p1 func_p1;
  Func_Reduce_Normal_p2 func_p2;
  Func_Reduce_Normal_p3 func_p3;

  Matrix2< int > result = rtree->reduce
    < Matrix2< int >, int, Func_Reduce_Normal_k1, Func_Reduce_Normal_k2, Func_Reduce_Normal_oplus, Func_Reduce_Normal_otimes, Func_Reduce_Normal_p1, Func_Reduce_Normal_p2, Func_Reduce_Normal_p3 >
    ( func_k1,
      func_k2,
      func_oplus,
      func_otimes,
      func_p1,
      func_p2,
      func_p3,
      reduce_normal_e_oplus,
      reduce_normal_e_otimes );

  //  if ( skeleton::rank == 0 ) {
    std::cout << "Result of reduce_normal is ["
	      << skeleton::rank << "]" <<  " "
	      << result.at( 0, 0 ) << " " << result.at( 0, 1 ) << " "
	      << result.at( 1, 0 ) << " " << result.at( 1, 1 ) << std::endl;
    //  }
    
//     std::cout << "[" << skeleton::rank << "]"
// 	      << num_nodes << std::endl;
  
  if ( rtree ) delete rtree;
}
void test_Dist_Rose_Tree_Reduce( )
{
  const char filename[] = "dist_rose_tree_data.txt";
  CDistRoseTree< double >* rtree
    = new CDistRoseTree< double >( filename );

  Func_Reduce_Commutative_k1 func_k1;
  Func_Reduce_Commutative_k2 func_k2;
  Func_Reduce_Commutative_oplus func_oplus;
  Func_Reduce_Commutative_otimes func_otimes;
  Func_Reduce_Commutative_p1 func_p1;
  Func_Reduce_Commutative_p2 func_p2;
  
  int num_nodes = rtree->reduce
    /*< int, char, Func_Reduce_Commutative_k1, Func_Reduce_Commutative_k2, Func_Reduce_Commutative_oplus, Func_Reduce_Commutative_otimes, Func_Reduce_Commutative_p1, Func_Reduce_Commutative_p2 >*/
    ( func_k1, func_k2, func_oplus, func_otimes, func_p1, func_p2, reduce_e_oplus, reduce_e_otimes );

  std::cout << "[" << skeleton::rank << "]"
	    << num_nodes << std::endl;
//   if ( skeleton::rank == 0 ) {
//     // assert( num_nodes == 13 );
//   }

  if ( rtree ) delete rtree;
}

void test_Dist_Rose_Tree_mis( )
{
  const char filename[] = "dist_rose_tree_data.txt";
  CDistRoseTree< double >* rtree
    = new CDistRoseTree< double >( filename );

  Func_mis_k1 func_k1;
  Func_mis_k2 func_k2;
  Func_mis_oplus func_oplus;
  Func_mis_otimes func_otimes;
  Func_mis_p1 func_p1;
  Func_mis_p2 func_p2;
  
  mis_pair mis = rtree->reduce
    /*< mis_pair, mis_4tup, Func_mis_k1, Func_mis_k2, Func_mis_oplus, Func_mis_otimes, Func_mis_p1, Func_mis_p2 >*/
    ( func_k1, func_k2, func_oplus, func_otimes, func_p1, func_p2, mis_e_oplus, mis_e_otimes );

  std::cout << "[" << skeleton::rank << "] mis = "
	    << mis.s << std::endl;
//   if ( skeleton::rank == 0 ) {
//     // assert( num_nodes == 13 );
//   }

  if ( rtree ) delete rtree;
}

void test_Dist_Rose_Tree_Zip( )
{
  const char filename_double[] = "dist_rose_tree_data.txt";
  const char filename_int[] = "dist_rose_tree_data_int.txt";
  const char filename_out[] = "dist_rose_tree_data_out_zipwith.txt";
  
  CDistRoseTree< double >* tree_double
    = new CDistRoseTree< double >( filename_double );
  CDistRoseTree< int >* tree_int
    = new CDistRoseTree< int >( filename_int );

  Func_zip_k1 func_k1;
  Func_zip_k2 func_k2;
  
  CDistRoseTree< char >* tree_char
    = tree_double->zipwith< char, int, Func_zip_k1, Func_zip_k2 >( func_k1, func_k2, tree_int );

  tree_char->writeToFile( filename_out );
  if ( skeleton::rank == 0 ) {
    std::cerr << "Output tree of zip skeleton is written into ""dist_rose_tree_data_out_zipwith.txt""" << std::endl;
  }

  delete tree_char;
  delete tree_int;
  delete tree_double;
}

// void test_Dist_Rose_TreeuAcc( )
// {
//   const char filename[] = "testdata2.txt";
//   CDistTree< double >* tree
//     = CDistTree< double >::readFromFile( filename );

//   CDistTree< double >* sumtree = tree->uAcc( k1, k2, psi, phiL, phiR, G );

//   //  sumtree->print( );

//   if ( sumtree ) delete sumtree;
//   if ( tree ) delete tree;
// }

// void test_Dist_Rose_TreedAcc( )
// {
//   const char filename[] = "testdata2.txt";
//   CDistTree< double >* tree
//     = CDistTree< double >::readFromFile( filename );

//   CDistTree< int >* newtree = tree->dAcc( plus, 0, dAcc_gl, dAcc_gr, 100 );

//   // newtree->print( );

//   delete newtree;
//   delete tree;
// }


////////////////////////////////////////////////////////////////////////
#endif 
////////////////////////////////////////////////////////////////////////
