/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file tree_skeletons.h
 * @brief The Header of class @a tree_skeletons, consists of tree skeletons
 *        on binary trees.
 * @author Kiminori MATSUZAKI
 */

#ifndef __TREE_SKELETONS_H__
#define __TREE_SKELETONS_H__

#include "node.h"
#include "dist_tree.h"
#include "../util/compose.h"

/**
 * @brief Collection of skeletons on binary trees.
 */
class tree_skeletons
{
  ////////////////////////////////////////////////////////////////////////
  // Data parallel Skeletons

  /**
   * @brief Map skeleton with Function Objects
   *
   * This skeleton applies function @a k1 to every leaf and function @a k2
   * to every internal node.
   *
   * Definition (in Haskell):
   * @code
   *   map :: (a -> c) -> (a -> c) -> BTree a a -> BTree c c
   *   map k1 k2 (Leaf a)     = Leaf (k1 a)
   *   map k1 k2 (Node b l r) = Node (k2 b) (map k1 k2 l) (map k1 k2 r)
   * @endcode
   *
   * To call this skeleton, you need to specify the types below.
   *  - K1 = type of the function for each leaf (option).
   *  - K2 = type of the function for each internal node (option).
   *
   * Sample: \n
   * < Specification (in Haskell) >
   * @code 
   *   t2 = map k1 k2 t1
   * @endcode
   * < C++ code >
   * @code
   *   t2 = tree_skeletons::map( k1, k2, t1 );
   * @endcode
   *
   * @param k1: Function object applied to every leaf.
   *            This function needs to inherit @c unary_function
   *            with suitable types.
   * @param k2: Function object applied to every internal node.
   *            This function needs to inherit @c unary function
   *            with suitable types. 
   * @param t : A distributed tree for input. 
   */
public:
  template< typename K1, typename K2 >
  static dist_tree< typename K1::result_type > *
  map( const K1 &k1,
       const K2 &k2,
       dist_tree< typename K1::argument_type > *t );

private:
  template< typename C, typename A, typename K1, typename K2 >
  static dist_tree< C > *map_adapter( const K1 &k1,
				      const K2 &k2,
				      dist_tree< A > *t );

private:
  template< typename C, typename A, typename K1, typename K2 >
  static node< C >* map_rec( const K1 &k1,
			     const K2 &k2,
			     node< A > *root );

  /**
   * @brief Zipwith skeleton with function object
   *
   * This skeleton zips up two trees of the same shape into a tree
   * by applying @a k1 for every pair of leaves and @a k2 for every pair
   * of internal nodes.
   *
   * Definition (in Haskell):
   * @code
   *   zipwith :: (a -> c -> e) -> (a -> c -> e) -> BTree a a -> BTree c c -> BTree e e
   *   zipwith k1 k2 (Leaf a) (Leaf a')           = Leaf (k1 a a')
   *   zipwith k1 k2 (Node b l r) (Node b' l' r') = Node (k2 b b') (zipwith k1 k2 l l')
   *                                                               (zipwith k1 k2 r r')
   * @endcode
   *   
   * To call this skeleton, you need to specify the types below.
   *  - K1 = type of function object k1 (option),
   *  - K2 = type of function object k2 (option).
   *
   * Sample: \n
   * < Specification (in Haskell) >
   * @code
   *   t = zipwith k1 k2 at bt
   * @endcode
   * < C++ code >
   * @code
   *   t = tree_skeletons::zipwith( k1, k2, at, bt );
   * @endcode
   *
   * @param k1: Function object applied to every pair of leaves.
   *            This function needs to inherit @c binary_function
   *            with suitable types.
   * @param k2: Function object applied to every pair of internal nodes.
   *            This function needs to inherit @c binary_function
   *            with suitable types.
   * @param at: The first tree to zip up. 
   * @param bt: The second tree to zip up. 
   */
public:
  template< typename K1, typename K2 >
  static dist_tree< typename K1::result_type > *
  zipwith( const K1 &k1, 
	   const K2 &k2,
	   const dist_tree< typename K1::first_argument_type >* at, 
	   const dist_tree< typename K1::second_argument_type >* bt );

private:
  template< typename E, typename A, typename C, typename K1, typename K2 >
  static dist_tree< E >* zipwith_adapter( const K1 &k1, 
					  const K2 &k2,
					  const dist_tree< A >* at, 
					  const dist_tree< C >* bt );

private:
  template< typename E, typename A, typename C, typename K1, typename K2 >
  static node< E >* zipwith_rec( const K1 &k1, 
				 const K2 &k2, 
				 const node< A >* root_a,
				 const node< C >* root_b );

  /**
   * @brief Reduce skeleton with Function Objects
   *
   * This skeleton collapses the tree in a bottom-up manner by applying
   * k1 for every leaf and k2 for every internal node and its children.
   *
   * Definition (in Haskell):
   * @code
   *   reduce :: (a -> c) -> (a -> c -> c -> c) -> BTree a a -> c
   *   reduce k1 k2 (Leaf a)     = k1 a
   *   reduce k1 k2 (Node b l r) = k2 b (reduce k1 k2 l) (reduce k1 k2 r)
   * @endcode
   *
   * To guarantee the correctness of the parallel implementation,
   * the function @c k2 must satisfy the conditions of the tree contraction
   * algorithm, i.e., it is needed to derive functions 
   * @a psi, @a phiL, @a phiR, and @a G as in the conditions below. 
   *
   * Conditions (in Haskell):
   * @code
   *   psi :: c -> d
   *   phiL :: d -> c -> d -> d
   *   phiR :: d -> c -> d -> d
   *   phiG :: d -> c -> c -> c
   *   k2 b l' r' = G (psi b) l' r'
   *   k2 b l' (k2 rb rl' rr') = G (phiL (psi b) l' (psi rb)) rl' rr'
   *   k2 b (k2 lb ll' lr') r' = G (phiR (psi b) r' (psi lb)) ll' lr'
   * @endcode
   *
   * To call this skeleton, you need to specify the types below.
   *  - K1   = type of function object k1 (option), 
   *  - K2   = type of function object k2 (option), 
   *  - PSI  = type of function object psi (option), 
   *  - PHIL = type of function object phiL (option),
   *  - PHIR = type of function object phiR (option),
   *  - GG   = type of function object G (option).
   *
   * Sample: \n
   * < specification (two functions only) >
   * @code
   *   v = reduce k1 k2 t
   * @endcode
   * < C++ code (with all functions in the conditions) >
   * @code
   *   v = tree_skeletons::reduce( k1, k2, psi, phiL, phiR, G, t );
   * @endcode
   *
   * @param k1: Function object applied to every leaf.
   *            This function needs to inherit
   *            @code unary_function< A, C > @endcode. 
   * @param k2: Function object applied to every internal node.
   *            This function needs to inherit
   *            @code ternary_function< A, C, C, D > @endcode. 
   * @param psi: Function object for Initialization of the tree contraction.
   *            This function needs to inherit
   *            @code unary_function< A, D > @endcode. 
   * @param phiL: Function object for ContractL operation.
   *            This function needs to inherit
   *            @code ternary_function< D, C, D, D > @endcode. 
   * @param phiR: Function object for ContractR operation.
   *            This function needs to inherit
   *            @code ternary_function< D, C, D, D > @endcode. 
   * @param G: Function object for parametrized function for tree contraction. 
   *            This function needs to inherit
   *            @code ternary_function< D, C, C, C > @endcode. 
   * @param t: The distributed tree for input
   *
   */
public:
  template< typename K1, typename K2,
	    typename PSI, typename PHIL, typename PHIR, typename GG >
  static typename K1::result_type
  reduce( const K1 &k1, 
	  const K2 &k2, 
	  const PSI &psi,
	  const PHIL &phiL,
	  const PHIR &phiR,
	  const GG &G,
	  const dist_tree< typename K1::argument_type >* t );

private:
  template< typename C, typename D, typename A, typename K1, typename K2,
	    typename PSI, typename PHIL, typename PHIR, typename GG >
  static C reduce_adapter( const K1 &k1, 
			   const K2 &k2, 
			   const PSI &psi,
			   const PHIL &phiL,
			   const PHIR &phiR,
			   const GG &G,
			   const dist_tree< A >* t );

  /**
   * @brief Structure to represent the result of local reduce and local uAcc, 
   *        used in the reduce and uAcc skeletons.
   * @todo To avoid compile error caused by union, this structure has
   *       both valC and valD. This is not good in the sense of memory size. 
   */
private:
  template< typename C, typename D >
  struct reduce_result_t {
    enum {
      TYPEC,
      TYPED,
    } resulttype;
    C valC;
    D valD;
  };

private:
  template< typename C, typename D, typename A, typename K1, typename K2,
	    typename PSI, typename PHIL, typename PHIR >
  static tree_skeletons::reduce_result_t< C, D >
  reduce_local_rec( const K1 &k1,
		    const K2 &k2,
		    const PSI &psi,
		    const PHIL &phiL,
		    const PHIR &phiR, 
		    const node< A >* root );

private:
  template< typename C, typename D, typename A, typename GG >
  static C reduce_global( const GG &G,
			  const tree_skeletons::reduce_result_t< C, D > &local_result,
			  const dist_tree< A > *t );
  
  /**
   * @brief Upwards accumulate skeleton with Function Objects
   *
   * This skeleton computes on the tree in a bottom-up manner like @a reduce, 
   * and the difference is that this skeleton returns a tree of the same
   * shape as input tree by leaving the intermediate result on every node.
   *
   * The meaning of functions and the conditions of additional functions for
   * parallel implementation are the same as @a reduce skeleton.
   *
   * Definition (in Haskell):
   * @code
   *   uAcc :: (a -> c) -> (a -> c -> c -> c) -> BTree a a -> BTree c c
   *   uAcc k1 k2 (Leaf a)     = Leaf (k1 a)
   *   uAcc k1 k2 (Node b l r) = let l' = uAcc k1 k2 l
   *                                 r' = uAcc k1 k2 r
   *                             in  Node (k2 b (root l') (root r')) l' r'
   * @endcode
   *
   * Conditions: \n
   * For conditions of @a psi, @a phiL, @a phiR, @a G, see conditions in @a
   * reduce. 
   *
   * To call this skeleton, you need to specify the types below. 
   *  - K1   = type of function object k1 (option), 
   *  - K2   = type of function object k2 (option), 
   *  - PSI  = type of function object psi (option), 
   *  - PHIL = type of function object phiL (option),
   *  - PHIR = type of function object phiR (option),
   *  - GG   = type of function object G (option).
   * 
   * Sample: \n
   * < specification (two functions only) >
   * @code
   *   t2 = uAcc k1 k2 t1
   * @endcode
   * < C++ code (with all functions in the conditions) >
   * @code
   *   t2 = tree_skeletons::uAcc( k1, k2, psi, phiL, phiR, G, t1 );
   * @endcode
   * 
   * @param k1: Function object applied to every leaf.
   *            This function needs to inherit
   *            @code unary_function< A, C > @endcode. 
   * @param k2: Function object applied to every internal node.
   *            This function needs to inherit
   *            @code ternary_function< A, C, C, D > @endcode. 
   * @param psi: Function object for Initialization of the tree contraction.
   *            This function needs to inherit
   *            @code unary_function< A, D > @endcode. 
   * @param phiL: Function object for ContractL operation.
   *            This function needs to inherit
   *            @code ternary_function< D, C, D, D > @endcode. 
   * @param phiR: Function object for ContractR operation.
   *            This function needs to inherit
   *            @code ternary_function< D, C, D, D > @endcode. 
   * @param G: Function object for parametrized function for tree contraction. 
   *            This function needs to inherit
   *            @code ternary_function< D, C, C, C > @endcode. 
   * @param t: The distributed tree for input
   */
public:
  template< typename K1, typename K2,
	    typename PSI, typename PHIL, typename PHIR, typename GG >
  static dist_tree< typename K1::result_type > *
  uAcc( const K1 &k1,
	const K2 &k2,
	const PSI &psi,
	const PHIL &phiL,
	const PHIR &phiR,
	const GG &G,
	const dist_tree< typename K1::argument_type > *t );

private:
  template< typename C, typename D, typename A, typename K1, typename K2,
	    typename PSI, typename PHIL, typename PHIR, typename GG >
  static dist_tree< C >* uAcc_adapter( const K1 &k1,
				       const K2 &k2,
				       const PSI &psi,
				       const PHIL &phiL,
				       const PHIR &phiR,
				       const GG &G,
				       const dist_tree< A > *t );

private:
  template< typename C, typename D, typename A, typename GG >
  static void uAcc_global_uAcc( const GG &G, 
				const tree_skeletons::reduce_result_t< C, D > &local_result,
				C *vall,
				C *valr,
				const dist_tree< A > *t );

private:
  template< typename C, typename D, typename A, typename K1, typename K2 >
  static node< C >* uAcc_local_uAcc_rec( const K1 &k1,
					 const K2 &k2, 
					 const C& left_result,
					 const C& right_result,
					 const node< A > *root );

  /**
   * @brief Downwards accumulate skeleton with Function Objects
   *
   * This skeleton computes in a top-down manner updating an accumulative
   * parameter (initially @a c) with an associative operator (@a oplus) and
   * functions (@a gl for left subtree and @a gr for right subtree).
   * 
   * For parallel implementation, the binary operator @c oplus must be
   * associative with the unit (@a e_oplus). 
   *
   * Definition (in Haskell):
   * @code
   *   dAcc :: (c -> c -> c) <assoc> -> c <unit> -> (a -> c) -> (a -> c)
   *               -> c -> BTree a a -> BTree c c
   *   dAcc (oplus) e_oplus gl gr c (Leaf a)     = Leaf c
   *   dAcc (oplus) e_oplus gl gr c (Node b l r)
   *       = Node c (dAcc (oplus) e_oplus gl gr (c oplus gl b) l)
   *                (dAcc (oplus) e_oplus gl gr (c oplus gr b) r)
   * @endcode
   *
   * To call this skeleton, you don't need to specify the template types.
   *  - OPLUS = type of function object for oplus (option),
   *  - GL    = type of function object for gl (option),
   *  - GR    = type of function object for gr (option).
   *
   * Sample: \n
   * < Specification (in Haskell) >
   * @code
   *   t2 = dAcc (oplus) e_oplus gl gr c t1
   * @endcode
   * < C++ code >
   * @code
   *   t2 = tree_skeletons::dAcc( oplus, e_oplus, gl, gr, c, t1 );
   * @endcode
   * 
   * @param oplus: A function object for binary associative operator.
   *            This function needs to inherit
   *            @code binary_function< C, C, C > @endcode
   * @param e_oplus: The unit of operator of @a oplus. 
   * @param gl: Function object for function used in updating accumulative
   *            parameter for left subtree.
   *            This function needs to inherit
   *            @code unary_function< A, C > @endcode
   * @param gr: Function object for function used in updating accumulative
   *            parameter for right subtree.
   *            This function needs to inherit
   *            @code unary_function< A, C > @endcode
   * @param c: An initial value for accumulation.
   * @param t: A distributed tree for input. 
   */
public:
  template< typename OPLUS, typename GL, typename GR >
  static dist_tree< typename OPLUS::result_type > *
  dAcc( const OPLUS &oplus,
	const typename OPLUS::result_type &e_oplus,
	const GL &gl,
	const GR &gr,
	const typename OPLUS::result_type &c,
	const dist_tree< typename GL::argument_type >* t );

private:
  template< typename C, typename A, typename OPLUS, typename GL, typename GR >
  static dist_tree< C >* dAcc_adapter( const OPLUS &oplus,
				       const C &e_oplus,
				       const GL &gl,
				       const GR &gr,
				       const C &c,
				       const dist_tree< A >* t );

private:
  template< typename C, typename A, typename OPLUS, typename GL, typename GR >
  static bool dAcc_local_path_rec( const OPLUS &oplus,
				   const GL &gl,
				   const GR &gr,
				   const node< A >* root,
				   C *c_to_left,
				   C *c_to_right );

private:
  template< typename C, typename A, typename OPLUS >
  static C dAcc_global_dAcc( const OPLUS &oplus,
			     const C &c,
			     const C &c_to_left,
			     const C &c_to_right,
			     const dist_tree< A >* t );

private:
  template< typename C, typename A, typename OPLUS, typename GL, typename GR >
  static node< C >* dAcc_local_dAcc_rec( const OPLUS &oplus,
					 const GL &gl,
					 const GR &gr, 
					 const C &c,
					 const node< A >* root );

  ////////////////////////////////////////////////////////////////////////
  // Communication Skeletons

  /**
   * @brief Distribute skeleton.
   *
   * This skeleton distributes an input tree to processors.
   *
   * The input tree must be a perfect binary tree (i.e., all the
   * internal node has just two children).
   *
   * @param tree: a tree to distribute. 
   */
public:
  template< typename A >
  static dist_tree< A >* distribute( const node< A >* tree );

  /**
   * @brief Distribute skeleton.
   *
   * This skeleton distributes an input tree to processors.
   *
   * The input tree must be a perfect binary tree (i.e., all the
   * internal node has just two children).
   *
   * @param tree: a tree to distribute. 
   */
public:
  template< typename A >
  static dist_tree< A >* distribute2( const node< A >* tree );

private:
  template< typename A >
  struct size_and_critical {
    int left_size;
    int right_size;
    const node< A >* critical_node;
  };

private:
  template< typename A >
  static bool compute_size_and_get_global_tree( const node< A >* tree,
						int nprocs,
						node< size_and_critical< A > >** global_root,
						int *np,
						int *size );

private:
  template< typename A >
  static void get_global_tree_rec( node< size_and_critical< A > >** global_tree,
				   int *size_root,
				   const node< A >* tree,
				   int m );
  
private:
  template< typename A >
  static void serialize_global_structure( const node< A >* tree,
					  const node< size_and_critical< A > >* global_root,
					  int size, 
					  int *flags,
					  int *localsizes,
					  const node< A >** subroots,
					  const node< A >** subcrits );

private:
  template< typename A >
  static void serialize_global_structure_rec( const node< A >* tree,
					      const node< size_and_critical< A > >* global_root,
					      int size,
					      int *flags,
					      int *localsizes,
					      const node< A >** subroots,
					      const node< A >** subcrits,
					      int *index );
					      
private:
  template< typename A >
  static void get_local_tree( const node< A >* root,
			      const node< A >* crit_node,
			      int *flags,
			      A *values );

private:
  template< typename A >
  static void get_local_tree_rec( const node< A >* root,
				  const node< A >* crit_node,
				  int *flags,
				  A *values,
				  int *index );

  /**
   * @brief Gather skeleton.
   *
   * This skeleton gathers nodes from a distributed tree into a tree.
   *
   * @param tree: a distributed tree to be gathered.
   */
public:
  template< typename A >
  static node< A >* gather( const dist_tree< A >* tree );
  
private:
  template< typename A >
  static void gather_rec( node< A >* root,
			  const node< int >* global_root,
			  int *proc );

private:
  template< typename A >
  static node< A >* gather_find_cutnode( node< A >* root );

private:
  template< typename A >
  static node< A >* gather_create_subtree( const int* flags,
					   const A* values );
  
private:
  template< typename A >
  static node< A >* gather_copy_tree( const node< A >* root );

  /**
   * @brief Shift each value to its children
   *
   * Shiftp skeleton shift each value to its children, that is, for each
   * node this skeleton substitute the value for children's value.
   * The value of root node is specified by the input.
   * The values of leaves will be dropped.
   *
   * Definition (in Haskell):
   * @code
   *   shiftp :: a -> BTree a a -> BTree a a
   *   shiftp c (Leaf a)     = Leaf c
   *   shiftp c (Node b l r) = Node c (shitp b l) (shift b r)
   * @endcode
   *
   * @param rootval: The value which will be assigned to the root node.
   *                 This value is used only on the process 0.
   * @param tree: The distributed tree for input.
   */
public:
  template< typename A >
  static dist_tree< A >* shiftp( const A& rootval,
				 const dist_tree< A >* tree );

private:
  template< typename A >
  static node< A > *shiftp_rec( const node< A >* root,
				const A& root_val, 
				A *cut_node_val );

  /**
   * @brief Shift each left-child's value to its parent
   *
   * Shiftl skeleton shift each left-child's value to its parent, that is,
   * for each node this skeleton substitute the left-child's value for
   * the node's value.
   * The values of leaves are specified by the input. 
   * The value of root node will be dropped.
   *
   * Definition (in Haskell):
   * @code
   *   shiftl :: a -> BTree a a -> BTree a a
   *   shiftl a (Leaf b)     = Leaf a
   *   shiftl a (Node b l r) = Node (root l) (shiftl a l) (shiftl a r)
   * @endcode
   *
   * @param leafval: The value which will be assigned to the leaves' node.
   *                 This value is used on all processes. 
   * @param tree: The distributed tree for input.
   *
   * @attention: Note that @c leafval should be copied on all processes. 
   */
public:
  template< typename A >
  static dist_tree< A >* shiftl( const A& leafval,
				 const dist_tree< A >* tree );

private:
  template< typename A >
  static node< A >* shiftl_rec( const node< A > *root,
				const A& leafval,
				const A& left_child_root_val );
  
  /**
   * @brief Shift each right-child's value to its parent
   *
   * Shiftr skeleton shift each right-child's value to its parent, that is,
   * for each node this skeleton substitute the right-child's value for
   * the node's value.
   * The values of leaves are specified by the input. 
   * The value of root node will be dropped.
   *
   * Definition (in Haskell):
   * @code
   *   shiftr :: a -> BTree a a -> BTree a a
   *   shiftr a (Leaf b)     = Leaf a
   *   shiftr a (Node b l r) = Node (root r) (shiftr a l) (shiftr a r)
   * @endcode
   *
   * @param leafval: The value which will be assigned to the leaves' node.
   *                 This value is used on all processes. 
   * @param tree: The distributed tree for input.
   *
   * @attention: Note that @c leafval should be copied on all processes. 
   */
public:
  template< typename A >
  static dist_tree< A >* shiftr( const A& leafval,
				 const dist_tree< A >* tree );

private:
  template< typename A >
  static node< A >* shiftr_rec( const node< A > *root,
				const A& leafval,
				const A& right_child_root_val );

};

#include "tree_skeletons.tpp"

#endif /* __TREE_SKELETONS_H__ */

