/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file tree_util.h
 * @brief The Header of namespace @a tree_util. 
 * @author Kiminori MATSUZAKI
 */

#ifndef __TREE_UTIL_H__
#define __TREE_UTIL_H__

#include "node.h"

/**
 * @brief Utility for manipulating binary trees in sequential programs.
 */
namespace tree_util {

  /**
   * @brief Read a non-distributed tree from a file
   *
   * @param filename : The file to read from
   * @return If this function succeed to read from file then it returns
   *         the pointer to the root node, and if it fails, returns @c NULL.
   */
  template< typename A >
  node< A >*
  read_tree_from_file( const char* filename );

  /**
   * @brief Construct a non-distributed tree from serialized arrays.
   *
   * @param flags : Pointer to the array of flags.
   *                ( 0 == Internal node, 1 == Leaf )
   * @param values : Pointer to the array of values.
   * @return The pointer to the root node. 
   */
  template< typename A >
  node< A >*
  construct_tree_from_array( const int* flags,
			     const A* values );
  
  /**
   * @brief Free all the node recursively.
   *
   * @param root : Pointer to the root node to free. 
   */
  template< typename A >
  void
  free_all_tree( node< A >* root );

  /**
   * @brief Check the input tree is a perfect binary tree.
   *
   * @param tree : A rootnode of a tree to check (in)
   *
   * @retval true : @a tree is a perfect binary tree.
   * @retval false : @a tree is not a perfect binary tree.
   */
  template< typename A >
  bool
  check_perfect_binary_tree( const node< A >* tree );

  /**
   * @brief Distribute and write the distributed tree to file.
   *
   * @param tree : A tree to distribute (in)
   * @param nprocs : Number of processors (in)
   * @param filename : File to write to. (in)
   * @retval true : Successfully distributed and written to file.
   * @retval false : Something wrong occurred.
   */
  template< typename A >
  bool
  dist_and_write_to_file( const node< A >* tree,
			  const int nprocs,
			  const char *filename );

  /**
   * @brief Write a tree to file as it is.
   *
   * @param tree : A tree to write to file
   * @param filename : Filename to write to.
   */
  template< typename A >
  bool
  write_to_file( const node< A >* tree,
		 const char *filename );

  /**
   * @brief Compute the size of tree.
   *
   * @param tree : A tree whose size you want to know
   */
  template< typename A >
  int
  get_size_rec( const node< A >* tree );
  
   
}

#include "tree_util.tpp"

#endif /* __TREE_UTIL_H__ */
