/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @mainpage
 *
 * SkeTo (Skeletons in Tokyo) library provides many parallel skeletons
 * for lists, matrices, and trees. A parallel skeleton is a
 * computation pattern which is often used in parallel programming,
 * and is implemented in parallel efficiently. These parallel
 * skeletons gives a more straightforward way to compose parallel
 * programs.
 *
 * Quick links.
 *
 * - List skeletons:
 *  - @a dist_list
 *  - @a list_skeletons
 * - Matrix skeletons:
 *  - @a dist_matrix
 *  - @a matrix_skeletons
 * - Tree skeletons:
 *  - @a dist_tree
 *  - @a dist_rose_tree
 *  - @a tree_skeletons
 *  - @a rose_tree_skeletons
 * - Utilities
 *  - @a skeleton.h
 *  - @a functions.h
 *
 * Copyright (c) 2005, SkeTo Project.  All rights reserved.
 */
 
/**
 * @file skeleton.h
 * @brief The skeleton common file.
 * @author Yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 */

#ifndef __SKELETON_H__
#define __SKELETON_H__

#include <string>
#include <cstdio>
#include <iostream>
#include <cstdarg>
#include <vector>
#include <utility>
#include <mpi.h>
#include "exception.h"

namespace skeleton {
  int rank;  // MPI rank
  int procs; // Number of processors

  inline void init(int &argc, char **&argv)
  {
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &procs);
  }

  inline void finalize()
  {
    MPI_Finalize();
  }

  inline bool is_master( )
  {
    return ( rank == 0 );
  }

  inline void mprintf( const char* format, ... )
  {
    if ( !is_master( ) ) return;
    va_list args;
    va_start( args, format );
    vprintf( format, args );
    va_end( args );
  }
    
  template< class A >
  inline void safe_delete( A *a )
  {
    if ( a ) delete a;
  }

  template< class A >
  inline void safe_delete_array( A *a )
  {
    if ( a ) delete[] a;
  }

  namespace time
  {
    std::vector< std::pair< double, std::string > > time_table;
    inline void initialize( )
    {
      if ( is_master( ) ) time_table.clear( );
      MPI_Barrier( MPI_COMM_WORLD );
      if ( is_master( ) ) time_table.push_back( std::pair< double, std::string >( MPI_Wtime( ), std::string( "" ) ) );
    }

    void set_time( const char* name, bool barrier = true )
    {
      if ( barrier ) MPI_Barrier( MPI_COMM_WORLD );
      if ( is_master( ) ) time_table.push_back( std::pair< double, std::string >( MPI_Wtime( ), std::string( name ) ) );
    }

    void print( FILE *fp = stdout )
    {
      if ( !is_master( ) ) return;
      fprintf(fp, "%10s %10s %s\n", "total", "difference", "comment");
      for (unsigned int i = 1; i < time_table.size(); i++) {
	fprintf(fp, "%10.6f %10.6f %s\n",
		time_table[i].first - time_table[0].first,
		time_table[i].first - time_table[i-1].first,
		time_table[i].second.c_str());
      }
    }
  }
};

// metaclass optimizer Main;
// class Main{
// public:
//   static int main(int argc, char **argv);
// };
int SketoMain( int argc, char **argv );

int main(int argc, char **argv)
{
  skeleton::init(argc, argv);

  int retval = 0;
  try {
    retval = SketoMain( argc, argv );
  }
  catch (skeleton::bad_size e) {
    std::cerr << "bad_size exception: (" << e.size_a << ", " << e.size_b << ")" << std::endl;
    retval = -1;
  }
  catch (skeleton::error e) {
    std::cerr << "Error: " << e.msg << std::endl;
    retval = e.errno;
  }     

  skeleton::finalize();
  return retval;
}

#endif // __SKELETON_H__
