/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
//
// @author: Hideo NISHIMURA (nish@ipl.cs.uec.ac.jp)
// 
// This program is using C++ parallel skeleton on distributed list.

//-----------------------------//
//         How to use          //
//-----------------------------//
//
// This is One-way ANOVA program which outputs anova table.
// ( If you want to know about ANOVA, ask it of GOOGLE. )
// 
//  Argument of console command :  %s size number
//     "size" is the number of elemnts of each groups
//     "number" is the number of groups
//

//-----------------------------//
//           Compile           //
//-----------------------------//
//
//  put C++ parallel library and "anova.cpp" in the same directory
//
//     $ mpiCC anova.cpp
//
//  to execute
//     $ mpirun -np "N" a.out "size" "number"

#include "list_skeletons.h"

#include <cstdlib>

//---------------------------------//
//  functions using with Skeleton  //
//---------------------------------//

//--- function that generates proper data
struct Gen : public unary_function< int, double > {
  double operator() ( int index ) const {
    //--- return [0,99] data generated at random
    return static_cast< double >( ( rand( ) % 100 ) );

    //--- return index number (for debug. with this data, var might be 0)
    //return static_cast< double >( index );
  }
} generator;



//--- dyadic ADD operator function
template < typename A >
struct Add : public binary_function< A, A, A > {
  A operator() ( A x, A y ) const {
    return x + y;
  }
};


//--- function that squares an argument
template < typename A >
struct Square : public unary_function< A, A > {
  A operator() ( A x ) const {
    return x * x;
  }
};


//--- function that subtracts member variable ( given when constructed ) 
//    from an argument
struct Sub_ave : public unary_function< double, double > {
  double ave_;

  // constructer with an argument to give a variable
  Sub_ave( double a ) { ave_ = a; }

  // main content of function
  double operator() ( double x ) const {
    return x - ave_;
  }
};




//------------------------//
//     main function      //
//------------------------//
int SketoMain(int argc, char **argv)
{
 
  if ( argc < 3 ) {
    skeleton::mprintf( "usage: %s size number\n", argv[ 0 ] );
	return 0;
  }

  int size = atoi( argv[ 1 ] );   // "size" is the number of elements of each group
  int number = atoi( argv[ 2 ] );  // "number" is the number of groups

  skeleton::time::initialize( ); // for time ekeeping

  const double zero = 0.0; // constant variable which means zero

  // for each dist_list "*a[ i ]", generates datas
  dist_list< double > **a = NULL;
  a = new dist_list< double >*[ number ];
  for ( int i = 0; i < number; i++) 
	  a[i] = new dist_list< double>( generator, size );

  double ave = 0.0;

  // store average of each groups in "ave_a"
  double *ave_a;
  ave_a = new double[ number ];

  Add< double > add; // make an instance of an add operator object

  for ( int i = 0; i < number; i++ ) {
    ave_a[ i ] = list_skeletons::reduce( add, zero, a[ i ] ) / size;
    ave += ave_a[ i ];  // store sum of averages of each groups in ave
  }

  // divides by number
  // to store average of all elements in "ave"
  ave /= number;


  double sm = 0.0;
  for ( int i = 0; i < number; i++ ) {
    sm += ( ave_a[ i ] - ave ) * ( ave_a[ i ] - ave );
  }

  delete [] ave_a;
  
  // multiple by size 
  // to store between group sum of squares in "sm"
  sm *= size;

  Sub_ave sub_ave( ave ); // make an instance of sub "ave" operator object
  Square< double > square; // make an instance of square operator object

  // below, get variations of each group
  for ( int i = 0; i < number; i++ ) {
    list_skeletons::map_ow( sub_ave, a[ i ] );
    list_skeletons::map_ow( square, a[ i ] );
  }
  
  // store sum of square of each group in "ss" temporary
  double st = 0.0;
  for ( int i = 0; i < number; i++ ) {
    st += list_skeletons::reduce( add, zero, a[ i ] );
  }

  for ( int i = 0; i < number; i++ ) delete a[i];
  delete [] a;

  // store within group sum of squares in "se"
  double se = st - sm;

  // store
  //   total degrees of freedom in "dft"
  //   between group degrees of freedom in "dfm"
  //   within group degrees of freedom in "dfe"
  int dft = number * size - 1;
  int dfm = number - 1;
  int dfe = number * (size - 1);

  // store
  //   between group mean square in "vm"
  //   within group mean square in "ve"
  double vm = sm / dfm;
  double ve = se / dfe;

  skeleton::time::set_time( "calculate" );  // for time ekeeping

  // output anova table
  skeleton::mprintf("factor            SS         DF             MS              F\n");
  skeleton::mprintf("source   %.5e    %7d    %.5e    %3.5e\n", sm, dfm, vm, vm/ve);
  skeleton::mprintf("error    %.5e    %7d    %.5e\n", se, dfe, ve);
  skeleton::mprintf("total    %.5e    %7d\n", st, dft);

  skeleton::time::print( ); // show the result of time ekeeping

  return 0;
}
