/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file bracket_matching.cpp
 * @brief This program check bracket matching.
 * @author yoshiki AKASHI (yoshiki@ipl.cs.uec.ac.jp)
 * @attention: If brackets does not match, this program may freeze.
 */

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <stack>
#include "list_skeletons.h"
using namespace std;

const int SIZE = 10000;
const int KIND = 4;
const int DEPTH = 10;
const int MAXDEPTH = 10;
const char openbr[] = "({[<";
const char closebr[] = ")}]>";

//---------------------------------------------------------------------
// Stack
class Stack{
public:
  char in;
  char out;
  char data[MAXDEPTH];
  Stack():in(0), out(0){}
  Stack(const Stack &s):in(s.in), out(s.out){
    for(int i = 0; i < in; i++ )
      data[ i ] = s.data[ i ];
  } 
  void push(char c){
    data[in++] = c;
  }
  void pop(){
    if(in > 0)
      in--;
    else
      out++;
  }
  char top()const{
    return (in > 0) ? data[in-1] : 0;
  }
  int isempty( )const{
    return (in == 0) && (out == 0);
  }
  Stack& operator=(const Stack &s){
    if(this != &s){
      in = s.in;
      out = s.out;
      for( int i = 0; i < in; i++ )
	data[ i ] = s.data[ i ];
    }
    return *this;
  }
};

template<>
class template_mpitype<Stack>{
public:
  static const MPI_Datatype type(){
    MPI_Datatype newtype;
    MPI_Type_contiguous(MAXDEPTH+2, MPI_CHAR, &newtype);
    MPI_Type_commit(&newtype);
    return newtype;
  }
};

//---------------------------------------------------------------------
// function declaration
int ran(int n);
void makestr(char *array, int size, int kind, int maxdepth);
int isopen(char c);
int isclose(char c);

//---------------------------------------------------------------------
// function definition
inline int ran(int n){
  return rand() % n;
}

inline void makestr(char *array, int size, int kind, int maxdepth){
  int n = 0;
  int *s = new int[maxdepth+1];

  srand(time(NULL));

  while(size > (n + 2)){
    switch ((n == 0) ? ran(2) : ran(3)){
    case 0: // open
      if(n >= maxdepth) continue;
      *s = ran(kind);
      *array++ = openbr[*s++];
      n++;
      break;
    case 1:
      *array++ = 'X';
      break;
    case 2: // close
      *array++ = closebr[*--s];
      n--;
      break;
    }
    size--;
  }
  while(n-- > 0){
    *array++ = closebr[*--s];
    size--;
  }
  while(size-- > 1){
    *array++ = 'X';
  }
  *array = '\0';
  delete[] s;
}

inline int isopen(char c){
  return (c == '(') || (c == '{') || (c == '[') || (c == '<');
}

inline int isclose(char c){
  return (c == ')') || (c == '}') || (c == ']') || (c == '>');
}

inline int match(char a, char b){
  if(a == ')') return (b == '(');
  if(a == '}') return (b == '{');
  if(a == ']') return (b == '[');
  if(a == '>') return (b == '<');
  return 0;
}

struct G : public unary_ptr_function<int, Stack>{
  void operator()(int *b, const Stack *s)const{
    *b = s->isempty();
  }
} g;

struct P : public binary_ptr_function<int, char, Stack>{
  void operator()(int *b, const char *c, const Stack *s)const{
    *b = isclose(*c) ? (!s->isempty() && match(*c, s->top())) : 1;
  }
} p;

struct Oplus : public binary_ptr_function<int, int, int>{
  void operator()(int *r, const int *x, const int *y)const{
    *r = *x && *y;
  }
} oplus;

struct Q : public unary_ptr_function<Stack, char>{
  void operator()(Stack *s, const char *c)const{
    if(isopen(*c))
      s->push(*c);
    else if(isclose(*c))
      s->pop();
  }
} q;

struct Otimes : public binary_ptr_function<Stack, Stack, Stack>{
  void operator()(Stack *c, const Stack *a, const Stack *b)const{
    if(a->in <= b->out){
      *c = *b;
      c->out += a->out - a->in;
    }
    else{
      *c = *a;
      for(int i = 0; i < b->in; i++){
	c->data[a->in - b->out + i] = b->data[i];
      }
      c->in += b->in - b->out;
    }
  }
} otimes;

int SketoMain(int argc, char **argv){

  // generate string include brackets
  char str[SIZE];
  makestr(str, SIZE, KIND, DEPTH);

  // output string
  cout << str << "\n";

  // distribute data
  dist_list<char> *as = new dist_list<char>(SIZE, str);

  // calculate bracket matching
  int result;
  Stack st;
  list_skeletons::accumulate(g, p, oplus, q, otimes, &st, as, &result);
  delete as;

  // output result
  if(result == 1){
    if(skeleton::rank == 0){
      cout << "match\n";
    }
  }
  else{
    if(skeleton::rank == 0){
      cout << "no match\n";
    }
  }

  return 0;
}
