/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
// @author: Hideo NISHIMURA (nish@ipl.cs.uec.ac.jp)
// 
// This program is using C++ parallel skeleton on distributed list.
//
//-----------------------------//
//         How to use          //
//-----------------------------//
//
// This program is solving method of one-dimensional heat equation.
// (  , "partial differential equation" or "boundary value problem" )
//
//  Argument of console command :  %s size number
//     "length" is the length of x-coordinate
//     "time" is the passing time
//
//  After running this program, we  get u(x, t) of each x (interval : 1)
//  u(x, t) is a condition after "time" has passed. 
//
//  
//  We use "0 degrees C with both ends" as boundary value.
//  And initial condtion phi(x) , i.e. u(x, 0), is 30 degrees C ( constant )
//   
//  If you want to change initial condition, rewrite function "double phi( double )"



//-----------------------------//
//           Compile           //
//-----------------------------//
//
//  put C++ parallel library and "heat_equation.cpp" in the same directory
//
//     $ mpiCC heat_equation.cpp
//
//  to execute
//     $ mpirun -np "N" a.out "length" "time"

#include "list_skeletons.h"

#include <cstdlib>

//------------------------------------------------//
// global variables ( parameters for calculation) //
//------------------------------------------------//

// delta x
const double delta_x = 1;

// delta t
const double delta_t = 0.001;

// heat diffusivity ( of copper )
const double diffuse = 0.094099517;

// boundary condition
const double boundary = 0.0;


//----------------------------------//
//  functions using with Skelton    //
//----------------------------------//

//--- initial condition  i.e. u( x, 0 )
// phi( x ) returns 30 degrees C for each x
struct Phi : public unary_function< int, double > {
  double operator()  ( int index ) const {
    return 30.0;
  }
} phi;


//--- function multiples argument by member variable 
//    ( given when constructed ) 
  struct Mul_element: public unary_function< double, double > {
    double element;

    //--- constructer with an argument to give a variable
    Mul_element( double elem ) { element = elem; }

    //--- main content of function
    double operator() ( double x ) const {
      return x * element;
    }
  };


//--- dyadic ADD operator function
template < typename A >
struct Add : public binary_function< A, A, A > {
  A operator() ( A x, A y ) const {
    return x + y;
  }
};



//------------------------//
//     main function      //
//------------------------//


int SketoMain(int argc, char **argv){

  if ( argc != 3 ) {
    skeleton::mprintf( "usage: %s length time\n" );
	return 1;
  }

  const int length = atoi( argv[ 1 ] ); // length of bar
  const int time = atoi( argv[ 2 ] );   // passed time

  skeleton::time::initialize( ); // for time ekeeping

  // Generate a dist_list "bar" which means a condition of the bar.
  // Now, each element of dist_list is an inital value.
  dist_list< double > *bar = NULL;
  bar = new dist_list< double >( phi, length );

  
  // ---- operator instances ----
  Add< double > add; // make an instance of an add operator object

  // make an instance of multiple operator object with an argument
  Mul_element mul_minus_2( - 2 );
  // ---------------------------- 

  int count = 0;
  for ( double t = delta_t; count < time / delta_t; t += delta_t, count++ ) {
    // ----- algorithm -----
    // We approximate Heat Equation Problem by Finite Difference Method
    //
    //  bar( x, t+1 ) =
    //    gamma * ( bar( x+1, t ) + bar( x-1, t ) - 2 * bar( x, t ) ) + bar( x, t )
    //
    //        bar( x, t ) : Temperature with location ( X-axis ) "x", time "t"
    //           gamma    : = diffuse * delta_t / delta_x * delta_x
    //          diffuse   : Heat Diffusivity
    //

    dist_list< double >
      *left = NULL, *right = NULL, *zipped1 = NULL, *zipped2 = NULL, 
      *bar_mul = NULL, *renewed_bar = NULL;

    // left shifted array ( immediate right )
    right = list_skeletons::shiftl( boundary, bar );
    // right shifted array ( immediate left )
    left = list_skeletons::shiftr( boundary, bar );

    // store dist_list 
    //    [ bar( x-1, t ) + bar( x+1, t ) ]
    // in "zipped"
    zipped1 = list_skeletons::zipwith( add, left, right );

    if ( left ) delete left;
    if ( right ) delete right;
    
    // store dist_list 
    //    [ bar( x-1, t ) + bar( x+1, t ) - 2*bar( x, t ) ]
    // in "zipped2"
    bar_mul = list_skeletons::map( mul_minus_2, bar ); // multiple "bar" by -2;
    zipped2 = list_skeletons::zipwith( add, zipped1, bar_mul );

    if ( zipped1 ) delete zipped1;
    if ( bar_mul ) delete bar_mul;


    // make an instance of multiple operator object with an argument( gamma )
    Mul_element mul_gamma( ( diffuse * delta_t ) / ( delta_x * delta_x ) );

    // multiple dist_list "zipped2" by gamma
    list_skeletons::map_ow( mul_gamma, zipped2 );
    
    // store
    //    gamma * [ bar( x-1, t ) + bar( x+1, t ) - 2*bar( x, t ) ] + bar( x, t )
    // in "renewed_bar"
    renewed_bar = list_skeletons::zipwith( add, bar, zipped2 );

    if ( zipped2 ) delete zipped2;

    if ( bar ) delete bar;

    // renew "bar"
    bar = renewed_bar;
  }

  skeleton::time::set_time( "calculate" ); // for time ekeeping

  /*
    list_skeletons::print( bar );
  */

  if ( bar ) delete bar;

  skeleton::time::print( ); // show the result of time ekeeping
  

  return 0;
}
