/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file MM.cpp
 * @brief Matrix Multiplication by Data Parallel Skeletons. 
 * @author Kento EMOTO
 *
 */ 
#include <iostream>
#include <iomanip>
#include <mpi.h>
#include <cmath>
#include <cstdlib>
#include "matrix_skeletons.h"

#include "primitive_functions.h"

using namespace primitive_functions;


template <typename A>
class Iprod2  : public binary_function< matrix<A>, matrix<A> , A>{
public:
    void operator()(const matrix <A> &a, const matrix <A> &b, A *res) const{
		A r = 0;
		int n = a.getCols();
		for(int i = 0; i < n; i++){
			r+=a.at(0,i) * b.at(0, i);
		}
		*res = r;
	}
};

template <typename A>
class Tri {
public:
  	Tri(){}
    void operator()(A *r) const{
		r->tr();
	}
};

template <typename A>
dist_matrix <matrix <A> > * all_cols2(const dist_matrix <A> *a){
	dist_matrix<matrix <A> >*x = new dist_matrix<matrix <A> >();
	x->shapeCopy(a);	
	matrix_skeletons::map(Singleton<A>(), a, x);
	dist_matrix < matrix <A> > *y = matrix_skeletons::allreduce_col(Abv<A>(), x);
	delete x;
	return y;
}
template <typename A>
dist_matrix <matrix <A> > * all_rows2(const dist_matrix <A> *a){
	dist_matrix<matrix <A> >*x = new dist_matrix<matrix <A> >();
	x->shapeCopy(a);
	matrix_skeletons::map(Singleton<A>(), a, x);
	dist_matrix < matrix <A> > *y = matrix_skeletons::allreduce_row(Bsd<A>(), x);
	delete x;
	return y;
}




template <class C, class A, class B>
inline void mmP4(dist_matrix<C> &Z2, const dist_matrix<A> &X2, const dist_matrix<B> &Y2)
{
	dist_matrix < matrix < int > > *A2;
	dist_matrix < matrix < int > > *B2;
	A2 = all_rows2(&X2);
	B2 = all_cols2(&Y2);
	matrix_skeletons::map_i(Tri< matrix <B> >(), B2);
	matrix_skeletons::zipwith(Iprod2<C>(), A2, B2, &Z2);
	delete B2;
	delete A2;
}


/***
 * main routine.
 * command line arguments: n
 *   n : #rows of A
 */

int SketoMain( int argc, char **argv )
{
	int n = 1000;
	int debug = 0;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	if(argc > 2){
		debug = atoi(argv[2]);
	}

	int nproc = skeleton::procs;
	int myrank = skeleton::rank;
	
	int q = (int)(sqrt(nproc)+0.1);
	
	int r = q;
	for(int k = 1; k <= q; k++){
		if((nproc % k)==0) r = k;
	}
	q = nproc / r;
	if(myrank==0){
		std::cout << "A * B : (" << n << " x " << n << ") * (" << n << " x " << n << ")\n";
		std::cout << " total prosess = " << nproc << "\n";
		std::cout << " using prosess = " << q << " x "<< r << "\n";
	}




	dist_matrix<int> A(n, n, GenAdd<int>(), q, r);
	dist_matrix<int> B(n, n, GenSub<int>(), q, r);
	if(myrank==0){
		std::cout << "A = \n";
	}
	if(debug){
		A.print(std::cout);
	}
	if(myrank==0){
		std::cout << "B = \n";
	}
	if(debug){
		B.print(std::cout);
	}
	dist_matrix<int> C(n, n, GenZero<int>(), q, r);
	mmP4(C,A,B);
	
	if(myrank==0){
		std::cout << "C = A * B = \n";
	}
	if(debug){
		C.print(std::cout);
	}
	return 0;
}

