/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/***
 * @file MMT4.cpp
 * @brief Matrix Multiiplication by parallel computing.
 *  Cannon's Algorithm.
 *
 * @author Kento EMOTO
 */
 

#include <iostream>
#include <iomanip>
#include <mpi.h>
#include <cmath>
#include <cstdlib>
#include "matrix_skeletons.h"

#include "primitive_functions.h"

using namespace primitive_functions;



/***
 * two associative (and distrivutie) operators
 */

inline void mul(const int &a, const int &b, int*c)
{
	*c = a * b;
}

inline void add(const int &a, const int &b, int*c)
{
	*c = a + b;
}



/***
 * used by rotations in Cannon's algorithm
 */
class Neg {
public:
    int operator()(const int i) const { return -i; }
};


/***
 * multiplication of local block-matrices.
 */
template <class C, class A, class B>
inline void MatrixMultiplyLocal(dist_matrix<C>& Z, const dist_matrix<A>& X, const dist_matrix<B>& Y)
{
	int rows = X.getLocalRows();
	int cols = Y.getLocalCols();
	int inner = X.getLocalCols();
	for(int i = 0; i < rows; i++){
		for(int j = 0; j < cols; j++){
			C v = Z.get(i, j);
			for(int k = 0; k < inner; k++){
				v += X.get(i, k) * Y.get(k, j);
			}
			Z.set(i, j, v);
		}
	}
}

/***
 * multiplication of matrices.
 */
template <class C, class A, class B>
inline void MatrixMultiply(dist_matrix<C>& Z, dist_matrix<A>& X, dist_matrix<B>& Y)
{
	matrix_skeletons::rotateCols(Neg(), &X);          // Cannon's algorithm
	matrix_skeletons::rotateRows(Neg(), &Y);
	int q = X.getBlocksInRow();
	for(int i = 0; i < q; i++){
		MatrixMultiplyLocal(Z, X, Y);
		matrix_skeletons::rotateColsConst(-1, &X);
		matrix_skeletons::rotateRowsConst(-1, &Y);
	}
}


/***
 * generalized multiplication of local block-matrices.
 */

template <class C, class A, class B>
inline void GeneralizedMatrixMultiplyLocal(dist_matrix<C>& Z, const dist_matrix<A>& X, const dist_matrix<B>& Y, void (&oplus)( const C&, const C&, C*), void (&otimes)(const A&, const B&, C*))
{
	int rows = X.getLocalRows();
	int cols = Y.getLocalCols();
	int inner = X.getLocalCols();
	for(int i = 0; i < rows; i++){
		for(int j = 0; j < cols; j++){
			C v = Z.get(i, j);
			for(int k = 0; k < inner; k++){
				C u;
				otimes(X.get(i, k), Y.get(k, j), &u);
				C tmp = v;
				oplus(tmp, u, &v);
			}
			Z.set(i, j, v);
		}
	}
}

/***
 * generalized multiplication of matrices.
 */
template <class C, class A, class B>
inline void GeneralizedMatrixMultiply(dist_matrix<C>& Z, dist_matrix<A>& X, dist_matrix<B>& Y, void (&oplus)(const C&, const C&, C*), void (&otimes)(const A&, const B&, C*))
{
	matrix_skeletons::rotateCols(Neg(), &X);          // Cannon's algorithm
	matrix_skeletons::rotateRows(Neg(), &Y);
	int q = X.getBlocksInRow();
	for(int i = 0; i < q; i++){
		GeneralizedMatrixMultiplyLocal(Z, X, Y, oplus, otimes);
		matrix_skeletons::rotateColsConst(-1, &X);
		matrix_skeletons::rotateRowsConst(-1, &Y);
	}
}


/***
 * main routine.
 * command line arguments: n
 *   n : #rows of A
 *   d : if d != 0 then print matrices
 */


int SketoMain( int argc, char **argv )
{
	int n = 1000;
	int debug = 0;
	if(argc > 1){
		n = atoi(argv[1]);
	}
	if(argc > 2){
		debug = atoi(argv[2]);
	}

	int nproc = skeleton::procs;
	int myrank = skeleton::rank;
	
	int q = (int)(sqrt(nproc)+0.1);
	
	if(myrank==0){
		std::cout << "A * B : (" << n << " x " << n << ") * (" << n << " x " << n << ")\n";
		std::cout << " total prosess = " << nproc << "\n";
		std::cout << " using prosess = " << q*q << "\n";
	}

	dist_matrix<int> A(n, n, GenAdd<int>(), q, q);
	dist_matrix<int> B(n, n, GenSub<int>(), q, q);
	if(myrank==0){
		std::cout << "A = \n";
	}
	if(debug){
		A.print(std::cout);
	}
	if(myrank==0){
		std::cout << "B = \n";
	}
	if(debug){
		B.print(std::cout);
	}

	dist_matrix<int> C(n, n, GenZero<int>(), q, q);
//	MatrixMultiply(C,A,B);
	GeneralizedMatrixMultiply(C,A,B,add,mul);
	
	if(myrank==0){
		std::cout << "C = A * B = \n";
	}
	if(debug){
		C.print(std::cout);
	}
	return 0;
}

 
