/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file qrlocal.cpp
 * @brief QR factorization (sequential)
 * @author Kento EMOTO
 */

#include <iostream>
#include <iomanip>
#include <cmath>
#include "matrix.h"

#include "primitive_functions.h"

using namespace primitive_functions;


template <typename A>
class Neg {
public:
	void operator()(A *r) const {
		*r = - *r;
	}
};

template <typename A>
class DivByi : public unary_function < A, A >{
	A d;
public:
	DivByi(A dd){
		d = dd;
	}
	void operator()(A *r) const {
		*r = *r / d;
	}
};

template <typename A>
class Abs {
public:
	void operator()(A *r) const {
		*r = (*r < 0) ? -*r : *r;
	}
};

template <class C, class A, class B>
inline void mm(matrix<C>& Z, const matrix<A>& X, const matrix<B>& Y)
{
	int rows = X.getRows();
	int cols = Y.getCols();
	int inner = X.getCols();
	for(int i = 0; i < rows; i++){
		for(int j = 0; j < cols; j++){
			C v = Z.get(i, j);
			for(int k = 0; k < inner; k++){
				v += X.get(i, k)*Y.get(k, j);
			}
			Z.set(i, j, v);
		}
	}
}

template <typename A>
class GenQh {
	const matrix<A> &e;
public:
	GenQh(const matrix<A> &ee) : e(ee) {
	}
	void operator()(int i, int j, A *r) const {
		*r = - e.at(i,0) * e.at(j,0);
		if(i==j) *r+=1;
	}
};

void householder(const matrix<double>& v, matrix<double>& Q, matrix<double>& R)
{
	double v0 = v.at(0,0); 
	double e0 = 1;
	double vnorm = 0;
	matrix<double> v2;
	m_skeletons::zipwith(Mul<double>(), &v, &v, &v2);
	m_skeletons::reduce(Add<double>(), Add<double>(), &v2, &vnorm);
	e0 = sqrt(vnorm);

	if(e0==0){
		Q.setSize(v.getRows(), v.getRows());
		Q.eye(1, 0);
		R.setSize(v.getRows(), 1);
		R.fill(0);
		return;
	}
	if(fabs(v0-e0)<e0){
		e0 = - e0;
	}

	
	matrix<double> e = v;
	e.at(0,0) -= e0;
	double a = 0;
	matrix<double> e2;
	m_skeletons::zipwith(Mul<double>(), &e, &e, &e2);
	m_skeletons::reduce(Add<double>(), Add<double>(), &e2, &a);
	a = sqrt(a/2);
	m_skeletons::map_i(DivByi<double>(a), &e);
	Q.setSize(e.getRows(), e.getRows());
	Q.generate(GenQh<double>(e));

	R.setSize(e.getRows(), 1);
	R.fill(0);
	R.at(0,0) = e0;
}
template <typename A>
class EmbedEye {
	const matrix<A> &q3;
	const int n, m;
public:
	EmbedEye(int mm, int nn,const matrix<A> &qq3) : q3(qq3), n(nn), m(mm) {
	}
	void operator()(int i, int j, A*r) const {
		if ((i >= m) && (j >= n)){
			*r = q3.get(i-m, j-n);
		} else if (i == j){
			*r = 1;
		} else {
			*r = 0;
		}
	}
};


void qrfact(const matrix<double>& A, matrix<double>& Q, matrix<double>& R)
{
	if(A.getCols()==1){
		householder(A, Q, R);
		return;
	}
	matrix<double> A1, A2;
	int n = (A.getCols() + 1)/2;
	
	A.splitv(n, &A1, &A2);
	matrix<double> q1, r1, r11, z;
	qrfact(A1, q1, r1);
	r1.splith(n, &r11, &z);
	matrix<double> a(A2.getRows(), A2.getCols()), qt;
	qt = q1;
	qt.tr();
	a.fill(0);
	mm(a, qt, A2);
	matrix<double> r12, A22;
	a.splith(n, &r12, &A22);
	matrix<double> q2, r22;
	qrfact(A22, q2, r22);
	matrix<double> q3(A2.getRows(), A2.getRows());
	q3.generate(EmbedEye<double>(A2.getRows() - q2.getRows(), A2.getRows() - q2.getRows(), q2));
	Q.setSize(A2.getRows(), A2.getRows());
	Q.fill(0);
	mm(Q, q1, q3);
	m_skeletons::uppertriangular(&r11, &r12, &r22, 0.0, &R);
}

// return max(abs(A-Q*R))
double getErr(const matrix<double> &A, const matrix<double> &Q, const matrix<double> &R)
{
	matrix<double> tmp(Q.getRows(), R.getCols());
	tmp.fill(0);
 	mm(tmp,Q,R);
	matrix<double> tmp2(A.getRows(), A.getCols());
	m_skeletons::zipwith(Sub<double>(), &A, &tmp, &tmp2);
	m_skeletons::map_i(Abs<double>(), &tmp2);
	double merr = 0;
	m_skeletons::reduce(Max<double>(), Max<double>(), &tmp2, &merr);
	return merr;
}

int main(int argc, char *argv[])
{
	int m = 5;
	int n = 4;
	int db = 0;
	if(argc > 1){
		m = atoi(argv[1]);
	}
	if(argc > 2){
		n = atoi(argv[2]);
	}
	if(argc > 3){
		db = atoi(argv[3]);
	}

	matrix< double > a(m, n);
	for(int i = 0; i < m; i++){
		for(int j = 0; j < n; j++){
			a.at(i, j) = (i+j);
		}
	}
	std::cout << "QR factorization of ( " << m << " x " << n << " ) matrix\n";
	if(db){
		std::cout << "a = \n" << a << std::endl;
	}
	matrix<double> q, r;
	qrfact(a, q, r);
	if(db){
		std::cout << "q = \n" << q << std::endl;
		std::cout << "r = \n" << r << std::endl;
	}
	matrix<double> qt = q;
	qt.tr();
	matrix<double> eye(q.getRows(), q.getCols());
	eye.eye(1,0);
	double merr = getErr(eye, q, qt);
	std::cout << "max(abs(I-Q'*Q)) = " << merr << std::endl;
	double merr2 = getErr(a, q, r);
	std::cout << "max(abs(A-Q*R)) = " << merr2 << std::endl;
}
