/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file qrlocal.cpp
 * @brief QR factorization (quadtree, sequential)
 * @author Kento EMOTO
 */

#include <iostream>
#include <iomanip>
#include <cmath>
#include "matrix.h"


#include "primitive_functions.h"

using namespace primitive_functions;

template <typename A>
class Neg {
public:
	void operator()(A *r) const {
		*r = - *r;
	}
};

template <typename A>
class DivByi : public unary_function < A, A >{
	A d;
public:
	DivByi(A dd){
		d = dd;
	}
	void operator()(A *r) const {
		*r = *r / d;
	}
};

template <typename A>
class Abs {
public:
	void operator()(A *r) const {
		*r = (*r < 0) ? -*r : *r;
	}
};


template <typename A>
class GenQ12 {
	const matrix<A> &q1;
	const matrix<A> &q2;
	const int n;
public:
	GenQ12(const matrix<A> &qq1, const matrix<A> &qq2) : q1(qq1), q2(qq2), n(qq1.getRows()) {
	}
	void operator()(int i, int j, A *r) const {
		if((i < n) && (j < n)) {
			*r = q1.get(i, j);
		} else if ((i >= n) && (j >= n)){
			*r = q2.get(i-n, j-n);
		} else {
			*r = 0;
		}
	}
};

template <typename A>
class GenQ12e {
	const matrix<A> &q1;
	const matrix<A> &q2;
	const int n;
public:
	GenQ12e(const matrix<A> &qq1, const matrix<A> &qq2) : q1(qq1), q2(qq2), n(qq1.getRows()/2) {
	}
	void operator()(int i, int j, A*r) const {
		if(i < n){
			if(j < n){
				*r = q1.at(i, j);
			} else if (n*2 <= j && j < 3*n) {
				*r = q1.at(i, j-n);
			} else {
				*r = 0;
			}
		} else if (i < 2*n) {
			if(n <= j && j < n*2){
				*r = q2.at(i-n, j-n);
			} else if (3*n <= j) {
				*r = q2.at(i-n, j-2*n);
			} else {
				*r = 0;
			}
		} else if (i < 3*n) {
			if(j < n){
				*r = q1.at(i-n, j);
			} else if (n*2 <= j && j < 3*n) {
				*r = q1.at(i-n, j-n);
			} else {
				*r = 0;
			}
		} else {
			if(n <= j && j < n*2){
				*r = q2.at(i-n*2, j-n);
			} else if (3*n <= j) {
				*r = q2.at(i-n*2, j-2*n);
			} else {
				*r = 0;
			}
		}
	}
};

template <typename A>
class GenQ62 {
	const matrix<A> &q6;
	const int n;
public:
	GenQ62(const matrix<A> &qq6) : q6(qq6), n(qq6.getRows()) {
	}
	void operator()(int i, int j, A*r) const {
		if ((i >= n) && (j >= n)){
			*r = q6.get(i-n, j-n);
		} else if (i == j) {
			*r = 1;
		} else {
			*r = 0;
		}
	}
};

template <typename A>
class GenQ62e {
	const matrix<A> &q6;
	const int n;
public:
	GenQ62e(const matrix<A> &qq6) : q6(qq6), n(qq6.getRows()/2) {
	}
	void operator()(int i, int j, A*r) const {
		if((n<= i) && (i < n*3) && (n<= j) && (j < n*3)) {
			*r = q6.get(i-n, j-n);
		} else if (i == j){
			*r = 1;
		} else {
			*r = 0;
		}
	}
};

template <typename A>
class GenQ42e {
	const matrix<A> &q4;
	const int n;
public:
	GenQ42e(const matrix<A> &qq4) : q4(qq4), n(qq4.getRows()) {
	}
	void operator()(int i, int j, A*r) const {
		if((n*2<= i) && (i < n*3) && (n*2<= j) && (j < n*3)) {
			*r = q4.get(i - 2*n, j - 2*n);;
		} else if (i == j){
			*r = 1;
		} else {
			*r = 0;
		}
	}
};

template <class C, class A, class B>
inline void mm(matrix<C>& Z, const matrix<A>& X, const matrix<B>& Y)
{
	int rows = X.getRows();
	int cols = Y.getCols();
	int inner = X.getCols();
	for(int i = 0; i < rows; i++){
		for(int j = 0; j < cols; j++){
			C v = Z.get(i, j);
			for(int k = 0; k < inner; k++){
				v += X.get(i, k)*Y.get(k, j);
			}
			Z.set(i, j, v);
		}
	}
}
void qrfactQ(const matrix<double>& A, matrix<double>& Q, matrix<double>& R);
double getErr(const matrix<double> &A, const matrix<double> &Q, const matrix<double> &R);
double getErr2(const matrix<double> &A, const matrix<double> &B, const matrix<double> &Q, const matrix<double> &R);

void efun(const matrix<double>& N, const matrix<double>& S, matrix<double>& Q, matrix<double>& R){
// 	std::cout << "efun" << std::endl;
// 	std::cout << "N = " << std::endl << N;
// 	std::cout << "S = " << std::endl << S;
	if(N.getCols()==1){
		Q.setSize(2,2);
		double a = N.get(0,0);
		double b = S.get(0,0);
		if(b == 0){
			Q.at(0,0) = 1;
			Q.at(0,1) = 0;
			Q.at(1,0) = 0;
			Q.at(1,1) = 1;
			R.singleton(a);
// 	std::cout << "efun Gibbens zero" << std::endl;
// 	std::cout << "N = \n" << N;
// 	std::cout << "S = \n" << S;
// 	std::cout << "Q = \n" << Q;
// 	std::cout << "R = \n" << R;
// 	std::cout << "efun err = " << getErr2(N, S, Q, R) << std::endl;
			return;
		}
		double r = sqrt(a*a + b*b);
		double c = a / r;
		double s = -b / r;
		Q.at(0,0) = c;
		Q.at(0,1) = s;
		Q.at(1,0) = -s;
		Q.at(1,1) = c;
		R.singleton(r);
// 	std::cout << "efun Gibbens" << std::endl;
// 	std::cout << "N = \n" << N;
// 	std::cout << "S = \n" << S;
// 	std::cout << "Q = \n" << Q;
// 	std::cout << "R = \n" << R;
// 	std::cout << "efun err = " << getErr2(N, S, Q, R) << std::endl;
		return;
	}
	int n2 = N.getRows();
	int n = n2 / 2;
	matrix<double> N11, N12, N21, N22;
	N.splitQ(n, n, &N11, &N12, &N21, &N22);
	matrix<double> S11, S12, S21, S22;
	S.splitQ(n, n, &S11, &S12, &S21, &S22);
	matrix<double> q1, N1, q2, N2;
	efun(N11, S11, q1, N1);
// 	std::cout << "efun(N11, S11, q1, N1) ret" << std::endl;
	efun(N22, S22, q2, N2);
// 	std::cout << "efun(N22, S22, q2, N2)" << std::endl;
	matrix<double> q12(n2*2, n2*2);
 // 	std::cout << "q1 = \n" << q1 << "q2 = \n" << q2 << std::endl;
	q12.generate(GenQ12e<double>(q1, q2));
 // 	std::cout << "q12 = \n" << q12 << std::endl;


	matrix<double> A2, U(n2, n), Us, Un;
	m_skeletons::abv(&N12, &S12, &A2);
	q1.tr();
	U.fill(0);
	mm(U, q1, A2);
	U.splith(n, &Un, &Us);
	matrix<double> q4, r4;
	qrfactQ(Us, q4, r4);
	matrix<double> q42(n2*2, n2*2);
	q42.generate(GenQ42e<double>(q4));
	matrix<double> q5(n2*2, n2*2);
	q5.fill(0);
//  	std::cout << "q4 = \n" << q4 << std::endl;
	mm(q5, q12, q42);

	matrix<double> q6, n6;
	efun(N2, r4, q6, n6);
// 	std::cout << "efun(N2, r4, q6, n6)" << std::endl;
	matrix<double> q62(n2*2, n2*2);
	q62.generate(GenQ62e<double>(q6));
// 	std::cout << "q62" << std::endl;
	
	Q.setSize(n2*2,n2*2);
	Q.fill(0);
//  	std::cout << "q5 = \n" << q5 << "q62 = \n" << q62 << std::endl;
//  	std::cout << "q12 = \n" << q12 << "q42 = \n" << q42 << std::endl;
	mm(Q, q5, q62);
// 	std::cout << "Q" << std::endl;
	m_skeletons::uppertriangular(&N1, &Un, &n6, 0.0, &R);
// 	std::cout << "N = \n" << N;
// 	std::cout << "S = \n" << S;
// 	std::cout << "Q = \n" << Q;
// 	std::cout << "R = \n" << R;
// 	std::cout << "efun err = " << getErr2(N, S, Q, R) << std::endl;
	
}
void qrfactQ(const matrix<double>& A, matrix<double>& Q, matrix<double>& R)
{
 // 	std::cout << "qrfactQ: A = \n" << A << std::endl; 
	if(A.getCols()==1){
		Q.singleton(1);
		R.singleton(A.get(0,0));
// 	std::cout << "A = \n" << A;
// 	std::cout << "Q = \n" << Q;
// 	std::cout << "R = \n" << R;
// 	std::cout << "err = " << getErr(A, Q, R) << std::endl;
		return;
	}
	matrix<double> A11, A12, A21, A22;
	int n2 = A.getRows();
	int n = n2 / 2;
	assert((n*2 == n2) && (A.getCols()==n2));
	
	A.splitQ(n, n, &A11, &A12, &A21, &A22);
	matrix<double> q1, r1, q2, r2;
	qrfactQ(A11, q1, r1);
	qrfactQ(A21, q2, r2);
	matrix<double> q12(n2, n2);
	q12.generate(GenQ12<double>(q1, q2));
//  	std::cout << "q1 = " << std::endl << q1;
//  	std::cout << "q2 = " << std::endl << q2;
//  	std::cout << "q12 = " << std::endl << q12;
	matrix<double> q3, r3;
	efun(r1, r2, q3, r3);
 // 	std::cout << "r1 = " << std::endl << r1;
//  	std::cout << "r2 = " << std::endl << r2;
//  	std::cout << "q3 = " << std::endl << q3;
//  	std::cout << "r3 = " << std::endl << r3;

	

	// 	std::cout << "efun(r1, r2, q3, r3)" << std::endl;
	matrix<double> q4(n2, n2), q4tr;
	q4.fill(0);
	mm(q4, q12, q3);
	q4tr.copyTr(q4);
	matrix<double> A2, U(n2, n), Us, Un;
	m_skeletons::abv(&A12, &A22, &A2);
	U.fill(0);
	mm(U, q4tr, A2);
// 	std::cout << "U = " << std::endl << U;
// 	std::cout << "q4tr = " << std::endl << q4tr;
	U.splith(n, &Un, &Us);
	matrix<double> q6, r6;
	qrfactQ(Us, q6, r6);
	matrix<double> q62(n2, n2);
	q62.generate(GenQ62<double>(q6));
	Q.setSize(n2,n2);
	Q.fill(0);
// 	std::cout << "q4 = \n" << q4;
// 	std::cout << "q62 = \n" << q62;
	mm(Q, q4, q62);
// 	std::cout << "r3 = \n" << r3;
// 	std::cout << "Un = \n" << Un;
// 	std::cout << "r6 = \n" << r6;
	m_skeletons::uppertriangular(&r3, &Un, &r6, 0.0, &R);
// 	std::cout << "A = \n" << A;
// 	std::cout << "Q = \n" << Q;
// 	std::cout << "R = \n" << R;
// 	std::cout << "err = " << getErr(A, Q, R) << std::endl;
}

// return max(abs(A-Q*R))
double getErr(const matrix<double> &A, const matrix<double> &Q, const matrix<double> &R)
{
	matrix<double> tmp(Q.getRows(), R.getCols());
	tmp.fill(0);
 	mm(tmp,Q,R);
	matrix<double> tmp2(A.getRows(), A.getCols());
	m_skeletons::zipwith(Sub<double>(), &A, &tmp, &tmp2);
	m_skeletons::map_i(Abs<double>(), &tmp2);
	double merr = 0;
	m_skeletons::reduce(Max<double>(), Max<double>(), &tmp2, &merr);
	return merr;
}
double getErr2(const matrix<double> &A, const matrix<double> &B, const matrix<double> &Q, const matrix<double> &R)
{
	matrix<double> tmpR(R.getRows()*2, R.getCols());
	tmpR.fill(0);
	tmpR.embed(0, 0, &R);
	matrix<double> tmpA(A.getRows() + B.getRows(), A.getCols());
	tmpA.embed(0, 0, &A);
	tmpA.embed(A.getRows(), 0, &B);
	
	matrix<double> tmp(Q.getRows(), tmpR.getCols());
	tmp.fill(0);
 	mm(tmp,Q,tmpR);
	matrix<double> tmp2(tmpA.getRows(), tmpA.getCols());
	m_skeletons::zipwith(Sub<double>(), &tmpA, &tmp, &tmp2);
	m_skeletons::map_i(Abs<double>(), &tmp2);
	double merr = 0;
	m_skeletons::reduce(Max<double>(), Max<double>(), &tmp2, &merr);
	return merr;
}

int main(int argc, char *argv[])
{
	int n = 4;
	int k = 2;
	int db = 0;
	if(argc > 1){
		k = atoi(argv[1]);
	}
	if(argc > 2){
		db = atoi(argv[2]);
	}
	n = 1 << k;
	matrix< double > a(n, n);
	for(int i = 0; i < n; i++){
		for(int j = 0; j < n; j++){
			a.at(i, j) = (i+j);
		}
	}
	std::cout << "QR factorization of ( " << n << " x " << n << " ) matrix\n";
	if(db){
		std::cout << "a = \n" << a << std::endl;
	}
	matrix<double> q, r;
	qrfactQ(a, q, r);
	if(db){
		std::cout << "q = \n" << q << std::endl;
		std::cout << "r = \n" << r << std::endl;
	}
	matrix<double> qt = q;
	qt.tr();
	matrix<double> eye(q.getRows(), q.getCols());
	eye.eye(1,0);
	double merr = getErr(eye, q, qt);
	std::cout << "max(abs(I-Q'*Q)) = " << merr << std::endl;
	double merr2 = getErr(a, q, r);
	std::cout << "max(abs(A-Q*R)) = " << merr2 << std::endl;
}
