/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file mis.cpp
 * @brief A sample program which solves maximum independent-set sum
 *        problem on rose trees. 
 * @author Kiminori MATSUZAKI
 */

#include <assert.h>
#include <iostream>
#include <algorithm>

#include <mpi.h>

#include <skeleton.h>
#include <rose_tree_skeletons.h>

/**
 * @brief Read a rose tree in a list format.
 *
 * The first line shows the size of the rose tree. Then a list of the 
 * nodes with depth information follows in a prefix traversal on the tree.
 * One more line with depth 0 is added for sentinel.
 *
 * For example, the tree in Haskell
 * @code RNode a [RLeaf b, RNode c [RLeaf d, RLeaf e], RLeaf f] @endcode
 * is arranged into a file as follows,
 * @code
 * 6
 * 0 a
 * 1 b
 * 1 c
 * 2 d
 * 2 e
 * 1 f
 * 0 x
 * @endcode
 *
 * @warning This function does not work correctly in case that the tree
 * is only a leaf. 
 */
template< typename A >
node< rose_value< A > > *
read_rose_tree_from_list( const char *filename )
{
  std::fstream fin( filename, std::ios::in );
  if ( fin.fail( ) ) return NULL;

  int size;
  fin >> size;

  int depth0, depth1;
  A value0, value1;

  fin >> depth0 >> value0;
  fin >> depth1 >> value1; 

  // we asume that the tree is not a leaf.
  if ( depth1 == 0 ) {
    rose_value< A > rval( value0, rose_value< A >::ORGLEAF );
    node< rose_value< A > > *tree = new node< rose_value< A > >( rval );
    return tree;
  }
  else {
    rose_value< A > rval( value0, rose_value< A >::ORGNODE );
    node< rose_value< A > > *tree = new node< rose_value< A > >( rval );

    read_rose_tree_from_list_rec( tree, 0, &depth1, &value1, fin );
    return tree;
  }
}

template< typename A >
static void
read_rose_tree_from_list_rec( node< rose_value< A > >* root,
			      int root_depth,
			      int *next_depth,
			      A *next_value,
			      std::fstream &fin )
{
  //  int cur_depth = *next_depth;
  A   cur_value = *next_value;

  int ndepth;
  A nvalue;
  
  fin >> ndepth >> nvalue;
  *next_depth = ndepth;
  *next_value = nvalue;
  
  if ( root_depth + 1 >= *next_depth ) {
    // add node as a leaf.
    rose_value< A > rval( cur_value, rose_value< A >::ORGLEAF );
    node< rose_value< A > > *child = new node< rose_value< A > >( rval );
    root->left = child; child->parent = root;
  }
  else {
    // add node as a normal node
    rose_value< A > rval( cur_value, rose_value< A >::ORGNODE );
    node< rose_value< A > > *child = new node< rose_value< A > >( rval );
    root->left = child; child->parent = root;

    read_rose_tree_from_list_rec( child,
				  root_depth + 1,
				  next_depth,
				  next_value,
				  fin );
  }

  if ( root_depth + 1 == *next_depth ) {
    // add node as dummy
    rose_value< A > rval;
    rval.valuetype = rose_value< A >::DUMMY;
    node< rose_value< A > > *child = new node< rose_value< A > >( rval );
    root->right = child; child->parent = root;
    
    read_rose_tree_from_list_rec( child,
				  root_depth,
				  next_depth,
				  next_value,
				  fin );
  } else {
    // add node as nil
    rose_value< A > rval;
    rval.valuetype = rose_value< A >::ORGNIL;

    node< rose_value< A > > *child = new node< rose_value< A > >( rval );
    root->right = child; child->parent = root;
  }
}

////////////////////////////////////////////////////////////////////////
// Utility functions

// // Definition exist in <algorithm>
// template< typename A > inline A max( const A &x, const A &y )
// { return ( x > y ) ? x : y; }

template< typename A > inline A plus( const A &x, const A &y )
{
  if ( x < 0 ) return -1;
  if ( y < 0 ) return -1;
  return x + y;
}

////////////////////////////////////////////////////////////////////////
// Definitions of types of intermediate data structures.

/**
 * @brief A pair representing the arguments of otimes, i.e., (r, s)
 */
struct mis_pair {
  int r, s;
  mis_pair( int r_ = 0, int s_ = 0 )
    : r( r_ ), s( s_ ) { }
  mis_pair( bool )
    : r( 0 ), s( 0 ) { }
};

/**
 * @brief A pair representing the left-argument of oplus, i.e., (a, b, c, d)
 */
struct mis_4tup {
  int a, b, c, d;
  mis_4tup( int a_ = 0, int b_ = 0, int c_ = 0, int d_ = 0 )
    : a( a_ ), b( b_ ), c( c_ ), d( d_ ) {}
  mis_4tup( bool )
    : a( 0 ), b( 0 ), c( 0 ), d( 0 ) {}
};

////////////////////////////////////////////////////////////////////////
// Definitions of Function Objects

class Func_mis_k1 : public unary_function< int, mis_pair > {
public:
  mis_pair operator()( const int &a ) const {
    return mis_pair( 0, std::max( a, 0 ) );
  }
};

class Func_mis_k2 : public unary_function< int, mis_4tup > {
public:
  mis_4tup operator()( const int &b ) const {
    return mis_4tup( -1, b, 0, 0 );
  }
};

class Func_mis_oplus : public binary_function< mis_4tup, mis_pair, mis_pair > {
public:
  mis_pair operator()( const mis_4tup &abcd, const mis_pair &rs ) const {
    const int ar = plus( abcd.a, rs.r );
    const int br = plus( abcd.b, rs.r );
    const int cs = plus( abcd.c, rs.s );
    const int ds = plus( abcd.d, rs.s );
    return mis_pair( std::max( ar, cs ), std::max( br, ds ) );
  }
};

class Func_mis_otimes : public binary_function< mis_pair, mis_pair, mis_pair > {
public:
  mis_pair operator()( const mis_pair &rs1, const mis_pair &rs2 ) const {
    return mis_pair( plus( rs1.r, rs2.r) , plus( rs1.s, rs2.s ) );
  }
};

class Func_mis_p1 {
public:
  mis_4tup operator()( const mis_4tup &abcd1, const mis_pair &rs1,
		       const mis_4tup &abcd2, const mis_pair &rs2 ) const {
    return mis_4tup( std::max( plus( plus( abcd1.a, rs1.r ), plus( abcd2.a, rs2.r )),
			       plus( plus( abcd1.c, rs1.s ), plus( abcd2.b, rs2.r )) ),
		     std::max( plus( plus( abcd1.b, rs1.r ), plus( abcd2.a, rs2.r )),
			       plus( plus( abcd1.d, rs1.s ), plus( abcd2.b, rs2.r )) ),
		     std::max( plus( plus( abcd1.a, rs1.r ), plus( abcd2.c, rs2.s )),
			       plus( plus( abcd1.c, rs1.s ), plus( abcd2.d, rs2.s )) ),
		     std::max( plus( plus( abcd1.b, rs1.r ), plus( abcd2.c, rs2.s )),
			       plus( plus( abcd1.d, rs1.s ), plus( abcd2.d, rs2.s )) ) );
  }
};

class Func_mis_p2 {
public:
  mis_pair operator()( const mis_4tup &, const mis_pair &,
		       const mis_4tup &, const mis_pair & ) const {
    return mis_pair( 0, 0 );
  }
};

const mis_4tup mis_e_oplus( 0, -1, -1, 0 );
const mis_pair mis_e_otimes( 0, 0 );

////////////////////////////////////////////////////////////////////////
// Main routine

void usage( )
{
  throw skeleton::error( 1, "mprun -np <size> mis <inputfile>" );
}

int SketoMain( int argc, char *argv[] )
{
  if ( argc < 2 ) {
    usage( );
  }

  // create tree (local)
  const char *filename = argv[ 1 ];
  node< rose_value< int > > *s_tree = NULL;
  if ( skeleton::rank == 0 ) {
    s_tree = read_rose_tree_from_list< int >( filename );
    assert( s_tree );
  }

  skeleton::time::initialize( );

  // distribute
  dist_rose_tree< int > *d_tree
    = rose_tree_skeletons::distribute< int >( s_tree );

  skeleton::time::set_time( "after distribute" );

  // reduce
  mis_pair mis = rose_tree_skeletons::reduce_c( Func_mis_k1( ),
						Func_mis_k2( ),
						Func_mis_oplus( ),
						mis_e_oplus, 
						Func_mis_otimes( ),
						mis_e_otimes, 
						Func_mis_p1( ),
						Func_mis_p2( ),
						d_tree );

  skeleton::time::set_time( "after computation" );
  skeleton::time::print( );

  if ( skeleton::rank == 0 ) {
    std::cout << "result = " << mis.s << std::endl;
  }

  // free all memory
  delete d_tree;
  tree_util::free_all_tree( s_tree );

  return 0;
}
