/***
 * Copyright (c) 2005, SkeTo Project
 * All rights reserved.
 */
/**
 * @file party_planning_problem_binary.cpp
 * @brief A sample program that solves party planning problem
 *        on Binary Trees.
 * @author Kiminori MATSUZAKI
 */

#include <assert.h>
#include <iostream>
#include <algorithm>

#include <mpi.h>

#include <skeleton.h>
#include <rose_tree_skeletons.h>

const int MIN = -1000000;

// Definition exist in <algorithm>
// /**
//  * @brief Return the larger of x and y.
//  */
// template< typename A >
// inline A
// max( const A &x, const A &y )
// { return ( x > y ) ? x : y; }

/**
 * @brief Compute x + y on the domain of R and {-\infty}. 
 */
inline int
plus( const int &x, const int &y )
{
  if ( x == MIN || y == MIN ) return MIN;
  return x + y;
}

/**
 * @brief A pair constructed by two intergers (i, e).
 */
struct ppp_pair {
  int i, e;
  ppp_pair( int i_ = 0, int e_ = 0 )
    : i( i_ ), e( e_ )
  {
  }
  ppp_pair( bool )
    : i( 0 ), e( 0 )
  {
  }
};

/**
 * @brief Simple serializing function. 
 */
std::ostream& operator<< ( std::ostream& sout, const ppp_pair& p ) {
  sout << "(" << p.i << "," << p.e << ")";
  return sout;
}

/**
 * @brief A structure of 2x2 matrix. 
 */
struct ppp_matrix {
  int a, b, c, d;
  ppp_matrix( int a_, int b_, int c_, int d_ )
    : a( a_ ), b( b_ ), c( c_ ), d( d_ )
  {
  }
};

/**
 * @brief A tuple for partial evaluated value of internal node.
 *        (a, b, c, d) represent a unary function, e represents the value of
 *        internal node
 */
struct ppp_5tup {
  ppp_matrix mat;
  int b;
  ppp_5tup( ppp_matrix mat_ = ppp_matrix( 0, 0, 0, 0 ), int b_ = 0 )
    : mat( mat_ ), b( b_ )
  {
  }
  ppp_5tup( bool )
    : mat( ppp_matrix( 0, 0, 0, 0 )), b( 0 )
  {
  }
};

ppp_matrix times( ppp_matrix x, ppp_matrix y ) {
  return ppp_matrix( std::max( plus( x.a, y.a ), plus( x.b, y.c ) ),
		     std::max( plus( x.a, y.b ), plus( x.b, y.d ) ),
		     std::max( plus( x.c, y.a ), plus( x.d, y.c ) ),
		     std::max( plus( x.c, y.b ), plus( x.d, y.d ) ) );
}
  
struct Func_ppp_k1 : unary_function< int, ppp_pair > {
  ppp_pair operator( )( const int &a ) const {
    return ppp_pair( a, 0 );
  }
};

struct Func_ppp_k2 : ternary_function< int, ppp_pair, ppp_pair, ppp_pair > {
  ppp_pair operator( )( const int &b,
			const ppp_pair &l,
			const ppp_pair &r ) const {
    return ppp_pair( plus( plus( b, l.e ), r.e ),
		     plus( std::max( l.i, l.e ), std::max( r.i, r.e ) ) );
  }
};

struct Func_ppp_psi : unary_function< int, ppp_5tup > {
  ppp_5tup operator( )( const int &b ) const {
    return ppp_5tup( ppp_matrix( 0, MIN, MIN, 0 ), b );
  }
};

struct Func_ppp_phiL : ternary_function< ppp_5tup, ppp_pair, ppp_5tup, ppp_5tup > {
  ppp_5tup operator( )( const ppp_5tup p,
			const ppp_pair l,
			const ppp_5tup r ) const {
    ppp_matrix mat2( MIN, plus( p.b, l.e ), std::max( l.i, l.e ), std::max( l.i, l.e ) );
    return ppp_5tup( times( p.mat, times( mat2, r.mat ) ), r.b );
  }
};

struct Func_ppp_phiR : ternary_function< ppp_5tup, ppp_pair, ppp_5tup, ppp_5tup > {
  ppp_5tup operator( )( const ppp_5tup p,
			const ppp_pair r,
			const ppp_5tup l ) const {
    ppp_matrix mat2( MIN, plus( p.b, r.e ), std::max( r.i, r.e ), std::max( r.i, r.e ) );
    return ppp_5tup( times( p.mat, times( mat2, l.mat ) ), l.b );
  }
};

struct Func_ppp_G : ternary_function< ppp_5tup, ppp_pair, ppp_pair, ppp_pair > {
  ppp_pair operator( )( const ppp_5tup p,
			const ppp_pair l,
			const ppp_pair r ) const {
    ppp_pair sub( plus( plus( p.b, l.e ), r.e ),
		  plus( std::max( l.i, l.e ), std::max( r.i, r.e ) ) );
    return ppp_pair( std::max( plus( p.mat.a, sub.i ), plus( p.mat.b, sub.e ) ),
		     std::max( plus( p.mat.c, sub.i ), plus( p.mat.d, sub.e ) ) );
  }
};

struct Func_ppp2_oplus : binary_function< ppp_pair, ppp_pair, ppp_pair > {
  ppp_pair operator( )( const ppp_pair &c,
			const ppp_pair &x ) const {
    return ppp_pair( ( c.i ) ? x.i : x.e,
		     ( c.e ) ? x.i : x.e );
  }
};
const ppp_pair unit_ppp2_oplus( true, false );

struct Func_ppp2_g : unary_function< ppp_pair, ppp_pair > {
  ppp_pair operator( )( const ppp_pair &b ) const {
    return ppp_pair( false, ( b.i > b.e ) );
  }
};

struct Func_ppp_zipfunc : binary_function< ppp_pair, ppp_pair, char > {
  char operator( )( const ppp_pair &b, const ppp_pair &c ) const {
    return ( c.e ) ? 'F' : ( ( b.i > b.e ) ? 'T' : 'F' );
  }
};

int SketoMain( int argc, char *argv[] ) {
  // const char filename_in[]  = "dist_tree_data.txt";
  if ( argc <= 0 + 1 ) {
    throw skeleton::error( 1, "usage: party_planning_problem_binary <inputfile>" );
  }
  
  const char* filename_in = argv[ 1 ];
#ifdef DEBUGOUTPUT
  const char filename_out1[] = "result_of_sum.txt";
  const char filename_out2[] = "result_of_ppp.txt";
#endif /* DEBUGOUTPUT */

  skeleton::time::initialize( );

  dist_tree< int > *tree1
    = dist_tree< int >::read_from_file( filename_in );
  assert( tree1 );

  skeleton::time::set_time( "after read and distribute" );

  dist_tree< ppp_pair > *tree2
    = tree_skeletons::uAcc( Func_ppp_k1( ), Func_ppp_k2( ),
			    Func_ppp_psi( ), Func_ppp_phiL( ),
			    Func_ppp_phiR( ), Func_ppp_G( ),
			    tree1 );
#ifdef DEBUGOUTPUT
    tree2->write_to_file( filename_out1 );
#endif /* DEBUGOUTPUT */

  dist_tree< ppp_pair > *tree3
    = tree_skeletons::dAcc( Func_ppp2_oplus( ), unit_ppp2_oplus,
			    Func_ppp2_g( ), Func_ppp2_g( ),
			    unit_ppp2_oplus,
			    tree2 );
  dist_tree< char > *tree4
    = tree_skeletons::zipwith( Func_ppp_zipfunc( ),
			       Func_ppp_zipfunc( ),
			       tree2,
			       tree3 );
#ifdef DEBUGOUTPUT
    tree4->write_to_file( filename_out2 );
#endif /* DEBUGOUTPUT */

  skeleton::time::set_time( "after computation" );
  skeleton::time::print( );

  skeleton::safe_delete( tree4 );
  skeleton::safe_delete( tree3 );
  skeleton::safe_delete( tree2 );
  skeleton::safe_delete( tree1 );
  return 0;
}
